package sk.tuke.bkpi.extensions;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Map;

public class CompiledImplementationManager {
	
	private Map <String, Hashtable <String, String>> MachineImplementationPairs = new Hashtable <String, Hashtable <String, String>>();
	
	public void resetMachineImplementationPairs ()
	{
		MachineImplementationPairs = new Hashtable <String, Hashtable <String, String>>();
	}
	
	public boolean MachineImplementationPairsEmpty()
	{
		if (this.MachineImplementationPairs.keySet().size() > 0)
		{
			return false;
		}
		else
		{
			return true;
		}
	}
	
	public void ResetMIPairs(String language)
	{
		for (String machine: this.MachineImplementationPairs.keySet())
		{
			RemoveMachineImplementationPair (machine, language);
		}
	}
	
	public boolean CanBeIdentifier (String identifier)
	{
		if (identifier != null)
		{
		  if (identifier.matches("(\\w)+"))
		   {
			return true;
		   }
		  else
		   {
			return false;
		   }
		}
		else
		{
			return false;
		}
		
	}

	public void InsertMachineImplementationPair (String language, String machine, String implementation)
	{
		if (this.MachineImplementationPairs.containsKey(machine))
		{
			if (!this.MachineImplementationPairs.get(machine).containsKey(language))
			{
				this.MachineImplementationPairs.get(machine).put(language, implementation);			
			}
		}
		else
		{
			this.MachineImplementationPairs.put(machine, new Hashtable<String, String>());
			this.MachineImplementationPairs.get(machine).put(language, implementation);
			
		}
	}
	
	public String GetCompiledImplementation (String language, String machine)
	{
		if (this.MachineImplementationPairs.containsKey(machine))
		{
			if (this.MachineImplementationPairs.get(machine).containsKey(language))
			{
				return this.MachineImplementationPairs.get(machine).get(language);
			}
			else
			{
				return null;
			}
		}
		else
		{
			return null;
		}
	}

	
	public boolean ContainsCompiledImplementation (String language, String machine)
	{
		if (this.MachineImplementationPairs.containsKey(machine))
		{
			if (this.MachineImplementationPairs.get(machine).containsKey(language))
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	
	public void RemoveMachineImplementationPair (String machine, String language)
	{
		if (this.MachineImplementationPairs.containsKey(machine))
		{
			if (this.MachineImplementationPairs.get(machine).containsKey(language))
			{
				this.MachineImplementationPairs.get(machine).remove(language);
				
				if (this.MachineImplementationPairs.get(machine).keySet().size() == 0)
				{
					this.MachineImplementationPairs.remove(machine);
				}
			}
		}
	}
	
	public Hashtable<String, String> AllMachineImplementationPairs (String Language)
	{
		Hashtable <String, String> MachineImplementationPairsPerLanguage = new Hashtable <String, String>();
		for (String machine: this.MachineImplementationPairs.keySet())
		{
			for (String language: this.MachineImplementationPairs.get(machine).keySet())
			{
				if (language.equals(Language))
				{
					MachineImplementationPairsPerLanguage.put(machine, this.MachineImplementationPairs.get(machine).get(language));
				}
			}
		}
		return MachineImplementationPairsPerLanguage; 
	}
	
	public ArrayList<String> DefinitionLanguages()
	{
		ArrayList<String> Languages = new ArrayList<String>();
		
		for (String machine : this.MachineImplementationPairs.keySet())
		{
			for (String Language: this.MachineImplementationPairs.get(machine).keySet())
			{
				if (!Languages.contains(Language))
				{
					Languages.add(Language);
				}
			}
		}
		return Languages;
	}

	public boolean CheckMIDefinitionSyntax (String MIDefinition)
	{
		if (MIDefinition.matches("(((\\s)*#LANGUAGES(\\s)*=(\\s)*\\[(\\s)*(\\w)+(\\s)*(,(\\s)*(\\w)+(\\s)*)*\\]((\\s)+#MIPAIR(\\s)*=(\\s)*'(\\w)+'(\\s)*:(\\s)*'(\\w)+'(\\s)*)+)+|(\\s)*)"))
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	
	
	public boolean CheckMIPairDuplicationError (String MIDefinition)
	{	
		if (CheckMIDefinitionSyntax(MIDefinition))
		{	
			String TrimmedMIDefinition = new String();
			
			for (int count = 0; count < MIDefinition.length(); count++)
			{
			
				if (MIDefinition.substring(count, count + 1).matches("[\\w,#\\[\\]=:']"))
				{
					TrimmedMIDefinition = TrimmedMIDefinition + MIDefinition.substring(count, count + 1);
				}
			}
			
			Map <String, Hashtable <String, String>> MIPairs = new Hashtable <String, Hashtable <String, String>>();
			
			String[] MIDefinitionOfLanguageGroups = TrimmedMIDefinition.split("#LANGUAGES"); 
			
			for (String MIDefinitionOfLanguageGroup: MIDefinitionOfLanguageGroups)
			{  
				if (MIDefinitionOfLanguageGroup.startsWith("="))
				{
					
					String languageList = MIDefinitionOfLanguageGroup.substring(MIDefinitionOfLanguageGroup.indexOf("[") + 1, MIDefinitionOfLanguageGroup.indexOf("]"));
					String[] languagesOfLanguageGroup = languageList.split(",");
					
					for (String languageOfLanguageGroup: languagesOfLanguageGroup)
					{
						if (!MIPairs.containsKey(languageOfLanguageGroup))
						{
							MIPairs.put(languageOfLanguageGroup, new Hashtable <String, String>());	
						}
					}
					
					MIDefinitionOfLanguageGroup = MIDefinitionOfLanguageGroup.substring(MIDefinitionOfLanguageGroup.indexOf("]")+1);
					
					String[] MIPairsOfLanguageGroup = MIDefinitionOfLanguageGroup.split("#MIPAIR");
					
					for (String MIPairOfLanguageGroup: MIPairsOfLanguageGroup)
					{
						if (MIPairOfLanguageGroup.startsWith("="))
						{
							String Machine = MIPairOfLanguageGroup.substring(2, (MIPairOfLanguageGroup.indexOf(":") - 1));
							String Implementation = MIPairOfLanguageGroup.substring((MIPairOfLanguageGroup.indexOf(":") + 2), (MIPairOfLanguageGroup.length() - 1));
							
							for (String languageOfLanguageGroup : languagesOfLanguageGroup)
							{
								if (MIPairs.get(languageOfLanguageGroup).containsKey(Machine))
								{
									return false;
								}
								else
								{
									MIPairs.get(languageOfLanguageGroup).put(Machine, Implementation);
								}
							}
	
						}
					}
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	public void ReadMIPairDefinition (String MIDefinition)
	{
		if (CheckMIPairDuplicationError(MIDefinition))
		{
			
			String TrimmedMIDefinition = new String();
			
			for (int count = 0; count < MIDefinition.length(); count++)
			{
			
				if (MIDefinition.substring(count, count + 1).matches("[\\w,#\\[\\]=:']"))
				{
					TrimmedMIDefinition = TrimmedMIDefinition + MIDefinition.substring(count, count + 1);
				}
			}
			
			String[] MIDefinitionOfLanguageGroups = TrimmedMIDefinition.split("#LANGUAGES"); 
			
			for (String MIDefinitionOfLanguageGroup: MIDefinitionOfLanguageGroups)
			{  
				if (MIDefinitionOfLanguageGroup.startsWith("="))
				{
					
					String languageList = MIDefinitionOfLanguageGroup.substring(MIDefinitionOfLanguageGroup.indexOf("[") + 1, MIDefinitionOfLanguageGroup.indexOf("]"));
					String[] languagesOfLanguageGroup = languageList.split(",");	
					
					MIDefinitionOfLanguageGroup = MIDefinitionOfLanguageGroup.substring(MIDefinitionOfLanguageGroup.indexOf("]")+1);
					
					String[] MIPairsOfLanguageGroup = MIDefinitionOfLanguageGroup.split("#MIPAIR");
					
					for (String MIPairOfLanguageGroup: MIPairsOfLanguageGroup)
					{
						if (MIPairOfLanguageGroup.startsWith("="))
						{
							String Machine = MIPairOfLanguageGroup.substring(2, (MIPairOfLanguageGroup.indexOf(":") - 1));
							String Implementation = MIPairOfLanguageGroup.substring((MIPairOfLanguageGroup.indexOf(":") + 2), (MIPairOfLanguageGroup.length() - 1));
							
							for (String languageOfLanguageGroup : languagesOfLanguageGroup)
							{   
								InsertMachineImplementationPair (languageOfLanguageGroup, Machine, Implementation);		         
							}
						}
					}
				}
			}
		}
	}
	
	public void ReadMIPairDefinition (File file)
	{
		StringBuffer contents = new StringBuffer();
		BufferedReader reader = null; 
		try {
		      reader = new BufferedReader(new FileReader(file));
              String text = null;
		      while ((text = reader.readLine()) != null) {
		               contents.append(text)
	                   .append(System.getProperty(
		                        "line.separator"));
		           }
		      
		      ReadMIPairDefinition (contents.toString());
		    } catch (FileNotFoundException e) {
		           e.printStackTrace();
		    } catch (IOException e) {
		            e.printStackTrace();
		    } finally {
		                try {
		                if (reader != null) {
		                    reader.close();
		                }
		              } catch (IOException e) {
		                e.printStackTrace();
		                                      }
		    }
	}
	
	public boolean CheckMIPairDuplicationError (File file)
	{
		StringBuffer contents = new StringBuffer();
		BufferedReader reader = null; 
		try {
		      reader = new BufferedReader(new FileReader(file));
              String text = null;
		      while ((text = reader.readLine()) != null) {
		               contents.append(text)
	                   .append(System.getProperty(
		                        "line.separator"));
		           }
		      
		      return CheckMIPairDuplicationError (contents.toString());
		    } catch (FileNotFoundException e) {
		           e.printStackTrace();
		    } catch (IOException e) {
		            e.printStackTrace();
		    } finally {
		                try {
		                if (reader != null) {
		                    reader.close();
		                }
		              } catch (IOException e) {
		                e.printStackTrace();
		                                      }
		    }
		    return false;
	}
	
	public String WriteMIPairDefinition()
	{
		String MIDefinition = new String ();
		for (String Language: DefinitionLanguages())
		{
			MIDefinition = MIDefinition + "#LANGUAGES = [" + Language + "]\n";
			for (String Machine : this.MachineImplementationPairs.keySet())
			{
				if (this.MachineImplementationPairs.get(Machine).containsKey(Language)) 
				{
					MIDefinition = MIDefinition + "   #MIPAIR = '" + Machine + "' : '" + this.MachineImplementationPairs.get(Machine).get(Language) + "'\n";	
				}
			}	
		}
		return MIDefinition;
	}
	
	public void WriteMIPairDefinition(File file)
	{
		FileWriter outFile;
		try {
			outFile = new FileWriter(file);
			PrintWriter out = new PrintWriter(outFile);
			out.printf(WriteMIPairDefinition());
			out.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
}


