package sk.tuke.bkpi.extensions;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Map;

public class ExcludedWordsManager {
	
	private Map <String, ArrayList<String>> excludedWords = new Hashtable <String, ArrayList<String>>();

	private ArrayList<String> implementedLanguages = new ArrayList<String> ();
	
	public void setImplementedLanguages(ArrayList<String> implementedLanguages) {
		this.implementedLanguages = implementedLanguages;
	}

	public ArrayList<String> getImplementedLanguages() {
		return implementedLanguages;
	}
	
	public ArrayList<String> getExcludedWords (String language)
	{
		if (this.excludedWords.containsKey(language))
		{
			return this.excludedWords.get(language);
		}
		else
		{
			return new ArrayList<String> ();
		}
	}
	
	public boolean CanBeExcludedWord (String word)
	{
		if (word.matches("(\\w)+"))
		{
			return true;
		}
		else
		{
			return false;
		}
				
	}
	
	public void AddExcludedWord (String language, String word)
	{
		if (CanBeExcludedWord(word))
		{
		    if (this.excludedWords.containsKey(language))
		     {		    	
			    if (!this.excludedWords.get(language).contains(word))
			     {
				    this.excludedWords.get(language).add(word);
			     }
		     }
		    else
		     {
			    this.excludedWords.put(language, new ArrayList<String>());
			    this.excludedWords.get(language).add(word);
		     }
		}
	}
	
	public void AddExcludedWord (String word)
	{
		for (String language: implementedLanguages)
		{
			AddExcludedWord (language, word);
		}
	}
	
	public void RemoveExcludedWord (String language, String word)
	{
		if (this.excludedWords.containsKey(language))
		{
			if (this.excludedWords.get(language).contains(word))
			{
				this.excludedWords.get(language).remove(this.excludedWords.get(language).indexOf(word));
				if (this.excludedWords.get(language).size() == 0)
				{
					this.excludedWords.remove(language);
				}
			}
		}
	}
	
	public void RemoveExcludedWord (String word)
	{
		for (String language: implementedLanguages)
		{
			RemoveExcludedWord (language, word);
		}
	}

	public void ResetExcludedWords(String language)
	{
		if (this.excludedWords.containsKey(language))
		{
			this.excludedWords.remove(language);
		}
	}
	
	public void ResetExcludedWords ()
	{
		for (String language: implementedLanguages)
		{
			ResetExcludedWords (language);
		}
	}
	
	public boolean ContainsExcludedWord (String language, String word)
	{
		if (this.excludedWords.containsKey(language))
		{
			if (this.excludedWords.get(language).contains(word))
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	
	public ArrayList<String> FindExcludedWords (String language, String regex)
	{
		ArrayList<String> ResultWords = new ArrayList<String> ();
		if (this.excludedWords.containsKey(language))
		{
			for (String word: this.excludedWords.get(language))
			{
				if (word.matches(regex))
				{
					ResultWords.add(word);
				}
			}	
		}
		return ResultWords;
	}
	
	public boolean checkExcludedWordsDefinition (String excludedWordsDefinition)
	{
		if (excludedWordsDefinition.matches("((\\s)*|((\\s)*#LANGUAGES(\\s)*=(\\s)*\\[(\\s)*(\\w)+(\\s)*(,(\\s)*(\\w)+(\\s)*)*(\\s)*\\](\\s)+#EXCLUDE(\\s)*=(\\s)*\\{(\\s)*((\\s)*|('(\\w)+'(\\s)*(,(\\s)*'(\\w)+'(\\s)*)*))(\\s)*\\}(\\s)*((\\s)+(\\s)*#LANGUAGES(\\s)*=(\\s)*\\[(\\s)*(\\w)+(\\s)*(,(\\s)*(\\w)+(\\s)*)*(\\s)*\\](\\s)+#EXCLUDE(\\s)*=(\\s)*\\{(\\s)*((\\s)*|('(\\w)+'(\\s)*(,(\\s)*'(\\w)+'(\\s)*)*))(\\s)*\\}(\\s)*)*))"))
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	
	public boolean checkExcludedWordsDefinition (File excludedWordsDefinition)
	{
		{
			StringBuffer contents = new StringBuffer();
			BufferedReader reader = null; 
			try {
			      reader = new BufferedReader(new FileReader(excludedWordsDefinition));
	              String text = null;
			      while ((text = reader.readLine()) != null) {
			               contents.append(text)
		                   .append(System.getProperty(
			                        "line.separator"));
			           }
			      return checkExcludedWordsDefinition(contents.toString());
			    } catch (FileNotFoundException e) {
			           e.printStackTrace();
			    } catch (IOException e) {
			            e.printStackTrace();
			    } finally {
			                try {
			                if (reader != null) {
			                    reader.close();
			                }
			              } catch (IOException e) {
			                e.printStackTrace();
			                                      }
			    }
		}
		return false;
	}
	
	public void ReadExcludedWords (String excludedWordsDefinition)
	{
		if (checkExcludedWordsDefinition(excludedWordsDefinition))
		{ 
			ArrayList <String> excludedWordsPerLanguageGroups = new ArrayList<String> (Arrays.asList((excludedWordsDefinition.split("(\\s)*#LANGUAGES(\\s)*=(\\s)*")))); 
			
			for (int count = 0; count < excludedWordsPerLanguageGroups.size(); count++)
			{
				if (!excludedWordsPerLanguageGroups.get(count).contains("#EXCLUDE"))
				{
					excludedWordsPerLanguageGroups.remove(count);
				}

			}
			
			for (String excludedWordsPerLanguageGroup: excludedWordsPerLanguageGroups)
			{
				
					String LanguageList = excludedWordsPerLanguageGroup.substring((excludedWordsPerLanguageGroup.indexOf("[") + 1), excludedWordsPerLanguageGroup.indexOf("]"));
					String TrimmedLanguageList = new String ();
					
					for (int count = 0; count < LanguageList.length(); count++)
					{
						if (LanguageList.substring(count, count + 1).matches("[\\w,]"))
						{
							TrimmedLanguageList = TrimmedLanguageList + LanguageList.substring(count, count + 1);
						}
					}
					
					ArrayList <String> LanguageGroup = new ArrayList<String> (Arrays.asList(TrimmedLanguageList.split(",")));
					
					String ExcludedWordsList = excludedWordsPerLanguageGroup.substring ((excludedWordsPerLanguageGroup.indexOf("{") + 1), excludedWordsPerLanguageGroup.indexOf("}")); 
			        String TrimmedExcludedWordsList = new String(); 
					for (int count = 0; count < ExcludedWordsList.length(); count++)
					{
						if (ExcludedWordsList.substring(count, count + 1).matches("[\\w,']"))
						{
							TrimmedExcludedWordsList = TrimmedExcludedWordsList + ExcludedWordsList.substring(count, count + 1);
						}
						
					}
					
					ArrayList <String> GroupExcludedWords = new ArrayList<String>();
					
					ArrayList <String> PreGroupExcludedWords;
					if (TrimmedExcludedWordsList.length() != 0)
					{
						PreGroupExcludedWords = new ArrayList<String> (Arrays.asList(TrimmedExcludedWordsList.split(",")));
						for (int count = 0; count < PreGroupExcludedWords.size(); count++)
						{
							String ExcludedWord = PreGroupExcludedWords.get(count).substring(1, (PreGroupExcludedWords.get(count).length() - 1));
							GroupExcludedWords.add(ExcludedWord);
						}
					}
					else
					{
						PreGroupExcludedWords = new ArrayList<String>();
					}
					
					
					if (GroupExcludedWords.size() != 0)
					{
						for (String Language: LanguageGroup)
						{
							if (Language.equals("ALL"))
							{
								for (String GroupExcludedWord: GroupExcludedWords)
								{
									AddExcludedWord(GroupExcludedWord);
								}
							}
							else
							{
								for (String GroupExcludedWord: GroupExcludedWords)
								{
									AddExcludedWord(Language, GroupExcludedWord);
								}
							}
						}
					}			
			}
		}
	}
	
	public void ReadExcludedWords (File file)
	{
		StringBuffer contents = new StringBuffer();
		BufferedReader reader = null; 
		try {
		      reader = new BufferedReader(new FileReader(file));
              String text = null;
		      while ((text = reader.readLine()) != null) {
		               contents.append(text)
	                   .append(System.getProperty(
		                        "line.separator"));
		           }
		      
		      ReadExcludedWords (contents.toString());
		    } catch (FileNotFoundException e) {
		           e.printStackTrace();
		    } catch (IOException e) {
		            e.printStackTrace();
		    } finally {
		                try {
		                if (reader != null) {
		                    reader.close();
		                }
		              } catch (IOException e) {
		                e.printStackTrace();
		                                      }
		    }
	}
	
	public String WriteExcludedWords()
	{
		String ExcludedWordsDefinition = new String();
		for (String Language: this.excludedWords.keySet())
		{
			ExcludedWordsDefinition = ExcludedWordsDefinition + "#LANGUAGES = [" + Language + "]\n";
			ExcludedWordsDefinition = ExcludedWordsDefinition + "  #EXCLUDE = {";
			
			for (String LanguageExcludedWord: this.excludedWords.get(Language))
			{
				ExcludedWordsDefinition = ExcludedWordsDefinition + "'" + LanguageExcludedWord + "'" + ",";
			}
			
			if (this.excludedWords.get(Language).size() != 0)
			{
				ExcludedWordsDefinition = ExcludedWordsDefinition.substring(0, (ExcludedWordsDefinition.length() - 1));
			}
			
			ExcludedWordsDefinition = ExcludedWordsDefinition + "}\n";
		}
		return ExcludedWordsDefinition;
	}
	
	public void WriteExcludedWords(File file)
	{
		FileWriter outFile;
		try {
			outFile = new FileWriter(file);
			PrintWriter out = new PrintWriter(outFile);
			out.printf(WriteExcludedWords());
			out.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
		
}
