package sk.tuke.bkpi.generator;

import static java.lang.String.format;
import static sk.tuke.bkpi.model.types.BasicType.BOOL;
import static sk.tuke.bkpi.model.types.BasicType.STRING;
import static sk.tuke.bkpi.model.types.BasicType.Z;
import static sk.tuke.bkpi.model.types.TypeBuilder.CartesianProduct;
import static sk.tuke.bkpi.model.types.TypeBuilder.PowerSet;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.extensions.ClassAccessPolicy;
import sk.tuke.bkpi.extensions.ClassModificationPolicy;
import sk.tuke.bkpi.generator.CsGenerator.CodeGenerator;
import sk.tuke.bkpi.generator.CsGenerator.MethodCodeGenerator;
import sk.tuke.bkpi.generator.CsGenerator.StaticCodeGenerator;
import sk.tuke.bkpi.generator.Utilities.ListFormatter;
import sk.tuke.bkpi.generator.support.AccessPolicyResolver;
import sk.tuke.bkpi.generator.support.AtributeSignature;
import sk.tuke.bkpi.generator.support.MethodSignature;
import sk.tuke.bkpi.generator.support.ModificationPolicyResolver;
import sk.tuke.bkpi.model.Module;
import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.model.b0data.ArrayFieldVariable;
import sk.tuke.bkpi.model.b0data.B0Variable;
import sk.tuke.bkpi.model.b0data.B0VariableReference;
import sk.tuke.bkpi.model.b0data.RecordFieldVariable;
import sk.tuke.bkpi.model.clause.EnumeratedSet;
import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.IncludedMachine;
import sk.tuke.bkpi.model.clause.Operation;
import sk.tuke.bkpi.model.clause.OperationB0;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.clause.Set;
import sk.tuke.bkpi.model.clause.EnumeratedSet.Item;
import sk.tuke.bkpi.model.component.AbstractComponent;
import sk.tuke.bkpi.model.component.Component;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.component.Refinement;
import sk.tuke.bkpi.model.condition.BinaryCondition;
import sk.tuke.bkpi.model.condition.ComparisonCondition;
import sk.tuke.bkpi.model.condition.Condition;
import sk.tuke.bkpi.model.condition.NegationCondition;
import sk.tuke.bkpi.model.condition.ComparisonCondition.Operator;
import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.data.ConcreteConstant;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.ForeignData;
import sk.tuke.bkpi.model.data.ForeignVariable;
import sk.tuke.bkpi.model.data.LocalVariable;
import sk.tuke.bkpi.model.data.OperationInputParameter;
import sk.tuke.bkpi.model.data.OperationOutputParameter;
import sk.tuke.bkpi.model.data.ScalarParameter;
import sk.tuke.bkpi.model.data.SetParameter;
import sk.tuke.bkpi.model.data.StringValue;
import sk.tuke.bkpi.model.instruction.AssertionInstruction;
import sk.tuke.bkpi.model.instruction.BecomesEqualInstruction;
import sk.tuke.bkpi.model.instruction.BlockInstruction;
import sk.tuke.bkpi.model.instruction.CallupInstruction;
import sk.tuke.bkpi.model.instruction.CaseConditionalInstruction;
import sk.tuke.bkpi.model.instruction.IdentityInstruction;
import sk.tuke.bkpi.model.instruction.IfConditionalInstruction;
import sk.tuke.bkpi.model.instruction.Instruction;
import sk.tuke.bkpi.model.instruction.LocalVariableInstruction;
import sk.tuke.bkpi.model.instruction.SequenceInstruction;
import sk.tuke.bkpi.model.instruction.WhileInstruction;
import sk.tuke.bkpi.model.instruction.IfConditionalInstruction.Branch;
import sk.tuke.bkpi.model.term.ArithmeticalBinaryExpression;
import sk.tuke.bkpi.model.term.ArrayExpression;
import sk.tuke.bkpi.model.term.BoolConversionValue;
import sk.tuke.bkpi.model.term.BooleanTerm;
import sk.tuke.bkpi.model.term.IntegerTerm;
import sk.tuke.bkpi.model.term.IntervalB0;
import sk.tuke.bkpi.model.term.NegativeTerm;
import sk.tuke.bkpi.model.term.RecordTerm;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Record;
import sk.tuke.bkpi.model.types.SetType;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeResolver;

public class CsGenerator {
	
	private Project project;
	private File outputFolder;
	
	private ModificationPolicyResolver MPR;
	private AccessPolicyResolver APR;
	//
    private Hashtable <String, ArrayList<String>> projectConsructorAccessRights = new Hashtable <String, ArrayList<String>>();
    private Hashtable <String, ArrayList<String>> projectMethodAccessRights = new Hashtable <String, ArrayList<String>>();
    
    public Hashtable <String, ArrayList<String>> getProjectConstructorAccessRight()
    {
    	return this.projectConsructorAccessRights;
    }
    
    public Hashtable <String, ArrayList<String>> getProjectMethodAccessRight()
    {
    	return this.projectMethodAccessRights;
    }
    
    private void generateAccessRightReport()
    {
    	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
    	  
    	  this.projectConsructorAccessRights.put(nameManager.getUniqueMachineName(), new ArrayList<String>());
    	  this.projectMethodAccessRights.put(nameManager.getUniqueMachineName(), new ArrayList<String>());
    	  
    	  for (Module foreignModule: this.project.modules())
    	  {
    		  for (Import foreignImport: foreignModule.implementation().imports())
    		  {
    			  
    			  if (foreignImport.machine().name().equals(this.module.machine().name()))
    			  {
    				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
    				  { 
    					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    					  this.projectConsructorAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    				  }
    			  }  
    		  }
    	  }
    	  
    		if (project.rootModules().size() > 0) {
    		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
    		  {
    			  this.projectConsructorAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.generateMainMachineName());
    		  }
    		}
        	
        	
        	for (Module Pmodule: project.modules())
        	{
        		this.projectMethodAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.getUniqueMachineName(Pmodule.machine().name()));
        	}
        	this.projectMethodAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.generateMainMachineName());
    	
    }
	//
	public void setAccessPolicy (ArrayList <ClassAccessPolicy> languageAccessPolicy)
	{
		APR = new AccessPolicyResolver (languageAccessPolicy);
	}
	
	public void setModificationPolicy(ArrayList <ClassModificationPolicy> languageModificationPolicy)
	{
		MPR = new ModificationPolicyResolver (languageModificationPolicy);
	}
	
	private static final ArrayList <String> csKeywords = new ArrayList<String>(Arrays.asList ("abstract", 
			                                    "as", "base", "bool", "break", "byte", "case", "while",
		                                        "catch", "char", "checked", "class", "const", "continue", 
		                                        "decimal", "default", "delegate", "do", "double", "else", 
		                                        "enum", "event", "explicit", "extern", "false", "finally",
		                                        "fixed", "float", "for", "foreach", "goto", "if", "implicit",
		                                        "in", "int", "interface", "internal", "is", "lock", "long",
		                                        "namespace", "new", "null", "object", "operator", "out",
		                                        "override", "params", "private", "protected", "public", 
		                                        "readonly", "ref", "return", "sbyte", "sealed", "short",
		                                        "sizeof", "stackalloc", "static", "string", "struct", "switch",
		                                        "this", "throw", "true", "try", "typeof", "uint", "ulong",
		                                        "unchecked", "unsafe", "ushort", "using", "virtual", "void",
		                                        "volatile"));
	
	private static ArrayList<String> csExcludedWords = new ArrayList<String>();
	
	private Implementation implementation;
	private Module module;
	private StringWriter buffer;
	private JavaWriter out;
	
	private Map<DataField, DataField> includedMapping;
	private Map<EnumeratedSet, Module> setsMapping;
	
	private String namespaceName;
	
	private NameManager nameManager;
	
	private Map<Type, String> typesAssociationTable = new Hashtable<Type, String>();
	
	private ArrayGenerator2 arrayGenerator = new CsNativeArrayGenerator(this);
	private RecordGenerator recordGenerator = new RecordGenerator2(this);
	
	public void setCsExcludedWords(ArrayList<String> csExcludedWords)
	{
		this.csExcludedWords = csExcludedWords;
	}
	
	public void ResetCsExcludedWords()
	{
		this.csExcludedWords = new ArrayList<String>();
	}
	
	public CsGenerator() {}
	
	public boolean is_keyword (String word)
	{
		if (csKeywords.contains(word))
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	
	public CsGenerator(Project project, File outputFolder) {
		this.project = project;
		this.outputFolder = outputFolder;
	}
	
	public CsGenerator(Project project, File outputFolder, String[] excludedWords) {
		this.project = project;
		this.outputFolder = outputFolder;
		csExcludedWords.addAll(Arrays.asList(excludedWords));
	}
	
	public CsGenerator(Implementation implementation, Writer out) {
		this.implementation = implementation;
		this.out = new JavaWriter(out);
	}
	
	public CsGenerator(Implementation implementation) {
		this.implementation = implementation;
		buffer = new StringWriter();
		out = new JavaWriter(new PrintWriter(buffer));
	}

	
	JavaWriter writter() {
		return out;
	}
	
	public Implementation implementation() {
		return implementation;
	}
	
	public PrintWriter newFileStream(String name) throws IOException {
		return new PrintWriter(new FileWriter(outputFolder.getAbsolutePath()+File.separator+name+".cs"));
	}

	public void generateCsCode(Writer writer) throws CsCodeGenerationException {
		
		if (!outputFolder.exists()) {
			if (!outputFolder.mkdir()) {
				throw new RuntimeException("Cannot create output directory");
			}
		}
		
		
		Module mainModule = null;
		String mainOperationName = null;
		if (project.rootModules().size() > 0) {
			mainModule = project.rootModules().get(0);
			for (Operation op : mainModule.machine().operations()) {
				if (op.inputParameters().size() == 0) {
					mainOperationName = op.name();
					break;
				}
			}
		}
		
		
		includedMapping = new Hashtable<DataField, DataField>();
		setsMapping = new Hashtable<EnumeratedSet, Module>();
		nameManager = new NameManager();
		nameManager.setKeywords(csExcludedWords);
		nameManager.generateUniqueMachineNames(this.project);
		nameManager.resolveOperationNames(this.project);
		nameManager.SetRecordExcludedWords();
		
		if (mainModule != null && mainOperationName != null) {
			String mainClassName = this.nameManager.generateMainMachineName();
			File mainClass = new File(outputFolder.getAbsolutePath()+File.separator+(csKeywords.contains(mainClassName) ? ("@" + mainClassName) : mainClassName)+".cs");
			try {
				FileOutputStream fileOut = new FileOutputStream(mainClass);
				String main = "class "+(csExcludedWords.contains(mainClassName) ? ("@" + mainClassName) : mainClassName)+" {\n\n"+
					"\tstatic void Main(string[] args) {\n"+ 
					"\t\tnew "+(csKeywords.contains(nameManager.getUniqueMachineName(mainModule.machine().name())) ? ("@" + nameManager.getUniqueMachineName(mainModule.machine().name())) : nameManager.getUniqueMachineName(mainModule.machine().name()))+"()."+ (csKeywords.contains(nameManager.get_resolved_operation_name(mainModule.machine(), mainOperationName)) ? ("@" + nameManager.get_resolved_operation_name(mainModule.machine(), mainOperationName)) : nameManager.get_resolved_operation_name(mainModule.machine(), mainOperationName)) +"();\n"+ 
					"\t}\n"+
					"}"; 
				fileOut.write(main.getBytes());
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		/*NOT SURE WHAT IT DOES, SO CAN'T GUARATEE SAFETY BEGIN*/
		for (Module module : project.modules()) {
			List<DataField> moduleDataFields = new ArrayList<DataField>();
			Implementation implementation = module.implementation();
			
			for (Component c : module.components()) {
				for (EnumeratedSet set : c.enumeratedSets()) {
					setsMapping.put(set, module);
				}
				moduleDataFields.addAll(c.concreteConstants());
				for (Set set : c.sets()) {
					moduleDataFields.add(set);
					if (set instanceof EnumeratedSet) {
						moduleDataFields.addAll(((EnumeratedSet) set).items());
					}
				}
				moduleDataFields.addAll(c.concreteVariables());
			}
			for (DataField data : moduleDataFields) {
				nameManager.moduleData(module, data);
			}
			
			if (implementation != null) {
				for (Import imp : implementation.imports()) { 
					nameManager.componentReferenceName(module, imp);
				}
			}
			for (SeenMachine seen : module.lastComponent().seenMachines()) {
				nameManager.componentReferenceName(module, seen);
			}
		}
		/*NOT SURE WHAT IT DOES, SO CAN'T GUARATEE SAFETY END*/	
		
		for (Module module : project.modules()) {
			if (module.lastComponent() instanceof Implementation) {

				File file = new File(outputFolder.getAbsolutePath()+File.separator+(csKeywords.contains(nameManager.getUniqueMachineName(module.machine().name())) ? ("@" + nameManager.getUniqueMachineName(module.machine().name())) : nameManager.getUniqueMachineName(module.machine().name()))+".cs");
				FileWriter fileOut = null;
				System.out.println(file.getAbsolutePath());
				
				try {
					fileOut = new FileWriter(file);
				} catch (FileNotFoundException e) {
					e.printStackTrace();
				} catch (IOException e) {
					e.printStackTrace();
				}
				
				buffer = new StringWriter();
				out = new JavaWriter(fileOut);
				
				try {
					generateCsCode(module, out);
				} catch (CsCodeGenerationException jcge) {
					if (jcge.getSource() == null) {
						jcge.setSource(module);
					}
					throw jcge;
				} catch (Exception e) {
					throw new CsCodeGenerationException(e, module);
				}
				
				try {
					fileOut.close();
				} catch (IOException e) {}		
			}
		}
	}
	
	public NameManager getNameManager() {
		return nameManager; 
	}
	
	public void generateCsCode(Module module, Writer writer) {
		this.module = module;
		this.implementation = module.implementation();
		nameManager.setWorkingModule(module);
		out = new JavaWriter(new PrintWriter(writer));
		generate();
	}
	
	private void find(DataField data) {
		List<Module> references = new ArrayList<Module>();
		DataField d = checkImported(module, data, references);
		references.remove(module);
		out.println("//"+references + " : "+d);
		out.println("//"+data.getClass() + " -> "+d.getClass() + "  "+data.hashCode() + " - "+ d.hashCode());
		if (d != null) {
			includedMapping.put(data, d);
		}
	}
	
	private DataField checkImported(Module module, DataField data, List<Module> references) {
		if (module.machine().dataByName(data.name()) != null) {
			references.add(module);
			return module.machine().dataByName(data.name());
		}
		if (module.implementation() != null) {
			for (Import imp : module.implementation().imports()) {
				DataField found = checkImported(project.moduleOfMachine(imp.machine()), data, references);
				if (found != null) {
					references.add(0, module);
					return found;
				}
			}
		}
		return null;
	}
	
    ArrayList <MethodSignature> methods = new ArrayList <MethodSignature>();
    ArrayList <AtributeSignature> atributes = new ArrayList <AtributeSignature>();
    ArrayList <MethodSignature> constructors = new ArrayList <MethodSignature>();
    
    // P = 0 NO POLICY APLIED
    // P = 1 FORBID REMOVING
    // P = 2 FORBID ADDING
    // P = 3 FORBID ADDING AND REMOVING
    
    private void generateModificationPolicyEnforcementMethod(int AP, int MP, int CP)
    {
        if ((AP != 0) || (MP != 0))
        {
        
      	  if (AP != 0)
    	  {  
    		  /*GENERATE CLASS AtributeSignature BEGIN*/   
    		  
    		  out.printf("private class %s\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private bool isStatic;\n");
    		  out.printf("private String type;\n");
    		  out.printf("private String name;\n\n");
 
    		  
    		  out.printf("public %s (bool isStatic, String type, String name)\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.isStatic = isStatic;\n");
    		    out.printf("this.type = type;\n");
    		    out.printf("this.name = name;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public bool getIsStatic()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.isStatic;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getType()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.type;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getName()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.name;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public override bool Equals (object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O is %s)\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s A = (%s) O;\n", nameManager.generateUniqueClassName("AtributeSignature"), nameManager.generateUniqueClassName("AtributeSignature"));
                  out.printf("if ((A.getIsStatic() == this.isStatic) && (A.getType().Equals(this.type)) && (A.getName().Equals(this.name)))\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return base.Equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n"); 	
    		  
    		  out.printf("public override int GetHashCode()\n");
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("return base.GetHashCode();\n");
    		  
    		  out.unindent();
    		  out.printf("}\n");   		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS AtributeSignature END*/   		  
    	  }
 
    	  if (MP != 0)
    	  {
    		  /*GENERATE CLASS MethodSignature BEGIN*/
    		  
    		  out.printf("private class %s\n", nameManager.generateUniqueClassName("MethodSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private bool isStatic;\n");
    		  out.printf("private String returnType;\n");
    		  out.printf("private String name;\n");
    		  out.printf("private List<String> parameterTypes;\n\n");
    		  
    		  out.printf("public %s (bool isStatic, String returnType, String name, List<String> parameterTypes)\n", nameManager.generateUniqueClassName("MethodSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.isStatic = isStatic;\n");
    		    out.printf("this.returnType = returnType;\n");
    		    out.printf("this.name = name;\n");
    		    out.printf("this.parameterTypes = parameterTypes;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public bool getIsStatic()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.isStatic;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getReturnType()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.returnType;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getName()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.name;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public List<String> getParameterTypes()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.parameterTypes;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public override bool Equals (object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O is %s)\n", nameManager.generateUniqueClassName("MethodSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s M = (%s) O;\n", nameManager.generateUniqueClassName("MethodSignature"), nameManager.generateUniqueClassName("MethodSignature"));
                  out.printf("if ((M.getIsStatic() == this.isStatic) && (M.getReturnType().Equals(this.returnType)) && (M.getName().Equals(this.name)))\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("List <String> FA = new List <String> (M.getParameterTypes());\n");
    		        
    		        out.printf("foreach (String TAE in this.parameterTypes)\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
        		      out.printf("if (!FA.Contains(TAE))\n");
          		      out.printf("{\n");
        		      out.indent(); 
        		      
        		        out.printf("return false;\n");
        		      
          		      out.unindent();
        		      out.printf("}\n"); 
        		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("FA.Remove(TAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		      
        		    
        		    out.unindent();
        		    out.printf("}\n\n"); 
        		    
        		    out.printf("List <String> TA = new List <String> (this.parameterTypes);\n");
        		    
    		        out.printf("foreach (String FAE in M.getParameterTypes())\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
      		          out.printf("if (!TA.Contains(FAE))\n");
      		          out.printf("{\n");
    		          out.indent(); 
    		      
    		          out.printf("return false;\n");
    		      
      		          out.unindent();
    		          out.printf("}\n");  
    		          
           		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("TA.Remove(FAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		    
        		    out.unindent();
        		    out.printf("}\n\n");   
        		    
        		    out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return base.Equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n"); 	
    		  
    		  out.printf("public override int GetHashCode()\n");
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("return base.GetHashCode();\n");
    		  
    		  out.unindent();
    		  out.printf("}\n");      		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS MethodSignature END*/
    	  }
    	  
    	  if (CP != 0)
    	  {
    		  /*GENERATE CLASS ConstructorSignature BEGIN*/
    		  
    		  out.printf("private class %s\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private bool isStatic;\n");
    		  out.printf("private List<String> parameterTypes;\n\n");
    		  
    		  out.printf("public %s (bool isStatic, List<String> parameterTypes)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.isStatic = isStatic;\n");
    		    out.printf("this.parameterTypes = parameterTypes;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public bool getIsStatic()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.isStatic;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public List<String> getParameterTypes()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.parameterTypes;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public override bool Equals (object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O is %s)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s C = (%s) O;\n", nameManager.generateUniqueClassName("ConstructorSignature"), nameManager.generateUniqueClassName("ConstructorSignature"));
                  out.printf("if (C.getIsStatic() == this.isStatic)\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("List <String> FA = new List <String> (C.getParameterTypes());\n");
    		        
    		        out.printf("foreach (String TAE in this.parameterTypes)\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
        		      out.printf("if (!FA.Contains(TAE))\n");
          		      out.printf("{\n");
        		      out.indent(); 
        		      
        		        out.printf("return false;\n");
        		      
          		      out.unindent();
        		      out.printf("}\n"); 
        		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("FA.Remove(TAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		      
        		    
        		    out.unindent();
        		    out.printf("}\n\n"); 
        		    
        		    out.printf("List <String> TA = new List <String> (this.parameterTypes);\n");
        		    
    		        out.printf("foreach (String FAE in C.getParameterTypes())\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
      		          out.printf("if (!TA.Contains(FAE))\n");
      		          out.printf("{\n");
    		          out.indent(); 
    		      
    		          out.printf("return false;\n");
    		      
      		          out.unindent();
    		          out.printf("}\n");  
    		          
           		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("TA.Remove(FAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		    
        		    out.unindent();
        		    out.printf("}\n\n");   
        		    
        		    out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return base.Equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n"); 	
    		  
    		  out.printf("public override int GetHashCode()\n");
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("return base.GetHashCode();\n");
    		  
    		  out.unindent();
    		  out.printf("}\n");      		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS ConstructorSignature END*/
    	  }    	  
    	  
      	  out.printf("private static void %s ()\n", nameManager.generate_unique_method_name(this.implementation, "checkClassStructure"));
    	  out.printf("{\n");
    	  out.indent();

    	  
    	    if (MP != 0)
    	    {
    	    	out.printf("List <%s> generatedMethods = new List <%s>();\n\n", nameManager.generateUniqueClassName("MethodSignature"), nameManager.generateUniqueClassName("MethodSignature"));
    	    	
    	    	Integer count = 0;
    	    	for (MethodSignature method: this.methods)
    	    	{
    	    		String outParameters = new String();
    	    		
                    for (String parameterType: method.parameterTypes)
                    {
                    	outParameters = outParameters + "\""+ parameterType +"\", "; 
                    }
                    
                    if (method.parameterTypes.size() > 0)
                    {
                    	outParameters = outParameters.substring(0, (outParameters.length() - 2));
                    	outParameters = "String[] parameterTypes" + count.toString() + "= {" + outParameters + "};";
                    	out.printf("%s\n", outParameters);
                    }
                                    
                    
    	    		out.printf("generatedMethods.Add (new %s(%b, \"%s\", \"%s\", new List<String>(%s)));\n", nameManager.generateUniqueClassName("MethodSignature")
    	    				                                                 , (method.modifiers == 0 ? false : true)
    	    				                                                 , method.returnType
    	    				                                                 , method.name
    	    				                                                 ,((method.parameterTypes.size() > 0) ? ("parameterTypes" + count.toString()) : ""));
    	    		if (method.parameterTypes.size() > 0)
    	    		{
    	    		  count++;
    	    		}
    	    	}
    	    	
    	    	switch (MP) 
    	    	{
    	    	
    	    	case 1 : {
    	    		       
    	    		       out.printf("foreach (%s generatedMethod in generatedMethods)\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		         out.printf("Boolean containsMethod = false;\n");
    	    		         out.printf("foreach (MethodInfo classMethod in typeof(%s).GetMethods())\n", nameManager.getUniqueMachineName());
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("List <String> parameterTypes = new List <String>();\n");
    	    		           
    	    		           out.printf("foreach (ParameterInfo  parameterType in classMethod.GetParameters())\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n\n");
    	    		           
    	    		           out.printf("if (generatedMethod.Equals (new %s(classMethod.IsStatic, classMethod.ReturnType.Name, classMethod.Name, parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("containsMethod = true;\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n");
    	    		           
    	    		           
    	    		           
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		         out.printf("if (containsMethod == false)\n");
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.Console.Write (\"Missing \" + (generatedMethod.getIsStatic() ? \"static\" : \"non-static\" ) + \" method: \" + generatedMethod.getReturnType() + \" \" + generatedMethod.getName());\n" );
    	    		           out.printf("generatedMethod.getParameterTypes().ForEach(delegate (String parType)\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("System.Console.Write(\" \" + parType);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("});\n");
    	    		           
    	    		           out.printf("System.Console.WriteLine();\n");
    	    		           
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");
    	    		
    	    		
    	    	         } break;
    	    	
    	    	case 2 : {
    	    		       out.printf("foreach (MethodInfo classMethod in typeof(%s).GetMethods())\n", nameManager.getUniqueMachineName());
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		        out.printf("if (!classMethod.DeclaringType.ToString().Equals(\"System.Object\"))\n");
    	    		        out.printf("{\n");
    	    		        out.indent();
    	    		       
	    		             out.printf("List <String> parameterTypes = new List <String>();\n");
	    		           
	    		             out.printf("foreach (ParameterInfo  parameterType in classMethod.GetParameters())\n");
	    		             out.printf("{\n");
	    		             out.indent();
	    		           
	    		               out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
	    		           
	    		             out.unindent();
	    		             out.printf("}\n\n");
    	    		       
    	    		         out.printf("if (!generatedMethods.Contains(new %s(classMethod.IsStatic, classMethod.ReturnType.Name, classMethod.Name, parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.Console.Write (\"Forbidden \" + (classMethod.IsStatic ? \"static\" : \"non-static\" ) + \" method: \" + classMethod.ReturnType.Name + \" \" + classMethod.Name);\n");
    	    		           
    	    		           out.printf("parameterTypes.ForEach(delegate (String parType)\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("System.Console.Write(\" \" + parType);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("});\n");
    	    		           
    	    		           out.printf("System.Console.WriteLine();\n");
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		       
    	    		        out.unindent(); 
    	    		        out.printf("}\n");
    	    		        
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");		
    	    		
    	    	         } break;
    	    	
    	    	case 3 :  {
    	    		
	    		       out.printf("foreach (%s generatedMethod in generatedMethods)\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsMethod = false;\n");
	    		         out.printf("foreach (MethodInfo classMethod in typeof(%s).GetMethods())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("List <String> parameterTypes = new List <String>();\n");
	    		           
	    		           out.printf("foreach (ParameterInfo  parameterType in classMethod.GetParameters())\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n\n");
	    		           
	    		           out.printf("if (generatedMethod.Equals (new %s(classMethod.IsStatic, classMethod.ReturnType.Name, classMethod.Name, parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsMethod = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsMethod == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.Write (\"Missing \" + (generatedMethod.getIsStatic() ? \"static\" : \"non-static\" ) + \" method: \" + generatedMethod.getReturnType() + \" \" + generatedMethod.getName());\n" );
	    		           out.printf("generatedMethod.getParameterTypes().ForEach(delegate (String parType)\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("System.Console.Write(\" \" + parType);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("});\n");
	    		           
	    		           out.printf("System.Console.WriteLine();\n");
	    		           
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");
	    		       
	    		       out.printf("foreach (MethodInfo classMethod in typeof(%s).GetMethods())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classMethod.DeclaringType.ToString().Equals(\"System.Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       
    		             out.printf("List <String> parameterTypes = new List <String>();\n");
    		           
    		             out.printf("foreach (ParameterInfo  parameterType in classMethod.GetParameters())\n");
    		             out.printf("{\n");
    		             out.indent();
    		           
    		               out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
    		           
    		             out.unindent();
    		             out.printf("}\n\n");
	    		       
	    		         out.printf("if (!generatedMethods.Contains(new %s(classMethod.IsStatic, classMethod.ReturnType.Name, classMethod.Name, parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.Write (\"Forbidden \" + (classMethod.IsStatic ? \"static\" : \"non-static\" ) + \" method: \" + classMethod.ReturnType.Name + \" \" + classMethod.Name);\n");
	    		           
	    		           out.printf("parameterTypes.ForEach(delegate (String parType)\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("System.Console.Write(\" \" + parType);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("});\n");
	    		           
	    		           out.printf("System.Console.WriteLine();\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");		    		       
	    		
	    		

    	    	          }break;
    	    	
    	    	
    	    	default : break;
    	        
    	    	}
	
    	    }
    	    
    	    if (CP != 0)
    	    {
    	    	out.printf("List <%s> generatedConstructors = new List <%s>();\n\n", nameManager.generateUniqueClassName("ConstructorSignature"), nameManager.generateUniqueClassName("ConstructorSignature"));
    	    	
    	    	Integer count = 0;
    	    	for (MethodSignature constructor: this.methods)
    	    	{
    	    		String outParameters = new String();
    	    		
                    for (String parameterType: constructor.parameterTypes)
                    {
                    	outParameters = outParameters + "\""+ parameterType +"\", "; 
                    }
                    
                    if (constructor.parameterTypes.size() > 0)
                    {
                    	outParameters = outParameters.substring(0, (outParameters.length() - 2));
                    	outParameters = "String[] constructorParameterTypes" + count.toString() + "= {" + outParameters + "};";
                    	out.printf("%s\n", outParameters);
                    }
                                    
                    
    	    		out.printf("generatedConstructors.Add (new %s(%b, new List<String>(%s)));\n", nameManager.generateUniqueClassName("ConstructorSignature")
    	    				                                                 , (constructor.modifiers == 0 ? false : true)
    	    				                                                 ,((constructor.parameterTypes.size() > 0) ? ("constructorParameterTypes" + count.toString()) : ""));
    	    		if (constructor.parameterTypes.size() > 0)
    	    		{
    	    		  count++;
    	    		}
    	    	}
    	    	
    	    	switch (CP) 
    	    	{
    	    	
    	    	case 1 : {
    	    		       
    	    		       out.printf("foreach (%s generatedConstructor in generatedConstructors)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		         out.printf("Boolean containsConstructor = false;\n");
    	    		         out.printf("foreach (ConstructorInfo classConstructor in typeof(%s).GetConstructors())\n", nameManager.getUniqueMachineName());
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("List <String> parameterTypes = new List <String>();\n");
    	    		           
    	    		           out.printf("foreach (ParameterInfo  parameterType in classConstructor.GetParameters())\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n\n");
    	    		           
    	    		           out.printf("if (generatedConstructor.Equals (new %s(classConstructor.IsStatic, parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("containsConstructor = true;\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n");
    	    		           
    	    		           
    	    		           
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		         out.printf("if (containsConstructor == false)\n");
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.Console.Write (\"Missing \" + (generatedConstructor.getIsStatic() ? \"static\" : \"non-static\" ) + \" constructor: %s\");\n", nameManager.getUniqueMachineName());
    	    		           out.printf("generatedConstructor.getParameterTypes().ForEach(delegate (String parType)\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("System.Console.Write(\" \" + parType);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("});\n");
    	    		           
    	    		           out.printf("System.Console.WriteLine();\n");
    	    		           
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");
    	    		
    	    		
    	    	         } break;
    	    	
    	    	case 2 : {
    	    		       out.printf("foreach (ConstructorInfo classConstructor in typeof(%s).GetConstructors())\n", nameManager.getUniqueMachineName());
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		        out.printf("if (!classConstructor.DeclaringType.ToString().Equals(\"System.Object\"))\n");
    	    		        out.printf("{\n");
    	    		        out.indent();
    	    		       
	    		             out.printf("List <String> parameterTypes = new List <String>();\n");
	    		           
	    		             out.printf("foreach (ParameterInfo  parameterType in classConstructor.GetParameters())\n");
	    		             out.printf("{\n");
	    		             out.indent();
	    		           
	    		               out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
	    		           
	    		             out.unindent();
	    		             out.printf("}\n\n");
    	    		       
    	    		         out.printf("if (!generatedConstructors.Contains(new %s(classConstructor.IsStatic, parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.Console.Write (\"Forbidden \" + (classConstructor.IsStatic ? \"static\" : \"non-static\" ) + \" constructor: %s\");\n", nameManager.getUniqueMachineName());
    	    		           
    	    		           out.printf("parameterTypes.ForEach(delegate (String parType)\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("System.Console.Write(\" \" + parType);\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("});\n");
    	    		           
    	    		           out.printf("System.Console.WriteLine();\n");
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		       
    	    		        out.unindent(); 
    	    		        out.printf("}\n");
    	    		        
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");		
    	    		
    	    	         } break;
    	    	
    	    	case 3 :  {
    	    		
	    		       out.printf("foreach (%s generatedConstructor in generatedConstructors)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsConstructor = false;\n");
	    		         out.printf("foreach (ConstructorInfo classConstructor in typeof(%s).GetConstructors())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("List <String> parameterTypes = new List <String>();\n");
	    		           
	    		           out.printf("foreach (ParameterInfo  parameterType in classConstructor.GetParameters())\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n\n");
	    		           
	    		           out.printf("if (generatedConstructor.Equals (new %s(classConstructor.IsStatic, parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsConstructor = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsConstructor == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.Write (\"Missing \" + (generatedConstructor.getIsStatic() ? \"static\" : \"non-static\" ) + \" constructor: %s\");\n", nameManager.getUniqueMachineName());
	    		           out.printf("generatedConstructor.getParameterTypes().ForEach(delegate (String parType)\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("System.Console.Write(\" \" + parType);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("});\n");
	    		           
	    		           out.printf("System.Console.WriteLine();\n");
	    		           
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");

	    		       
	    		       out.printf("foreach (ConstructorInfo classConstructor in typeof(%s).GetConstructors())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classConstructor.DeclaringType.ToString().Equals(\"System.Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       
    		             out.printf("List <String> parameterTypes = new List <String>();\n");
    		           
    		             out.printf("foreach (ParameterInfo  parameterType in classConstructor.GetParameters())\n");
    		             out.printf("{\n");
    		             out.indent();
    		           
    		               out.printf("parameterTypes.Add(parameterType.ParameterType.Name);\n");
    		           
    		             out.unindent();
    		             out.printf("}\n\n");
	    		       
	    		         out.printf("if (!generatedConstructors.Contains(new %s(classConstructor.IsStatic, parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.Write (\"Forbidden \" + (classConstructor.IsStatic ? \"static\" : \"non-static\" ) + \" constructor: %s\");\n", nameManager.getUniqueMachineName());
	    		           
	    		           out.printf("parameterTypes.ForEach(delegate (String parType)\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("System.Console.Write(\" \" + parType);\n");
	    		           
	    		           out.unindent();
	    		           out.printf("});\n");
	    		           
	    		           out.printf("System.Console.WriteLine();\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");		    		       
	    		       

    	    	          }break;
    	    	
    	    	
    	    	default : break;
    	        
    	    	}
	
    	    }    	    
    	    
    	    if (AP != 0)
    	    {
    	    	
    	    	out.printf("List <%s> generatedAtributes = new List <%s>();\n\n", nameManager.generateUniqueClassName("AtributeSignature"), nameManager.generateUniqueClassName("AtributeSignature"));
    	    	
    	    	for (AtributeSignature atribute: this.atributes)
    	    	{
    	    		if (atribute.modifiers != 2)
    	    		{
    	    		out.printf("generatedAtributes.Add(new %s(%b, \"%s\", \"%s\"));\n", nameManager.generateUniqueClassName("AtributeSignature")
    	    				                                                  , (atribute.modifiers == 1 ? true : false)
    	    				                                                  , atribute.type
    	    				                                                  , atribute.name);
    	    		}
    	    	}
    	    	
    	    	switch (AP)
    	    	{
    	    	case 1 : {
    	    		
	    		       out.printf("foreach (%s generatedAtribute in generatedAtributes)\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsAtribute = false;\n");
	    		         out.printf("foreach (FieldInfo classAtribute in typeof (%s).GetFields())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		      	    		           
	    		           out.printf("if (generatedAtribute.Equals (new %s(classAtribute.IsStatic, classAtribute.FieldType.Name, classAtribute.Name)))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsAtribute = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsAtribute == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.WriteLine (\"Missing \" + (generatedAtribute.getIsStatic() ? \"static\" : \"non-static\") + \" atribute: \"  + generatedAtribute.getType() + \" \" + generatedAtribute.getName());\n" );
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");    	    		
    	    		        
    	    	         }break;
    	    	
    	    	case 2: {
    	    		
	    		       out.printf("foreach (FieldInfo classAtribute in typeof (%s).GetFields())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classAtribute.DeclaringType.ToString().Equals(\"System.Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       	       
	    		         out.printf("if (!generatedAtributes.Contains(new %s(classAtribute.IsStatic, classAtribute.FieldType.Name, classAtribute.Name)))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.WriteLine (\"Forbidden \" + (classAtribute.IsStatic ? \"static\" : \"non-static\") + \" atribute: \" + classAtribute.FieldType.Name + \" \" + classAtribute.Name);\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	    	    		
    	    		 
    	    	        }break;
    	    	
    	    	case 3: { 
    	    		
	    		       out.printf("foreach (%s generatedAtribute in generatedAtributes)\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsAtribute = false;\n");
	    		         out.printf("foreach (FieldInfo classAtribute in typeof (%s).GetFields())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		      	    		           
	    		           out.printf("if (generatedAtribute.Equals (new %s(classAtribute.IsStatic, classAtribute.FieldType.Name, classAtribute.Name)))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsAtribute = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsAtribute == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.WriteLine (\"Missing \" + (generatedAtribute.getIsStatic() ? \"static\" : \"non-static\") + \" atribute: \"  + generatedAtribute.getType() + \" \" + generatedAtribute.getName());\n" );
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");   
	    		       
	    		       out.printf("foreach (FieldInfo classAtribute in typeof (%s).GetFields())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classAtribute.DeclaringType.ToString().Equals(\"System.Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       	       
	    		         out.printf("if (!generatedAtributes.Contains(new %s(classAtribute.IsStatic, classAtribute.FieldType.Name, classAtribute.Name)))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.Console.WriteLine (\"Forbidden \" + (classAtribute.IsStatic ? \"static\" : \"non-static\") + \" atribute: \" + classAtribute.FieldType.Name + \" \" + classAtribute.Name);\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	  
    	    	
    	    	        } break;
    	    	
    	    	default: break;
    	    	
    	    	}
    	    		
    	    }

      	  out.unindent();
    	  out.printf("}\n");    	  
    	  
    	  
  
    	  
    	  
        	
        	
        	
        	
        	
        	
        	
        	
        	
        	
        	
        	
        	
        }
    }
    
    
    
    private void generateConstructorAccessCheck ()
    {
    	
    	out.printf("private static void %s ()\n", nameManager.generate_unique_method_name(this.implementation, "checkConstructorAccessRight"));
    	out.printf("{\n");
    	out.indent();
    	
    	  out.printf("List <String> constructorAccessClasses = new List <String>();\n");
    	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
    	  
    	  for (Module foreignModule: this.project.modules())
    	  {
    		  for (Import foreignImport: foreignModule.implementation().imports())
    		  {
    			  
    			  if (foreignImport.machine().name().equals(this.module.machine().name()))
    			  {
    				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
    				  { 
    					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    					  out.printf("constructorAccessClasses.Add(\"%s\");\n", nameManager.getUniqueMachineName(foreignModule.machine().name()));
    				  }
    			  }  
    		  }
    	  }
    	  
  		if (project.rootModules().size() > 0) {
  		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
  		  {
  			out.printf("constructorAccessClasses.Add(\"%s\");\n", nameManager.generateMainMachineName());  
  		  }
  		}
			
    	  out.printf("if (!constructorAccessClasses.Contains(new StackFrame(2).GetMethod().DeclaringType.Name))\n");
    	  out.printf("{\n");
    	  out.indent();
    	  
    	    out.printf("System.Console.WriteLine (\"Class \" + new StackFrame(2).GetMethod().DeclaringType.Name + \" cannot create instance of class %s\");\n", nameManager.getUniqueMachineName());
    	  
    	  out.unindent();
    	  out.printf("}\n");
    	
    	out.unindent();
    	out.printf("}\n");
        	
    }
    
    
    private void generateMethodAccessRight ()
    {   
    	
    	out.printf("private static void %s(String methodName)\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
    	out.printf("{\n");
    	out.indent();
    	
    	
    	out.printf("List <String> methodAccessClasses = new List<String>();\n");
    	
    	for (Module Pmodule: project.modules())
    	{
    		out.printf("methodAccessClasses.Add(\"%s\");\n", nameManager.getUniqueMachineName(Pmodule.machine().name()));
    	}
    	
    	out.printf("methodAccessClasses.Add(\"%s\");\n", nameManager.generateMainMachineName());
    	
    	out.printf("if (methodAccessClasses.Contains(new StackFrame(2).GetMethod().DeclaringType.Name))\n"); 
    	out.printf("{\n");
    	out.indent();
    	
    	  out.printf("System.Console.WriteLine(\"Method \" + methodName + \" cannot be called from class \" + new StackFrame(2).GetMethod().DeclaringType.Name);\n"); 
    	
    	out.unindent();
    	out.printf("}\n");
    	
    	
    	out.unindent();
    	out.printf("}\n");
    	
    }
    
    private void generateAccessRightComments()
    {
    	
  	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
	  
  	  out.printf("/*Instances of class %s can be created by classes :\n", nameManager.getUniqueMachineName());
  	  out.indent();
	  for (Module foreignModule: this.project.modules())
	  {
		  for (Import foreignImport: foreignModule.implementation().imports())
		  {
			  
			  if (foreignImport.machine().name().equals(this.module.machine().name()))
			  {
				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
				  { 
					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
					  out.printf("%s\n", nameManager.getUniqueMachineName(foreignModule.machine().name()));
				  }
			  }  
		  }
	  }
	  
		if (project.rootModules().size() > 0) {
		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
		  {
			out.printf("%s\n", nameManager.generateMainMachineName());  
		  }
		}
    	
    	out.unindent();
    	out.printf("*/\n");
    	
    	out.printf("/*Methods of class %s can be invoked by classes :\n", nameManager.getUniqueMachineName());
    	out.indent();
    	
    	for (Module Pmodule: project.modules())
    	{
    		out.printf("%s\n", nameManager.getUniqueMachineName(Pmodule.machine().name()));
    	}
    	
    	out.printf("methodAccessClasses.add(\"%s\");\n", nameManager.generateMainMachineName());
    	out.unindent();
    	out.printf("*/\n");
    	
    	
    }
	
	private void generate() {
		
		this.methods = new ArrayList <MethodSignature>();
		this.atributes = new ArrayList <AtributeSignature>();
		this.constructors = new ArrayList <MethodSignature>();
	
		System.out.println("Generate code for: "+implementation+"\n");
		
		List<Refinement> refinements = new ArrayList<Refinement>();
		AbstractComponent refinedComponent = implementation.refinedComponent();
		while (refinedComponent instanceof Refinement) {
			refinements.add((Refinement) refinedComponent);
			refinedComponent = ((Refinement)refinedComponent).refinedComponent();
		}
		Machine machine = (Machine) refinedComponent;
		
		String generics = "";
		
		if (machine.setParameters().size() > 0) {
			ListFormatter typesList = new ListFormatter();
			for (SetParameter set : machine.setParameters()) {
				typesList.addItem(set.name());
				typesAssociationTable.put(((PowerSet)set.type()).baseType(), set.name());
			}
			generics = "<"+typesList.text()+">";
		}
		
		out.printf("using System;\n");
		out.printf("using System.Collections.Generic;\n");
		
		//
		
		if(
				MPR.getClassModificationPolicy(machine.name()).getAtributePolicy().index != 0 ||
				MPR.getClassModificationPolicy(machine.name()).getMethodPolicy().index != 0 ||
				MPR.getClassModificationPolicy(machine.name()).getConstructorPolicy().index != 0
				)
				{
					out.printf("using System.Reflection;\n");

					
				}
				


					if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
					{
						out.printf("using System.Diagnostics;\n");
					}
					
				
		
		//

		// OPEN NAMESPACE BRACKET
		if (namespaceName != null)
		{
			out.printf("namespace %s\n{", namespaceName);
			out.indent();
		}
		
		if (APR.applyAllForbyddenExceptAllowedByComment(machine.name()))
		{
		 generateAccessRightComments();
		}
		
		if (APR.applyAllForbyddenExceptAllowedByReport(machine.name()))
		{
			generateAccessRightReport(); 
		}
		
		out.printf("public class %s%s {\n", (csKeywords.contains (nameManager.getUniqueMachineName()) ? ("@" + nameManager.getUniqueMachineName()) : nameManager.getUniqueMachineName()), generics);
		out.indent();
		
		// SETS
		for (Component component : module.components()) {
			generateSets(component);
		}
		for (Set set : machine.includedSets()) {
			find(set);
		}
		
		// CONSTANTS
		for (Component component : module.components()) {
			generateConstants(component);
		}
		
		// VARIABLES
		for (Component component : module.components()) {
			generateVariables(component);
		}
		
		// EXTENDED MACHINES
		if (implementation.extendedMachines().size() > 0) {
			out.println("//Extended machines");
			generateCompositionChains(implementation.extendedMachines());
		}
		
		// IMPORTED MACHINES
		if (implementation.imports().size() > 0) {
			out.println("//Imported machines");
			generateCompositionChains(implementation.imports());
		}
		
		// SEEN MACHINES
		if (implementation.seenMachines().size() > 0) {
			out.println("//Seen machines");
		}
		for (SeenMachine seenMachine : implementation.seenMachines()) {
			if (seenMachine.isImported()) {
				out.printf("private %s %s;\n", (csKeywords.contains (nameManager.getUniqueMachineName(seenMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(seenMachine.machine().name())) : nameManager.getUniqueMachineName(seenMachine.machine().name())), (csKeywords.contains (nameManager.componentReferenceName(module, seenMachine)) ? ("@" + nameManager.componentReferenceName(module, seenMachine)) : nameManager.componentReferenceName(module, seenMachine)));
			    this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(seenMachine.machine().name()), nameManager.componentReferenceName(module, seenMachine)));
				// TODO BLINDSIGHT out.printf("private %s %s;\n", seenMachine.machine().name(), nameManager.componentReferenceName(module, seenMachine));
			}
		}
		
		// scalar formal parameters
		if (machine.scalarParameters().size() > 0) {
			out.println();
			out.println("// scalar formal parameters");
		}
		for (ScalarParameter scalar : machine.scalarParameters()) {
			out.printf("private %s %s;\n", csType(scalar.type()), (csKeywords.contains (nameManager.moduleData(module, scalar)) ? ("@" + nameManager.moduleData(module, scalar)) : nameManager.moduleData(module, scalar)));
			this.atributes.add(new AtributeSignature (2, csType(scalar.type()), nameManager.moduleData(module, scalar)));
		}
		
		
		
		
		
		// CONSTRUCTOR
		 MethodSignature generatedConstructor = new MethodSignature();
		ListFormatter argsList = new ListFormatter();
		for (ScalarParameter scalar : machine.scalarParameters()){
			argsList.addItem(csType(scalar.type()) +" "+ (csKeywords.contains (nameManager.moduleData(module, scalar)) ? ("@" + nameManager.moduleData(module, scalar)) : nameManager.moduleData(module, scalar)));
		    generatedConstructor.parameterTypes.add(csType(scalar.type()));
		}
		// collect seen references without duplications
		List<SeenMachine> seenMachinesReferences = new ArrayList<SeenMachine>();
		for (SeenMachine seenMachine : module.seesReferencesDelegations()) {
			if (!seenMachinesReferences.contains(seenMachine))
				seenMachinesReferences.add(seenMachine);
		}
		for (SeenMachine seenMachine : implementation.seenMachines()) {
			if (seenMachine.isImported() && !seenMachinesReferences.contains(seenMachine))
				seenMachinesReferences.add(seenMachine);
		}
		
		for (SeenMachine seenMachine : seenMachinesReferences) {
			argsList.addItem((csKeywords.contains (nameManager.getUniqueMachineName(seenMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(seenMachine.machine().name())) : nameManager.getUniqueMachineName(seenMachine.machine().name()))+" "+(csKeywords.contains (nameManager.componentReferenceName(module, seenMachine)) ? ("@" + nameManager.componentReferenceName(module, seenMachine)) : nameManager.componentReferenceName(module, seenMachine)));
			generatedConstructor.parameterTypes.add(nameManager.getUniqueMachineName(seenMachine.machine().name()));
			// TODO BLINDSIGHT argsList.addItem(seenMachine.machine().name()+" "+nameManager.componentReferenceName(module, seenMachine));
		}
		
		out.println();
		out.printf("public %s(%s) {\n", (csKeywords.contains (nameManager.getUniqueMachineName(machine.name())) ? ("@" + nameManager.getUniqueMachineName(machine.name())) : nameManager.getUniqueMachineName(machine.name())), argsList.text());
		generatedConstructor.modifiers = 0;
		this.constructors.add(generatedConstructor);
		
		
		out.indent();
		nameManager.openScope();
		
		if (MPR.ApplyModificationPolicy(machine.name()))
		{
			out.printf("checkClassStructure();\n");
		}
		
		if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
		{
			out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkConstructorAccessRight"));
		}
		
		for (ScalarParameter scalar : machine.scalarParameters()) {
			out.printf("this.%s = %s;\n", (csKeywords.contains (nameManager.moduleData(module, scalar)) ? ("@" + nameManager.moduleData(module, scalar)) : nameManager.moduleData(module, scalar)), (csKeywords.contains (nameManager.moduleData(module, scalar)) ? ("@" + nameManager.moduleData(module, scalar)) : nameManager.moduleData(module, scalar)));
			
			
			//out.printf("this.%s = %s;\n", scalar.name(), scalar.name());
		}
		
		// Create imported machines instances
		
		//out.println("//unsorted: "+implementation().imports());
		//out.println("//sorted: "+module.sortedImports());
		
		List<Import> allCompositions = new ArrayList<Import>(implementation.imports());
		allCompositions.addAll(implementation.extendedMachines());
		
		allCompositions = module.sortedImports(allCompositions);
		for (Import importedMachine : allCompositions) {
			generateMachinesInstantiation(importedMachine);
		}
		
		// Assign seen machines references
		for (SeenMachine seenMchine : implementation.seenMachines()) {
			if (seenMchine.isImported()) {
				String seenMachineReference = (csKeywords.contains (nameManager.componentReferenceName(module, seenMchine)) ? ("@" + nameManager.componentReferenceName(module, seenMchine)) : nameManager.componentReferenceName(module, seenMchine));
				out.printf("this.%s = %s;\n", seenMachineReference, seenMachineReference);
			}
		}
		
		if (implementation.initialization() != null)
			generateInstruction(implementation.initialization());
		
		
		nameManager.closeScope();
		out.unindent();
		
		out.println("}\n"); // constructor
		
		

		// DO TESTING AND EDITING
		// OPERATIONS
		List<OperationB0> operations = implementation.getOperations();
		for (OperationB0 operation : operations) { 
			generateMethodSignature((csKeywords.contains (nameManager.get_resolved_operation_name(implementation, operation.name())) ? ("@" + nameManager.get_resolved_operation_name(implementation, operation.name())) : nameManager.get_resolved_operation_name(implementation, operation.name())), operation.inputParameters(), operation.outputParameters());
          	//		generateMethodSignature(operation.name(), operation.inputParameters(), operation.outputParameters());
			out.print(" {\n");
			out.indent();
			nameManager.openScope();
			
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			for (OperationOutputParameter outParameter : operation.outputParameters()) {
				out.printf("%s %s;\n", csType(outParameter.type()), (csKeywords.contains (nameManager.moduleData(outParameter)) ? ("@" + nameManager.moduleData(outParameter)) : nameManager.moduleData(outParameter)));
			}
			generateInstruction(operation.body());
			if (operation.outputParameters().size() == 1) {
				out.printf("return %s;\n", (csKeywords.contains (nameManager.moduleData(operation.outputParameters().get(0))) ? ("@" + nameManager.moduleData(operation.outputParameters().get(0))) : nameManager.moduleData(operation.outputParameters().get(0))));
				//out.printf("return %s;\n", operation.outputParameters().get(0).name());
			}
			if (operation.outputParameters().size() > 1) {
				String retVarName = (csKeywords.contains (nameManager.generateUniqueName("ret")) ? ("@" + nameManager.generateUniqueName("ret")) : nameManager.generateUniqueName("ret"));// TODO register retVarName in nameManager ? is it needed ?
				//out.printf("java.util.ArrayList<Object> %s = new java.util.ArrayList<Object>(%d);\n", retVarName, operation.outputParameters().size());
				//for (OperationOutputParameter outParam : operation.outputParameters()) {
				//	out.printf("%s.add(%s);\n", retVarName, nameManager.moduleData(outParam));
				//}
				//out.printf("return %s;\n", retVarName);
				
				ListFormatter outParams = new ListFormatter();
				for (OperationOutputParameter outParam : operation.outputParameters()) {
					outParams.addItem((csKeywords.contains (nameManager.moduleData(outParam)) ? ("@" + nameManager.moduleData(outParam)) : nameManager.moduleData(outParam)));
				}
				out.printf("return new object[]{%s};\n", outParams.text());
			}
			nameManager.closeScope();
			out.unindent();
			out.println("}\n");
		}
		
		
		List<OperationDeclaration> promotedOperations = new ArrayList<OperationDeclaration>(implementation.promotedOperations());
		for (Import i : implementation.extendedMachines()) {
			promotedOperations.addAll(i.operationsDeclarations());
		}
		
		
		// NEXT
		
		for (OperationDeclaration promotedOperation : promotedOperations) {
			generateMethodSignature( (csKeywords.contains (nameManager.get_resolved_operation_name(implementation, promotedOperation.operationName())) ? ("@" + nameManager.get_resolved_operation_name(implementation, promotedOperation.operationName())) : nameManager.get_resolved_operation_name(implementation, promotedOperation.operationName())), promotedOperation.inputParameters(), promotedOperation.outputParameters());
			//  generateMethodSignature(promotedOperation.operationName(), promotedOperation.inputParameters(), promotedOperation.outputParameters());
			out.print(" {\n");
			out.indent();
			
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			
			assert promotedOperation.source() != null;
			
			String delegateObject = (csKeywords.contains (nameManager.componentReferenceName(promotedOperation.source())) ? ("@" + nameManager.componentReferenceName(promotedOperation.source())) : nameManager.componentReferenceName(promotedOperation.source()));
			ListFormatter inputParams = new ListFormatter();
			for (OperationInputParameter inputParam : promotedOperation.inputParameters())
				inputParams.addItem((csKeywords.contains (nameManager.moduleData(inputParam)) ? ("@" + nameManager.moduleData(inputParam)) : nameManager.moduleData(inputParam)));
			
			//inputParams.addItem(inputParam.name());			
			String format = promotedOperation.outputParameters().size() > 0? "return %s.%s(%s);\n" : "%s.%s(%s);\n";
			out.printf(format, delegateObject, (csKeywords.contains (nameManager.get_resolved_operation_name(promotedOperation.source().machine(), promotedOperation.operationName())) ? ("@" + nameManager.get_resolved_operation_name(promotedOperation.source().machine(), promotedOperation.operationName())) : nameManager.get_resolved_operation_name(promotedOperation.source().machine(), promotedOperation.operationName())), inputParams.text());
			//out.printf(format, delegateObject, promotedOperation.operationName(), inputParams.text());
			out.unindent();
			out.println("}");
		}
		
		// generate methods to return imported machines objects
		for (Import importedMachine : implementation.imports()) { // TODO check if other elements uses imported machine access operation
			out.printf("public %s %s() {\n", (csKeywords.contains (nameManager.getUniqueMachineName(importedMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(importedMachine.machine().name())) : nameManager.getUniqueMachineName(importedMachine.machine().name())), (csKeywords.contains (nameManager.generate_unique_method_name(implementation, (csKeywords.contains (nameManager.componentReferenceName(importedMachine)) ? ("@" + nameManager.componentReferenceName(importedMachine)) : nameManager.componentReferenceName(importedMachine)))) ? ("@" + nameManager.generate_unique_method_name(implementation, (csKeywords.contains (nameManager.componentReferenceName(importedMachine)) ? ("@" + nameManager.componentReferenceName(importedMachine)) : nameManager.componentReferenceName(importedMachine)))) : nameManager.generate_unique_method_name(implementation, (csKeywords.contains (nameManager.componentReferenceName(importedMachine)) ? ("@" + nameManager.componentReferenceName(importedMachine)) : nameManager.componentReferenceName(importedMachine)))));
//			out.printf("public %s %s() {\n", nameManager.getUniqueMachineName(importedMachine.machine().name()), nameManager.componentReferenceName(importedMachine));

			MethodSignature MS = new MethodSignature();
			MS.name = nameManager.generate_unique_method_name(implementation, nameManager.componentReferenceName(importedMachine));
			MS.modifiers = 0;
			MS.returnType = nameManager.getUniqueMachineName(importedMachine.machine().name());
			this.methods.add(MS);
			
			out.indent();
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			out.printf("return %s;\n", (csKeywords.contains (nameManager.componentReferenceName(importedMachine)) ? ("@" + nameManager.componentReferenceName(importedMachine)) : nameManager.componentReferenceName(importedMachine)));
			out.unindent();
			out.println("}\n");
		}
		
		generateModificationPolicyEnforcementMethod(MPR.getClassModificationPolicy(machine.name()).getAtributePolicy().index, MPR.getClassModificationPolicy(machine.name()).getMethodPolicy().index, MPR.getClassModificationPolicy(machine.name()).getConstructorPolicy().index);
		
		if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
		{
		  generateConstructorAccessCheck();
		  generateMethodAccessRight();
		}
		
		out.unindent();
		out.println("}"); // class
		
		//recordGenerator.generateRecordsClasses(this);
		
		out.flush();
		//System.out.println(buffer);
		

		// CLOSE NAMESPACE BRACKET
		if (namespaceName != null)
		{
			out.unindent();
			out.printf("\n}");
		}
		
		
		
		
	}
	
	
	Object[] ret() {
		return new Object[]{"g", 4};
	}
	
	private void generateCompositionChains(List<Import> compositions) {
		for (Import importedMachine : compositions) {
			String name = nameManager.componentReferenceName(module, importedMachine);
			if (importedMachine.machine().setParameters().size() > 0) {
				ListFormatter typesList = new ListFormatter();
				for (SetParameter setParam : importedMachine.machine().setParameters()) {
					
					Type paramType = importedMachine.formalParametersValues().get(setParam.index()).type();
					if (paramType instanceof PowerSet) {
						typesList.addItem(csType(((PowerSet) paramType).baseType()));
					} else {
						throw new CsCodeGenerationException("Set Parameter has wrong type: "+paramType, module);
					}
					
				}
				out.printf("private %s<%s> %s;\n", (csKeywords.contains(nameManager.getUniqueMachineName(importedMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(importedMachine.machine().name())) : nameManager.getUniqueMachineName(importedMachine.machine().name())), typesList.text(), (csKeywords.contains (name) ? ("@" + name) : name));
				this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(importedMachine.machine().name()), name));
			} else
			{
				out.printf("private %s %s;\n", (csKeywords.contains (nameManager.getUniqueMachineName(importedMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(importedMachine.machine().name())) : nameManager.getUniqueMachineName(importedMachine.machine().name())), (csKeywords.contains (name) ? ("@" + name) : name));
			    this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(importedMachine.machine().name()), name));
		
			}
		}
	}
	

	private void generateMachinesInstantiation(Import importedMachine) { //TODO deeper analyse
		
		String name = nameManager.componentReferenceName(module, importedMachine);
		
		ListFormatter values = new ListFormatter();
		for (ScalarParameter scalarParam : importedMachine.machine().scalarParameters()) {
			System.out.println(importedMachine);
			System.out.println(importedMachine.machine().scalarParameters());
			values.addItem(generateDataValue(importedMachine.formalParametersValues().get(scalarParam.index())));
		}
		// Seen machines references
		// as refinements must preserve previous sees, to get all seen machines
		// is used last refinement/implementation
		List<SeenMachine> seenReferences = new ArrayList<SeenMachine>();
		for (SeenMachine seen : project.moduleOfMachine(importedMachine.machine()).seesReferencesDelegations()) {
			if (!seenReferences.contains(seen))
				seenReferences.add(seen);
		}
		for (SeenMachine seen : project.moduleOfMachine(importedMachine.machine()).lastComponent().seenMachines()) {
			if (!seenReferences.contains(seen))
				seenReferences.add(seen);
		}
		for (SeenMachine seenMachine : seenReferences) {
			if (module.seesReferencesDelegations().contains(seenMachine)) {
				values.addItem(nameManager.componentReferenceName(module, seenMachine));
			} else {
				List<Import> importPath = module.getImportPath(seenMachine);
				String reference = "this."+nameManager.componentReferenceName(module, importPath.get(0));
				for (int i = 1 ; i < importPath.size(); i++) {
					reference += "."+nameManager.componentReferenceName(project.moduleOfMachine(importPath.get(1).machine()), importPath.get(1))+"()";
				}
				//values.addItem(module.getImportPath(seenMachine).toString());
				values.addItem(reference);
			}
		}
		
		if (importedMachine.machine().setParameters().size() > 0) {
			ListFormatter types = new ListFormatter();
			for (SetParameter setParam : importedMachine.machine().setParameters()) {
				Type paramType = importedMachine.formalParametersValues().get(setParam.index()).type();
				if (paramType instanceof PowerSet) {
					types.addItem(csType(((PowerSet) paramType).baseType()));
				} else {
					throw new CsCodeGenerationException("Set Parameter has wrong type: "+paramType, module);
				}
			}
			out.printf("this.%s = new %s<%s>(%s);\n", (csKeywords.contains (name) ? ("@" + name) : name), (csKeywords.contains (nameManager.getUniqueMachineName(importedMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(importedMachine.machine().name())) : nameManager.getUniqueMachineName(importedMachine.machine().name())), types.text(), values.text(), (csKeywords.contains (name) ? ("@" + name) : name));
		} else {
			out.printf("this.%s = new %s(%s);\n", (csKeywords.contains (name) ? ("@" + name) : name), (csKeywords.contains (nameManager.getUniqueMachineName(importedMachine.machine().name())) ? ("@" + nameManager.getUniqueMachineName(importedMachine.machine().name())) : nameManager.getUniqueMachineName(importedMachine.machine().name())), values.text(), (csKeywords.contains (name) ? ("@" + name) : name));
		}
		
	}
	
	
	private void generateMethodSignature(String operationName, List<OperationInputParameter> inputParameters, List<OperationOutputParameter> outputParameters) {
		//format input parameters to Java
		
		//AddedbyMe
		MethodSignature MS = new MethodSignature();
		MS.name = operationName;
		
		ListFormatter inputParams = new ListFormatter();
		for (OperationInputParameter inputParam : inputParameters) {
			inputParams.addItem(csType(inputParam.type()) +" "+ (csKeywords.contains (nameManager.moduleData(module, inputParam)) ? ("@" + nameManager.moduleData(module, inputParam)) : nameManager.moduleData(module, inputParam)));
			MS.parameterTypes.add(csType(inputParam.type()));
		}//inputParams.addItem(javaType(inputParam.type()) +" "+ inputParam.name());
		
		
		
		String outParamType;
		switch (outputParameters.size()) {
			case 0: outParamType = "void";break;
			case 1: outParamType = csType(outputParameters.get(0).type());break;
			default: outParamType = "object[]";
		}
		
		out.printf("public %s %s(%s)", outParamType, operationName, inputParams.text());		
		//out.printf("public %s %s(%s)", outParamType, operationName, inputParams.text());
		
		//AddedByMe
        MS.modifiers = 0;
        MS.returnType = (outParamType.equals("void") ? "Void" : outParamType);
        this.methods.add(MS);
	}	
	
	private void generateSets(Component component) {
		
		/* PRIMARI SET COMPILE IDEA
		if (component.enumeratedSets().size() > 0) {
			out.println();
			out.println("// sets of "+component);
			for (EnumeratedSet set : component.enumeratedSets()) {
				String csName = (csKeywords.contains(nameManager.moduleData(module, set)) ? ("@" + nameManager.moduleData(module, set)) : nameManager.moduleData(module, set));
				out.printf("public enum %s {\n", csName);
				out.indent();
				int index = 0;
				for (Item item : set.items()) {
					out.printf("%s = %d,\n", (csKeywords.contains (nameManager.moduleData(module, item)) ? ("@" + nameManager.moduleData(module, item)) : nameManager.moduleData(module, item)), index++);
				}
				out.unindent();
				out.println("}");
			}
		} 
		*/
		
		if (component.enumeratedSets().size() > 0) {
			out.println();
			out.println("// sets of "+component);
			for (EnumeratedSet set : component.enumeratedSets()) {
				String csName = (csKeywords.contains(nameManager.moduleData(module, set)) ? ("@" + nameManager.moduleData(module, set)) : nameManager.moduleData(module, set));
				out.printf("public enum %s {\n", csName);
				out.indent();
	
				for (Item item : set.items()) {
					out.printf("%s,\n", (csKeywords.contains (nameManager.moduleData(module, item)) ? ("@" + nameManager.moduleData(module, item)) : nameManager.moduleData(module, item)));
				}
				out.unindent();
				out.println("}");
			}
		} 
	}
	
	
	private void generateConstants(Component component) {
		if (component.concreteConstants().size() < 1)
			return;
		out.println();
		out.println("// constants of "+component);
		generateConstants: for (ConcreteConstant constant : component.concreteConstants()) {
			//out.println(implementation.valueOf(constant));
			//out.println(implementation.valuations2.get(constant.name()));
			
			B0DataValue value = implementation.valuations2.get(constant.name());
			// check all imported machines (transitively) for such data
			for (Import imp : implementation.imports()) {
				for (ConcreteConstant con : imp.machine().concreteConstants()) {
					if (con.name().equals(constant.name())) {
						out.printf("//included constant %s is in imported machine %s", constant, imp.machine().name());
						continue generateConstants;
					}
				}
			}
			// null value should have imported
			//if (value != null) {
				//out.printf("public static %s %s = %s;\n", javaType(constant.type()), constant.name(), generateDataValue(value));
				System.out.println(constant);
				System.out.println(value);
				String assignment = generateAssigment(constant, value, new StaticCodeGenerator());
				out.printf("public static %s %s;\n", csType(constant.type()), assignment);
				this.atributes.add(new AtributeSignature(1, csType(constant.type()), nameManager.moduleData(constant)));
			//}
		}
	}
	
	
	private void generateVariables(Component component) {
		
		if (component.concreteVariables().size() < 1)
			return;
		out.println(); 
		out.println("// variables of "+component);
		for (ConcreteVariable variable : component.concreteVariables()) {
		
			boolean generateSimpleVariable = true;
			
			if (nameManager.moduleData(variable).contains("."))
			{
				generateSimpleVariable = false;
			}
			
			for (IncludedMachine included : this.implementation.abstractMachine().includedMachines())
			{
				for (ConcreteVariable concreteVariable: included.includedConcreteVariables())
				{
					if (concreteVariable.name().equals(variable.name()))
					{
						generateSimpleVariable = false;
					}	
				}
			}
				
				// TODO DO FOR REFINEMENTS, BUT FIRST MODIFI PARSER ITSELF

			
			
			
			
			
			if (generateSimpleVariable)
			{
			out.printf("private %s %s;\n", csType(variable.type()), nameManager.moduleData(variable));
			this.atributes.add(new AtributeSignature (2,csType(variable.type()), nameManager.moduleData(variable)));
			}
			
		
		
		
		
		
		}
		
	}	
	
	private void generateInstruction(Instruction instruction) {
		if (instruction instanceof SequenceInstruction)
			generateInstruction((SequenceInstruction) instruction);
		else if (instruction instanceof BlockInstruction)
			generateInstruction((BlockInstruction) instruction);
		else if (instruction instanceof BecomesEqualInstruction)
			generateInstruction((BecomesEqualInstruction) instruction);
		
		else if (instruction instanceof IfConditionalInstruction)
			generateInstruction((IfConditionalInstruction) instruction);
		else if (instruction instanceof LocalVariableInstruction)
			generateInstruction((LocalVariableInstruction) instruction);
		else if (instruction instanceof CallupInstruction)
			generateInstruction((CallupInstruction) instruction);
		else if (instruction instanceof WhileInstruction)
			generateInstruction((WhileInstruction) instruction);
		else if (instruction instanceof CaseConditionalInstruction)
			generateInstruction((CaseConditionalInstruction) instruction);
		
		else if (instruction instanceof IdentityInstruction)
			;
		else if (instruction instanceof AssertionInstruction)
			;
		
		else throw new RuntimeException("Unsupported instruction type: "+instruction.getClass());
	}
	
	private String generateLabel(B0DataValue value) {
		if (value instanceof B0VariableReference) {
			DataField var = ((B0VariableReference) value).variable();
			return generateLabel(var);
		}
		return generateDataValue(value);
	}
	
	private String generateLabel(DataField data) {
		if (data instanceof ForeignVariable) {
			return generateLabel(((ForeignVariable) data).data()); // TODO ANALISE KEYWORD EXPLOSION
		} else if (data instanceof EnumeratedSet.Item) {
			EnumeratedSet.Item item = (EnumeratedSet.Item) data;
			
			//return generateDataValue(data.type(), data, new MethodCodeGenerator() )
			return (csType(item.type()) + "." + (csKeywords.contains (nameManager.moduleData(module, item)) ? ("@" + nameManager.moduleData(module, item)) : nameManager.moduleData(module, item)));
			//return (csKeywords.contains (nameManager.moduleData(module, item)) ? ("@" + nameManager.moduleData(module, item)) : nameManager.moduleData(module, item));
		}
		throw new CsCodeGenerationException("wrong case label", module);
	}
	
	// TODO CHECK IF DEFAULT HAS BREAK CONDITION
	private void generateInstruction(CaseConditionalInstruction instruction) {
		boolean convertToBool = instruction.expression().type().isEqual(BOOL);
		out.printf(convertToBool ? "switch (%s ? 1 : 0) {\n"
								 : "switch (%s) {\n", generateDataValue(instruction.expression()));
		
		out.indent();
		for (sk.tuke.bkpi.model.instruction.CaseConditionalInstruction.Branch branch : instruction.branches()) {
			for (B0DataValue literal : branch.literalConstants()) {
				out.printf(convertToBool? "case %s ? 1 : 0:\n"
										: "case %s:\n", generateLabel(literal));
			}
			out.indent();
			generateInstruction(branch.substitution());
			out.println("break;");
			out.unindent();
		}
		if (instruction.elseSubstitution() != null) {
			out.println("default:");
			out.indent();
			generateInstruction(instruction.elseSubstitution());
			out.println("break;");
			out.unindent();
		}
		out.unindent();
		out.println("}");
	}
	
	
	private void generateInstruction(LocalVariableInstruction instruction) {
		out.println("{");
		out.indent();
		nameManager.openScope();
		for (LocalVariable variable : instruction.localVariables()) {
			// declaration of local variables
			if (!variable.isTyped())
				throw new CsCodeGenerationException("variable not typed: "+variable, module);
			out.printf("%s %s;\n", csType(variable.type()), (csKeywords.contains (nameManager.moduleData(variable)) ? ("@" + nameManager.moduleData(variable)) : nameManager.moduleData(variable)));
		}
		// substitution code
		generateInstruction(instruction.substitution());
		nameManager.closeScope();
		out.unindent();
		out.println("}");
	}
	
	private void generateInstruction(BlockInstruction block) {
		//out.println("{");
		generateInstruction(block.body());
		//out.println("}");
	}
	
	private void generateInstruction(SequenceInstruction sequence) {
		for (Instruction instruction : sequence.instructions()) {
			generateInstruction(instruction);
		}
	}
	
	private void generateInstruction(IfConditionalInstruction instruction) {
		out.printf("if (%s) {\n", generateCondition(instruction.ifBranch().condition()));
		out.indent();
		generateInstruction(instruction.ifBranch().substitution());
		out.unindent();
		out.println("}");
		for (Branch elseif : instruction.elseifBranches()) {
			out.printf("else if (%s) {\n", generateCondition(elseif.condition()));
			out.indent();
			generateInstruction(elseif.substitution());
			out.unindent();
			out.println("}");
		}
		if (instruction.elseBranch() != null) {
			out.println("else {");
			out.indent();
			generateInstruction(instruction.elseBranch());
			out.unindent();
			out.println("}");
		}
	}
	
	private void generateInstruction(BecomesEqualInstruction instruction) {
		out.printf("%s;\n", generateAssigment(instruction.variable(), instruction.value(), new MethodCodeGenerator()));
		if (true) return;
		if (instruction.value() instanceof ArrayExpression) {
			out.println(arrayGenerator.arrayInitialization(instruction.variable(), (ArrayExpression) instruction.value(), null));
		} else {
			
			if (instruction.variable().type() instanceof Record) {
				out.println("//check record type: "+instruction.variable().type()+" data: "+instruction.value().type());
			}
			//out.printf("%s = %s;\n", generateVariable(instruction.variable()), generateDataValue(instruction.value()));
			out.printf("%s;\n", generateAssigment(instruction.variable(), instruction.value(), new MethodCodeGenerator()));
		}
	}
	
	
	private void generateInstruction(CallupInstruction instruction) {
		ListFormatter argumentsList = new ListFormatter();
		for (B0DataValue argument : instruction.arguments()) {
			argumentsList.addItem(generateDataValue(argument));
		}
		
		String instanceName = instruction.declaration().source() == null? "this" : (csKeywords.contains (nameManager.componentReferenceName(module, instruction.declaration().source())) ? ("@" + nameManager.componentReferenceName(module, instruction.declaration().source())) : nameManager.componentReferenceName(module, instruction.declaration().source()));
		String call = format("%s.%s(%s)", instanceName, (csKeywords.contains (nameManager.get_resolved_operation_name(instruction.declaration().source().machine(), instruction.declaration().operationName())) ? ("@" + nameManager.get_resolved_operation_name(instruction.declaration().source().machine(), instruction.declaration().operationName())) : nameManager.get_resolved_operation_name(instruction.declaration().source().machine(), instruction.declaration().operationName())), argumentsList.text());
		//  String call = format("%s.%s(%s)", instanceName, instruction.declaration().operationName(), argumentsList.text());
		if (instruction.assignVariables().size() == 0) {
			out.printf("%s;\n", call);
		} else if (instruction.assignVariables().size() == 1) {
			out.printf("%s = %s;\n", (csKeywords.contains (nameManager.moduleData(instruction.assignVariables().get(0))) ? ("@" + nameManager.moduleData(instruction.assignVariables().get(0))) : nameManager.moduleData(instruction.assignVariables().get(0))), call);
			//out.printf("%s = %s;\n", instruction.assignVariables().get(0).name(), call); 
		} else {
			LocalVariable result = new LocalVariable((csKeywords.contains (nameManager.generateUniqueName("out")) ? ("@" + nameManager.generateUniqueName("out")) : nameManager.generateUniqueName("out")));
			// Note: cannot directly use result.name(), cause it will not register
			// local variable and names collisions can occur later in code
			out.printf("Object[] %s = %s;\n", (csKeywords.contains (nameManager.moduleData(result)) ? ("@" + nameManager.moduleData(result)) : nameManager.moduleData(result)), call);
			int i = 0;
			for (DataField variable : instruction.assignVariables()) {
				out.printf("%s = (%s) %s[%d];\n", (csKeywords.contains (nameManager.moduleData(variable)) ? ("@" + nameManager.moduleData(variable)) : nameManager.moduleData(variable)), csType(variable.type()),
						(csKeywords.contains (nameManager.moduleData(result)) ? ("@" + nameManager.moduleData(result)) : nameManager.moduleData(result)), i++);
			}
		}
	}
	
	
	private void generateInstruction(WhileInstruction instruction) {
		out.printf("while(%s) {\n", generateCondition(instruction.condition()));
		out.indent();
		generateInstruction(instruction.body());
		out.unindent();
		out.println("}");
	}
	
	
	String generateDataField(DataField data, Module module) {
		//System.out.println("generate data field "+data);
		String prefix = "";
		String name = null;
		if (data instanceof ForeignVariable) { //&& ((ForeignData) data).data() instanceof ConcreteConstant or set
			Module m = project.moduleOfMachine(((ForeignVariable) data).componentReference().machine());
			DataField d = ((ForeignVariable) data).data();
			
			prefix = (csKeywords.contains (nameManager.componentReferenceName(((ForeignVariable) data).componentReference())) ? ("@" + nameManager.componentReferenceName(((ForeignVariable) data).componentReference())) : nameManager.componentReferenceName(((ForeignVariable) data).componentReference()))+".";
			name = (csKeywords.contains (nameManager.globalModuleData(m, data)) ? ("@" + nameManager.globalModuleData(m, data)) : nameManager.globalModuleData(m, data));
			module = m;
			data = d;
		}
		if (data instanceof ForeignData) {
			module = project.moduleOfMachine(((ForeignData) data).machine());
			DataField d = ((ForeignData) data).data();
			prefix = (csKeywords.contains (nameManager.getUniqueMachineName(((ForeignData) data).machine().name())) ? ("@" + nameManager.getUniqueMachineName(((ForeignData) data).machine().name())) : nameManager.getUniqueMachineName(((ForeignData) data).machine().name()))+"."; 
			name = (csKeywords.contains (nameManager.globalModuleData(module, data)) ? ("@" + nameManager.globalModuleData(module, data)) : nameManager.globalModuleData(module, data));
			data = d;
		}
		if (data instanceof Item) {
			Item item = (Item) data;
			prefix += (csKeywords.contains (nameManager.globalModuleData(module, item.type().set())) ? ("@" + nameManager.globalModuleData(module, item.type().set())) : nameManager.globalModuleData(module, item.type().set()))+".";
			name = (csKeywords.contains (nameManager.moduleData(module, item)) ? ("@" + nameManager.moduleData(module, item)) : nameManager.moduleData(module, item));
		}
		
		if (name == null)
			name = (csKeywords.contains (nameManager.moduleData(module, data)) ? ("@" + nameManager.moduleData(module, data)) : nameManager.moduleData(module, data));

		return prefix+name;
	}
	
	String generateDataField(DataField data) {
		return generateDataField(data, module);
	}
	
	
	String generateVariable(B0Variable variable) {
		//System.out.println("generate variable: "+variable);
		if (variable instanceof B0VariableReference) {
			DataField data = ((B0VariableReference) variable).variable();
			return generateDataField(data);
		}
		if (variable instanceof RecordFieldVariable) {
			// TODO: move to record generator
			RecordFieldVariable record = (RecordFieldVariable) variable;
			return format("%s.%s", generateVariable(record.recordVariable()), (is_keyword (nameManager.record_data(csType (record.recordVariable().type()), record.field())) ? ("@" + nameManager.record_data(csType (record.recordVariable().type()), record.field())) : nameManager.record_data(csType (record.recordVariable().type()), record.field())));
			//return format("%s.%s", generateVariable(((RecordFieldVariable) variable).recordVariable()), ((RecordFieldVariable) variable).field());
		}
		if (variable instanceof ArrayFieldVariable) {
			return arrayGenerator.getElement((ArrayFieldVariable) variable);
		}
		return "Unsupported variable :"+variable + " class: "+variable.getClass();
	}
	
	String generateDataValue(B0DataValue value) {
		return generateDataValue(value.type(), value, null);
	}
	
	
	private String generateCondition(Condition condition) {
		if (condition instanceof BinaryCondition)
			return generateCondition((BinaryCondition) condition);
		else if (condition instanceof ComparisonCondition)
			return generateCondition((ComparisonCondition) condition);
		else if (condition instanceof NegationCondition)
		    return "!("+generateCondition(((NegationCondition) condition).getOriginalCondition())+")";
		return "unimplemented";
	}
	
	private String generateCondition(BinaryCondition condition) {
		String operator;
		switch (condition.getOperator()) {
			case CONJUNCTION: operator = "&&"; break;
			case DISJUNCTION: operator = "||"; break;
			default: throw new CsCodeGenerationException("Unknown Binary Condition operator: "+condition.getOperator(), module);
		}
		//return generateCondition(condition.getLeft())+operator+generateCondition(condition.getRight());
		return format("(%s %s %s)", generateCondition(condition.getLeft()), operator, generateCondition(condition.getRight()));
	}
	
	private String generateCondition(ComparisonCondition condition) {
		Operator operator = condition.operator();
		/* transformation to nicer code
		if (condition.right().getType().isEqual(BasicType.BOOL)) {
			if (operator == Operator.EQUAL)
				return generateTerm(condition.left());
			else if (operator == Operator.NOT_EQUAL)
		}
		*/
		String csOperator;
		switch (operator) {
			case EQUAL: 			csOperator = "=="; break;
			case NOT_EQUAL: 		csOperator = "!="; break;
			case LESS: 				csOperator = "<";  break;
			case LESS_OR_EQUAL: 	csOperator = "<="; break;
			case GREATER: 			csOperator = ">";  break;
			case GREATER_OR_EQUAL: 	csOperator = ">="; break;
			default: throw new CsCodeGenerationException("Unimplemented Comparison Condition operator: "+operator, module);
		}
		return format("%s %s %s", generateDataValue(condition.left()), csOperator, generateDataValue(condition.right()));
	}
	
	
	private TypeResolver INDEXES = new TypeResolver();
	private TypeResolver ELEMENT_TYPE = new TypeResolver();
	private Type ARRAY_TYPE = PowerSet(CartesianProduct(INDEXES, ELEMENT_TYPE));
	
	String csType(Type type) {
		for (Type t1 : implementation.typesSubstitutions().keySet()) {
			if (type.contains(t1)) {
				type = type.replace(t1, implementation.typesSubstitutions().get(t1));
			}
		}
		
		if (typesAssociationTable.containsKey(type))
			return typesAssociationTable.get(type);
		
		// or add Basic Types to table ?
		else if (type.isEqual(Z))
			return "Int32";
		else if (type.isEqual(BOOL))
			return "Boolean";
		else if (type.isEqual(STRING))
			return "String";
		
		else if (type instanceof SetType) {
			Set set = ((SetType) type).set();
			if (set instanceof EnumeratedSet) {
				if (includedMapping.containsKey(set)) {
					DataField v = includedMapping.get(set);
					return 	(csKeywords.contains (nameManager.getUniqueMachineName(setsMapping.get(v).machine().name())) ? ("@" + nameManager.getUniqueMachineName(setsMapping.get(v).machine().name())) : nameManager.getUniqueMachineName(setsMapping.get(v).machine().name()))+
					"." + (csKeywords.contains (nameManager.moduleData(module, v)) ? ("@" + nameManager.moduleData(module, v)) : nameManager.moduleData(module, v)); 
				}
				return (csKeywords.contains (nameManager.getUniqueMachineName(setsMapping.get(set).machine().name())) ? ("@" + nameManager.getUniqueMachineName(setsMapping.get(set).machine().name())) : nameManager.getUniqueMachineName(setsMapping.get(set).machine().name()))+
				"." + (csKeywords.contains (nameManager.moduleData(module, set)) ? ("@" + nameManager.moduleData(module, set)) : nameManager.moduleData(module, set)); 
			}
			return (csKeywords.contains (nameManager.moduleData(module, set)) ? ("@" + nameManager.moduleData(module, set)) : nameManager.moduleData(module, set)); 
		}
		
		else if (type instanceof Record)
			return recordGenerator.recordType((Record) type);
		else if (type.isSubtype(ARRAY_TYPE))
			return arrayGenerator.arrayType(type);
		
		if (type.isEqual(PowerSet(Z))) {
			return "Interval";
		}
		throw new CsCodeGenerationException("Unimplemented Type:"+type);
	}
	
	String generateAssigment(B0Variable variable, B0DataValue value, CodeGenerator codeGenerator) {
		String csVariable = generateVariable(variable);
		String csValue = generateDataValue(variable.type(), value, codeGenerator);
		return format("%s = %s", csVariable, csValue);
	}
	
	String generateAssigment(DataField data, B0DataValue value, CodeGenerator codeGenerator) {
		String csValue = generateDataValue(data.type(), value, codeGenerator);
		codeGenerator.close();
		return format("%s = %s", (csKeywords.contains (nameManager.moduleData(module, data)) ? ("@" + nameManager.moduleData(module, data)) : nameManager.moduleData(module, data)), csValue);
	}
	
	
	public String generateDataValue(Type type, B0DataValue value, CodeGenerator codeGenerator) {
		try {
			Method m = CsGenerator.class.getMethod("generateDataValue", new Class[]{Type.class, value.getClass(), CodeGenerator.class});
			return (String) m.invoke(this, type, value, codeGenerator);
		} catch (Exception e) {
			try {
				System.out.println(value);
				System.out.println(value.getClass());
				System.out.println(value.getClass().getSuperclass());
				Method m = CsGenerator.class.getMethod("generateDataValue", new Class[]{Type.class, value.getClass().getSuperclass(), CodeGenerator.class});
				return (String) m.invoke(this, type, value, codeGenerator);
			} catch (Exception ex) {
				ex.printStackTrace();
				throw new CsCodeGenerationException(ex.getMessage());
			}
		}
	}
	
	
	public String generateDataValue(Type type, BooleanTerm value, CodeGenerator codeGenerator) {
		return value.value()? "true" : "false";
	}
	
	public String generateDataValue(Type type, IntegerTerm value, CodeGenerator codeGenerator) {
		switch(value.value()) {
			case Integer.MAX_VALUE: return "Int32.MaxValue";
			case Integer.MIN_VALUE: return "Int32.MinValue";
			default: return value.value().toString();
		}
	}
	
	public String generateDataValue(Type type, NegativeTerm value, CodeGenerator codeGenerator) {
		String csValue = generateDataValue(value.expression());
		return value.expression() instanceof ArithmeticalBinaryExpression? "-("+csValue+")" : "-"+csValue;
	}
	
	public String generateDataValue(Type type, RecordTerm record, CodeGenerator codeGenerator) {
		Record recordType = (Record) type;
		ListFormatter fields = new ListFormatter();
		for (RecordTerm.RecordField field : record.fields()) {
			Type fieldType = recordType.fieldType(field.label());
			fields.addItem(generateDataValue(fieldType, field.value(), codeGenerator));
		}
		return format("new %s(%s)", recordGenerator.recordType(recordType), fields.text());
	}
	
	public String generateDataValue(Type type, ArrayExpression value, CodeGenerator codeGenerator) {
		return arrayGenerator.arrayExpression(value, codeGenerator);
	}
	
	public String generateDataValue(Type type, ArrayFieldVariable value, CodeGenerator codeGenerator) {
		return arrayGenerator.getElement(value);
	}
	
	public String generateDataValue(Type type, B0Variable value, CodeGenerator codeGenerator) {
		return generateVariable(value);
	}
	
	public String generateDataValue(Type type, BoolConversionValue value, CodeGenerator codeGenerator) {
		return generateCondition(value.predicate());
	}
	
	public String generateDataValue(Type type, StringValue string, CodeGenerator codeGenerator) {
		return "\""+string.value()+"\"";
	}
	
	public String generateDataValue(Type type, ArithmeticalBinaryExpression expr, CodeGenerator codeGenerator) {
		String operator;
		B0DataValue left = expr.left();
		B0DataValue right = expr.right();
		
		switch (expr.operator()) {
			case PLUS: operator = "+";break;
			case MINUS:
				operator = "-";
				if (right instanceof NegativeTerm) {
					//return generateDataValue(expr.left())+"+"+generateDataValue(((NegativeTerm) expr.right()).expression());
					
					right = negativeTerm(right);
					if (negativeTermsCount(expr.right()) % 2 == 1) {
						right = ((NegativeTerm) right).expression();
						operator = "+";
					}
				}
				
				break;
			case TIMES: operator = "*";break;
			case DIVIDE: operator = "/";break;
			case MODULO: operator = "%";break;
			case POWER: return format("Math.Pow(%s, %s)", expr.left(), expr.right());
			default: throw new CsCodeGenerationException("Unsuported Arithmetical Binary operator: "+expr.operator(), module);
		}
		//String format = expr.right() instanceof NegativeTerm? "" : "";
		String format;
		
		format = left instanceof ArithmeticalBinaryExpression &&
			((ArithmeticalBinaryExpression) left).operator().priority>expr.operator().priority? "(%s)" : "%s";
		format += operator;
		format += right instanceof ArithmeticalBinaryExpression &&
			(expr.operator() == ArithmeticalBinaryExpression.Operator.MINUS
				|| ((ArithmeticalBinaryExpression) right).operator().priority>expr.operator().priority)? "(%s)" : "%s";
		
		//return format("(%s)%s(%s)", generateDataValue(left), operator, generateDataValue(right)); // safe way
		return format(format, generateDataValue(left), generateDataValue(right));
		//return "("+generateDataValue(left)+operator+generateDataValue(right)+")";
	}
	
	public String generateDataValue(Type type, IntervalB0 interval, CodeGenerator codeGenerator) {
		return format("new Interval(%s, %s)", generateDataValue(interval.minimum()), generateDataValue(interval.maximum()));
		
	}
	
	public B0DataValue negativeTerm2(B0DataValue term) {
		int count = 0;
		B0DataValue x = term;
		while(term instanceof NegativeTerm) {
			count++;
			term = ((NegativeTerm) term).expression();
		}
		return count % 2 == 0? term : new NegativeTerm(term);
	}
	
	public B0DataValue negativeTerm(B0DataValue term) {
		int count = 0;
		B0DataValue x = term;
		while(term instanceof NegativeTerm) {
			count++;
			if (count % 2 == 1) {
				x = term;
			}
			term = ((NegativeTerm) term).expression();
		}
		return count % 2 == 0? term : x;
	}
	
	public int negativeTermsCount(B0DataValue term) {
		int count = 0;
		while(term instanceof NegativeTerm) {
			count++;
			term = ((NegativeTerm) term).expression();
		}
		return count;
	}
	
	public abstract class CodeGenerator {
		protected int scope;
		
		public void openScope() {
			scope++;
			out.print(" {\n");
			out.indent();
		}
		
		void closeScope() {
			if (scope > 0) {
				scope--;
				out.unindent();
				out.println("}");
			} else {
				throw new IllegalStateException();
			}
		}
		
		void close() {}
		
		abstract void addDeclaration(String declaration);
		abstract void addStatement(String statement);
	}
	
	public class MethodCodeGenerator extends CodeGenerator {

		@Override
		void addDeclaration(String declaration) {
			out.printf(declaration);
		}

		@Override
		void addStatement(String statement) {
			out.printf(statement);
		}		
	}
	
	public class StaticCodeGenerator extends CodeGenerator {
		
		boolean declarationMode = true;
		
		public void addDeclaration(String declaration) {
			if (scope == 0) {
				if (!declarationMode) {
					out.println("}");
					declarationMode = true;
				}
				out.println("private static "+declaration);
			} else {
				out.println(declaration);
			}
		}
		
		public void close() {
			if (!declarationMode) {
				out.unindent();
				out.println("}");
			}
		}
		
		void addStatement(String statement) {
			if (scope == 0) {
				if (declarationMode) {
					out.println("static {");
					out.indent();
					declarationMode = false;
				}
				out.printf(statement);
			} else {
				out.printf(statement);
			}
		}
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
}
