package sk.tuke.bkpi.generator;

import static java.lang.String.format;
import static sk.tuke.bkpi.model.types.BasicType.Z;
import static sk.tuke.bkpi.model.types.TypeBuilder.CartesianProduct;
import static sk.tuke.bkpi.model.types.TypeBuilder.PowerSet;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;

import sk.tuke.bkpi.generator.CsGenerator.CodeGenerator;
import sk.tuke.bkpi.model.b0data.ArrayFieldVariable;
import sk.tuke.bkpi.model.b0data.B0Variable;
import sk.tuke.bkpi.model.b0data.B0VariableReference;
import sk.tuke.bkpi.model.clause.EnumeratedSet;
import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.LocalVariable;
import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.expression.ExpressionReference;
import sk.tuke.bkpi.model.expression.IntegerLiteral;
import sk.tuke.bkpi.model.expression.IntervalExpression;
import sk.tuke.bkpi.model.expression.arithmetical.BinaryArithmeticalExpression;
import sk.tuke.bkpi.model.expression.arithmetical.Negation;
import sk.tuke.bkpi.model.term.ArithmeticalBinaryExpression;
import sk.tuke.bkpi.model.term.ArrayExpression;
import sk.tuke.bkpi.model.term.ArrayInitialization;
import sk.tuke.bkpi.model.term.BooleanSetB0;
import sk.tuke.bkpi.model.term.CoupleSet;
import sk.tuke.bkpi.model.term.IntegerSetB0;
import sk.tuke.bkpi.model.term.IntegerTerm;
import sk.tuke.bkpi.model.term.IntervalB0;
import sk.tuke.bkpi.model.term.ArithmeticalBinaryExpression.Operator;
import sk.tuke.bkpi.model.term.CoupleSet.Couple;
import sk.tuke.bkpi.model.types.BasicType;
import sk.tuke.bkpi.model.types.CartesianProduct;
import sk.tuke.bkpi.model.types.SetType;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeResolver;
import sk.tuke.bkpi.model.types.Z;

public class CsNativeArrayGenerator implements ArrayGenerator2 {
	
	private TypeResolver INDEXES = new TypeResolver();
	private TypeResolver ELEMENT_TYPE = new TypeResolver();
	private Type ARRAY_TYPE = PowerSet(CartesianProduct(INDEXES, ELEMENT_TYPE));
	
	private CsGenerator generator;
	
	private Map<Type, ArrayExpression> map = new Hashtable<Type, ArrayExpression>();

	public CsNativeArrayGenerator(CsGenerator generator) {
		this.generator = generator;
	}

	@Override
	public String arrayType(Type arrayType) {
		ARRAY_TYPE.resolveType(arrayType);
		Type indexes = INDEXES.resolvedType();
		
		Stack<Type> types = new Stack<Type>(); 
		while(indexes instanceof CartesianProduct) {
			types.push(((CartesianProduct)indexes).second());
			indexes = ((CartesianProduct)indexes).first();
		}
		types.push(indexes);
		return generator.csType(ELEMENT_TYPE.resolvedType())+Utilities.repeat("[]", types.size());
	}

	public void couplesSetInitialization(B0Variable arrayVariable, CoupleSet couplesSet, CodeGenerator codeGenerator) {
		StringBuilder statement;
		for (Couple couple : couplesSet.couples()) {
			statement = new StringBuilder();
			statement.append(generator.generateVariable(arrayVariable));
			
			B0DataValue index = couple.left();

			while (index instanceof Couple) {
				statement.append("["+arrayIndex(((Couple)index).left())+"]");
				index = ((Couple) index).right();
			}
			
			statement.append("["+arrayIndex(index)+"]");
			statement.append(" = " + generator.generateDataValue(couple.right().type(), couple.right(), codeGenerator)+";\n");
			codeGenerator.addStatement(statement.toString());
		}
	}
	
	public String arrayExpression(ArrayExpression array, CodeGenerator codeGenerator) {
		if (!map.containsKey(array.type())) {
			map.put(array.type(), array);
		}
		
		if (array instanceof ArrayInitialization) {
			return arrayInitialization2((ArrayInitialization) array);
		}
		String csArrayType = generator.csType(array.type());
		String locVarName = generator.getNameManager().generateUniqueName("array");
		LocalVariable arrayVar = new LocalVariable(locVarName, array.type());
		B0Variable var = new B0VariableReference(arrayVar);
		String arrayDef = arrayDefinition(array);
		
		codeGenerator.addDeclaration(format("%s %s = new %s;\n",
				csArrayType, generator.getNameManager().moduleData(arrayVar), arrayDef));
		
		
		if (array instanceof ArrayInitialization) {
			arrayInitialization(var, array, codeGenerator);
			return locVarName;
		} else if (array instanceof CoupleSet) {
			couplesSetInitialization(var, (CoupleSet) array, codeGenerator);
			return locVarName;
		}
		throw new UnsupportedOperationException(array.getClass().getName());
	}
	
	private String arrayDefinition(ArrayExpression array) {
		ARRAY_TYPE.resolveType(array.type());
		Type indexes = INDEXES.resolvedType();
		System.out.println(array);
		Stack<Type> types = new Stack<Type>();

		String dimmensions = "";
		while(indexes instanceof CartesianProduct) {
			Type indexType = ((CartesianProduct)indexes).second();
			types.push(indexType);
			indexes = ((CartesianProduct)indexes).first();
		}
		types.push(indexes);
		if (array instanceof ArrayInitialization) {
			for (B0DataValue range : ((ArrayInitialization)array).indexes()) {
				dimmensions += "["+generator.generateDataValue(rangeCardinality(range))+"]";
			}
		} else if (array instanceof CoupleSet) {
			CoupleSet coupleSet = (CoupleSet) array;
			System.out.println("\tcouple set");
			int indexesCount = coupleToList(coupleSet.couples().get(0)).size() - 1;
			List<Set<String>> usedIndexes = new ArrayList<Set<String>>();
			for (int i = 0; i < indexesCount; i++) {
				usedIndexes.add(new TreeSet<String>());
			}
			
			List<B0DataValue> list;
			for (Couple couple : coupleSet.couples()) {
				list = coupleToList(couple);
				for (int i = 0; i <list.size()-1; i++) { // only indexes
					String index = generator.generateDataValue(list.get(i));
					usedIndexes.get(i).add(index);
				}
			}
			System.out.println("\t"+usedIndexes);
			for (Set<String> set : usedIndexes) {
				//writter.println(set);
				dimmensions+="["+set.size()+"]";
			}
		} else {
			//throw new 
		}
		
		System.out.println("new "+generator.csType(ELEMENT_TYPE.resolvedType())+dimmensions);
		ARRAY_TYPE.resolveType(array.type());
		if (ELEMENT_TYPE.resolvedType().isSubtype(ARRAY_TYPE)) {
			CoupleSet coupleSet = (CoupleSet) array;
			return arrayDefinition((ArrayExpression) coupleSet.couples().get(0).right())+dimmensions;
		}
		ARRAY_TYPE.resolveType(array.type());
		return generator.csType(ELEMENT_TYPE.resolvedType())+dimmensions;
	}
	
	private List<B0DataValue> coupleToList(Couple couple) {
		List<B0DataValue> list = new ArrayList<B0DataValue>();
		Stack<B0DataValue> stack = new Stack<B0DataValue>();
		
		B0DataValue data = couple;
		while(data instanceof Couple) {
			stack.push(((Couple) data).right());
			data = ((Couple) data).left();
		}
		stack.push(data);
		
		while(!stack.isEmpty()) {
			list.add(stack.pop());
		}
		return list;
	}
	
	public String arrayInitialization2(ArrayInitialization initialization) {		
		return format("ArrayUtils.initialiseArray(new %s, %s)", arrayDefinition(initialization),
				generator.generateDataValue(initialization.value()));
	}
	
	private B0DataValue rangeCardinality(B0DataValue range) {
		if (range instanceof BooleanSetB0) {
			return new IntegerTerm(2);
		} else if (range instanceof IntegerSetB0) {
			return new IntegerTerm(Integer.MAX_VALUE); 
		} else if (range instanceof IntervalB0) {
			IntervalB0 interval = (IntervalB0) range;
			return new ArithmeticalBinaryExpression(Operator.PLUS, new ArithmeticalBinaryExpression(Operator.MINUS, interval.maximum(), interval.minimum()), new IntegerTerm(1));
		} else if (range instanceof B0VariableReference) {
			DataField variable = ((B0VariableReference) range).variable();
			//TODO check if constant
			B0DataValue value = generator.implementation().valueOf(variable);
			if (variable instanceof EnumeratedSet) {
				return new IntegerTerm(((EnumeratedSet) variable).items().size());
			}
			System.out.println(generator.implementation().valuations());
			System.out.println(value);
			value = generator.implementation().valuations2.get(variable.name());
			if (value == null) throw new NullPointerException();
			
			return rangeCardinality(value);
		}
		
		throw new CsCodeGenerationException("Unsupported range for cardinality: "+range.getClass());
	}

	@Override
	public String arrayInitialization(B0Variable arrayVariable,
			ArrayExpression initialization, CodeGenerator codeGenerator) {
		
		//if (!initialized.contains(arrayVariable)) {
			String csArrayType = generator.csType(initialization.type());
			String arrayDef = arrayDefinition(initialization);
			codeGenerator.addStatement(format("%s = new %s;\n", generator.generateVariable(arrayVariable), arrayDef));
			
		//}
		//initialized.add(arrayVariable);
		
			if (initialization instanceof ArrayInitialization) {
				ArrayInitialization arrayInit = (ArrayInitialization) initialization;
				List<String> localVars = new ArrayList<String>();
				
				char nextVar = 'i';
				for (B0DataValue range : arrayInit.indexes()) {
					localVars.add(""+nextVar);
					codeGenerator.addStatement(format("for (int %s = 0; i < %s; i++)", nextVar, generator.generateDataValue(rangeCardinality(range))));
					codeGenerator.openScope();
					nextVar = (char) (nextVar+1);
				}
				//codeGenerator.openScope();
				
				String statement = generator.generateVariable(arrayVariable);
				for (String var : localVars) {
					statement += "["+var+"]";
				}
				statement += " = " + generator.generateDataValue(arrayInit.value())+";\n";
				codeGenerator.addStatement(statement);
				
				for (B0DataValue range : arrayInit.indexes()) {
					codeGenerator.closeScope();
				}
				
			}
			return "";
	}

	
	public String generateExpression(Expression expr) {
		if (expr instanceof BinaryArithmeticalExpression) {
			return generateExpression((BinaryArithmeticalExpression) expr);
		}
		if (expr instanceof ExpressionReference) {
			return generateExpression((ExpressionReference) expr);
		}
		if (expr instanceof IntegerLiteral) {
			return generateExpression((IntegerLiteral) expr);
		}
		if (expr instanceof Negation) {
			return generateExpression((Negation) expr);
		}
		throw new CsCodeGenerationException("Unsuported arithmetical expression class: "+expr.getClass());
	}
	
	
	public String generateExpression(ExpressionReference expr) {
		return generator.generateDataField(expr.getData());
	}
	
	public String generateExpression(IntegerLiteral expr) {
		return expr.value().toString();
	}
	
	public String generateExpression(Negation expr) {
		return "-("+generateExpression(expr.expression())+")";
	}
	
	public String generateExpression(BinaryArithmeticalExpression expr) {
		String left = generateExpression(expr.getLeft());
		String right = generateExpression(expr.getRight());
		
		String operator;
		switch(expr.getOperator()) {
			case PLUS: operator = "+"; break;
			case MINUS: operator = "-"; break;
			case TIMES: operator = "*"; break;
			case DIVIDE: operator = "/"; break;
			case MODULO: operator = "%"; break;
			case POWER: return String.format("Math.pow(%s, %s)", left, right);
			default: throw new CsCodeGenerationException("Unsuported arithmetical expression operator: "+expr.getOperator());
		}
		return String.format("(%s)%s(%s)", left, operator, right);
	}
	
	@Override
	public String getElement(ArrayFieldVariable expr) {
		String result;
		result = generator.generateVariable(expr.array());
		for (B0DataValue indexValue : expr.indexes()) {
			if (indexValue.type().isEqual(Z)) {
				
				//B0VariableReference var = (B0VariableReference) expr.array();
				System.out.println("INDEX TYPE: "+indexValue.type()+", array type: "+expr.array().type());
				ARRAY_TYPE.resolveType(expr.array().type());
				System.out.println(INDEXES.resolvedType());
				if (INDEXES.resolvedType() instanceof Z) {
					IntervalExpression interval = ((Z) INDEXES.resolvedType()).interval();
					if (interval != null) {
						System.out.println(interval.getLeft());
						
						result += String.format("[%s-(%s)]", generator.generateDataValue(indexValue), generateExpression(interval.getLeft()));
						//result += String.format("[%s-(%s)]", generator.generateDataValue(indexValue), interval.getLeft().toString());
					}
				}
				//Object typing = var.variable().typing;
				//ArrayExpression array;
				//if (typing instanceof BelongingPredicate) {
					//array = (ArrayExpression) ((BelongingPredicate) typing).getRight();
				//}
				//throw new JavaCodeGenerationException("zle: "+var.variable().typing+" "+expr.array().getClass()+" map: "+map);
				//new BinaryArithmeticalExpression(Operator.MINUS, indexValue, );
				
				
			} else {
				result += "["+arrayIndex(indexValue)+"]";
			}
		}
		return result;
	}

	private String arrayIndex(B0DataValue value) {
		if (value.type() instanceof SetType) {
			return generator.generateDataValue(value)+".index";
		} else if (value.type() == BasicType.BOOL) {
			return generator.generateDataValue(value)+"? 1:0";
		} else if (value.type().isEqual(BasicType.Z)) {
			return generator.generateDataValue(value)+"-1";
		} else throw new CsCodeGenerationException("unsupported array index type: "+value.type());

	}
	
	@Override
	public String setElement() {
		// TODO Auto-generated method stub
		return null;
	}
}
