package sk.tuke.bkpi.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;

import sk.tuke.bkpi.extensions.ClassAccessPolicy;
import sk.tuke.bkpi.extensions.ClassModificationPolicy;
import sk.tuke.bkpi.generator.Utilities.ListFormatter;
import sk.tuke.bkpi.generator.support.AccessPolicyResolver;
import sk.tuke.bkpi.generator.support.AtributeSignature;
import sk.tuke.bkpi.generator.support.MethodSignature;
import sk.tuke.bkpi.generator.support.ModificationPolicyResolver;
import sk.tuke.bkpi.model.Module;
import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.model.b0data.*;
import sk.tuke.bkpi.model.clause.EnumeratedSet;
import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.IncludedMachine;
import sk.tuke.bkpi.model.clause.Operation;
import sk.tuke.bkpi.model.clause.OperationB0;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.clause.Set;
import sk.tuke.bkpi.model.clause.EnumeratedSet.Item;
import sk.tuke.bkpi.model.component.*;
import sk.tuke.bkpi.model.condition.*;
import sk.tuke.bkpi.model.condition.ComparisonCondition.Operator;
import sk.tuke.bkpi.model.data.*;
import sk.tuke.bkpi.model.instruction.*;
import sk.tuke.bkpi.model.instruction.IfConditionalInstruction.Branch;
import sk.tuke.bkpi.model.term.*;
import sk.tuke.bkpi.model.types.*;

import static sk.tuke.bkpi.model.types.BasicType.*;
import static sk.tuke.bkpi.model.types.TypeBuilder.*;
import static java.lang.String.*;

/**
 * B-Method code generator to Java programming language.
 * Output files will be generated to the specified folder.
 * 
 * @author Marcel Dancak
 */
public class JavaGenerator {

	
	private ModificationPolicyResolver MPR;
	private AccessPolicyResolver APR;
	
	public void setAccessPolicy (ArrayList <ClassAccessPolicy> languageAccessPolicy)
	{
		APR = new AccessPolicyResolver (languageAccessPolicy);
	}
	
	public void setModificationPolicy(ArrayList <ClassModificationPolicy> languageModificationPolicy)
	{
		MPR = new ModificationPolicyResolver (languageModificationPolicy);
	}
	
	private static final String[] javaKeywords = {"abstract", "assert", "boolean", "break", "byte", "case",
		"catch", "char", "class", "const", "continue", "default", "do", "double", "else", "enum", "extends",
		"final", "finally", "float", "for", "goto", "if", "implements", "import", "instanceof", "int",
		"interface", "long", "native", "new", "package", "private", "protected", "public", "return", "short",
		"static", "strictfp", "super", "switch", "synchronized", "this", "throw", "throws", "transient",
		"try", "void", "volatile", "while", "true", "false", "null"};
	
	 /*Atribute of Ivan Halupka */
	private static ArrayList<String> excludedWords = new ArrayList<String>(Arrays.asList("abstract", "assert", "boolean", "break", "byte", "case",
		"catch", "char", "class", "const", "continue", "default", "do", "double", "else", "enum", "extends",
		"final", "finally", "float", "for", "goto", "if", "implements", "import", "instanceof", "int",
		"interface", "long", "native", "new", "package", "private", "protected", "public", "return", "short",
		"static", "strictfp", "super", "switch", "synchronized", "this", "throw", "throws", "transient",
		"try", "void", "volatile", "while", "true", "false", "null"));
	
	private Project project;
	private File outputFolder;
	
	private Implementation implementation;
	private Module module;
	private StringWriter buffer;
	private JavaWriter out;
	private ArrayGenerator arrayGenerator = new JavaNativeArrayGenerator(this);
	private RecordGenerator recordGenerator = new RecordGenerator1(this);
	private NameManager nameManager;
	private String packageName;
	
	
	

	

	
	private Map<DataField, DataField> includedMapping;
	private Map<EnumeratedSet, Module> setsMapping;
	
	public void setJavaExcludedWords(ArrayList<String> javaExcludedWords)
	{
		excludedWords.addAll(javaExcludedWords);
		
	}
	
	public void resetJavaExcludedWords()
	{
		this.excludedWords =  new ArrayList<String>(Arrays.asList(this.javaKeywords));
	}
	
	public JavaGenerator() {}
	
	public JavaGenerator(Project project, File outputFolder) {
		this.project = project;
		this.outputFolder = outputFolder;
		/*
		for (Module m: project.modules())
		{
			System.out.println("Implconst : "+ m.implementation().name());
		}*/
		/*
		Map<String, Method> dataGenerator = new Hashtable<String, Method>();
		dataGenerator.put("sk.tuke.bkpi.model.term.ArrayExpression", null);
		
		ArrayExpression expr = null;
		try {
			dataGenerator.get(expr.getClass()).invoke(this, expr);
		} catch (Exception e) {
			e.printStackTrace();
		}
		*/
	}
	
	
	/*Method of Ivan Halupka */
	public static ArrayList<String> getExcludedWords() {
		return excludedWords;
	}


	public JavaGenerator(Implementation implementation, Writer out) {
		this.implementation = implementation;
		 
		this.out = new JavaWriter(out);
	}
	
	public JavaGenerator(Implementation implementation) {
		this.implementation = implementation;
		
		buffer = new StringWriter();
		out = new JavaWriter(new PrintWriter(buffer));
	}
	
	JavaWriter writter() {
		return out;
	}
	
	public Implementation implementation() {
		return implementation;
	}
	
	public PrintWriter newFileStream(String name) throws IOException {
		//FileOutputStream fileOut = new FileOutputStream(outputFolder.getAbsolutePath()+File.separator+name);
		return new PrintWriter(new FileWriter(outputFolder.getAbsolutePath()+File.separator+name+".java"));
	}
	
	public void generateJavaCode(Writer writer) throws JavaCodeGenerationException {
		
		if (!outputFolder.exists()) {
			if (!outputFolder.mkdir()) {
				throw new RuntimeException("Cannot create output directory");
			}
		}
		
		Module mainModule = null;
		String mainOperationName = null;
		if (project.rootModules().size() > 0) {
			mainModule = project.rootModules().get(0);
			for (Operation op : mainModule.machine().operations()) {
				if (op.inputParameters().size() == 0) {
					mainOperationName = op.name();
					break;
				}
			}
		}

		includedMapping = new Hashtable<DataField, DataField>();
		setsMapping = new Hashtable<EnumeratedSet, Module>();
		nameManager = new NameManager();
		nameManager.setKeywords(excludedWords); 
		nameManager.generateUniqueMachineNames(this.project);
		nameManager.resolveOperationNames(this.project);
		nameManager.SetRecordExcludedWords();
	    
	
	  
		
		if (mainModule != null && mainOperationName != null) {
			String mainClassName = this.nameManager.generateMainMachineName();
			File mainClass = new File(outputFolder.getAbsolutePath()+File.separator+mainClassName+".java");
			try {
				FileOutputStream fileOut = new FileOutputStream(mainClass);
				String main = "public class "+mainClassName+" {\n\n"+
					"\tpublic static void main(String[] args) {\n"+
					"\t\tnew "+nameManager.getUniqueMachineName(mainModule.machine().name())+"()."+ nameManager.get_resolved_operation_name(mainModule.machine(), mainOperationName) +"();\n"+ // TODO: mapping machine -> class name
					//"\t\tnew "+nameManager.getUniqueMachineName(mainModule.machine().name())+"()."+mainOperationName+"();\n"+ 
					"\t}\n"+
					"}";
				fileOut.write(main.getBytes());
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		/*
		includedMapping = new Hashtable<DataField, DataField>();
		setsMapping = new Hashtable<EnumeratedSet, Module>();
		nameManager = new NameManager();
		nameManager.setKeywords(excludedWords); 
		nameManager.generateUniqueMachineNames(this.project);
		*/
		
		// register module (mainly implementation) data fields that will
		// be used in generator. So any local/temporary variable with
		// the same name will cannot get that name
		for (Module module : project.modules()) {
			//skip abstract modules
			//if (module.implementation() == null)
			//	continue;
			
			List<DataField> moduleDataFields = new ArrayList<DataField>();
			//Machine machine = module.machine();
			//List<Refinement> refinements = module.refinements(); 
			Implementation implementation = module.implementation();
			
			
			
			for (Component c : module.components()) {
				for (EnumeratedSet set : c.enumeratedSets()) {
					setsMapping.put(set, module);
				}
				moduleDataFields.addAll(c.concreteConstants());
				for (Set set : c.sets()) {
					moduleDataFields.add(set);
					if (set instanceof EnumeratedSet) {
						moduleDataFields.addAll(((EnumeratedSet) set).items());
					}
				}
				moduleDataFields.addAll(c.concreteVariables());
			}
			for (DataField data : moduleDataFields) {
				nameManager.moduleData(module, data);
				//System.out.println("register: "+data);
			}
			
			// imports/sees
			if (implementation != null) {
				for (Import imp : implementation.imports()) { // is it really needed ? sees chains should also do this
					nameManager.componentReferenceName(module, imp);
				}
			}
			for (SeenMachine seen : module.lastComponent().seenMachines()) {
				nameManager.componentReferenceName(module, seen);
			}
		}
		
	
		for (Module module : project.modules()) {
			//System.out.println("Kurva " + module.implementation()); // tu je null
			if (module.lastComponent() instanceof Implementation) {

				File file = new File(outputFolder.getAbsolutePath()+File.separator+nameManager.getUniqueMachineName(module.machine().name())+".java");
				FileWriter fileOut = null;
				System.out.println(file.getAbsolutePath());
				try {
					fileOut = new FileWriter(file);
				} catch (FileNotFoundException e) {
					e.printStackTrace();
				} catch (IOException e) {
					e.printStackTrace();
				}
				
				buffer = new StringWriter();
				//out = new JavaWriter(new PrintWriter(buffer));
				out = new JavaWriter(fileOut);
				try {
					generateJavaCode(module, out);
				} catch (JavaCodeGenerationException jcge) {
					if (jcge.getSource() == null) {
						jcge.setSource(module);
					}
					throw jcge;
				} catch (Exception e) {
					throw new JavaCodeGenerationException(e, module);
				}
				
				try {
					fileOut.close();
				} catch (IOException e) {}
				//String javaCode = buffer.toString();
				//System.out.println(javaCode);
				
			}
		}
	}
	
	public NameManager getNameManager() {
		return nameManager; 
	}
	
	public void generateJavaCode(Module module, Writer writer) {
		this.module = module;
		//System.out.println("polka: " + module.implementation().name());
		this.implementation = module.implementation();
		nameManager.setWorkingModule(module);
		out = new JavaWriter(new PrintWriter(writer));
		generate();
		
	}
	
	private void find(DataField data) {
		List<Module> references = new ArrayList<Module>();
		DataField d = checkImported(module, data, references);
		references.remove(module);
		out.println("//"+references + " : "+d);
		out.println("//"+data.getClass() + " -> "+d.getClass() + "  "+data.hashCode() + " - "+ d.hashCode());
		if (d != null) {
			includedMapping.put(data, d);
		}
	}
	
	private DataField checkImported(Module module, DataField data, List<Module> references) {
		if (module.machine().dataByName(data.name()) != null) {
			references.add(module);
			return module.machine().dataByName(data.name());
		}
		if (module.implementation() != null) {
			for (Import imp : module.implementation().imports()) {
				DataField found = checkImported(project.moduleOfMachine(imp.machine()), data, references);
				if (found != null) {
					references.add(0, module);
					return found;
				}
			}
		}
		return null;
	}
	
	
    ArrayList <MethodSignature> methods = new ArrayList <MethodSignature>();
    ArrayList <AtributeSignature> atributes = new ArrayList <AtributeSignature>();
    ArrayList <MethodSignature> constructors = new ArrayList <MethodSignature>();
    
    // P = 0 NO POLICY APLIED
    // P = 1 FORBID REMOVING
    // P = 2 FORBID ADDING
    // P = 3 FORBID ADDING AND REMOVING
    
    private void generateModificationPolicyEnforcementMethod(int AP, int MP, int CP)
    {
      if ((AP != 0) || (MP != 0) || (CP != 0))
      {
    	  
    	  if (AP != 0)
    	  {  
    		  /*GENERATE CLASS AtributeSignature BEGIN*/   
    		  
    		  out.printf("private static class %s\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private Integer modifier;\n");
    		  out.printf("private String type;\n");
    		  out.printf("private String name;\n\n");
 
    		  
    		  out.printf("public %s (Integer modifier, String type, String name)\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.modifier = modifier;\n");
    		    out.printf("this.type = type;\n");
    		    out.printf("this.name = name;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public Integer getModifier()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.modifier;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getType()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.type;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getName()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.name;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("@Override\n");
    		  out.printf("public boolean equals (Object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O instanceof %s)\n", nameManager.generateUniqueClassName("AtributeSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s A = (%s) O;\n", nameManager.generateUniqueClassName("AtributeSignature"), nameManager.generateUniqueClassName("AtributeSignature"));
                  out.printf("if ((A.getModifier() == this.modifier) && (A.getType().equals(this.type)) && (A.getName().equals(this.name)))\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return super.getClass().equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n"); 		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS AtributeSignature END*/   		  
    	  }
    	  
    	  if (MP != 0)
    	  {
    		  /*GENERATE CLASS MethodSignature BEGIN*/
    		  
    		  out.printf("private static class %s\n", nameManager.generateUniqueClassName("MethodSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private Integer modifier;\n");
    		  out.printf("private String returnType;\n");
    		  out.printf("private String name;\n");
    		  out.printf("private ArrayList<String> parameterTypes;\n\n");
    		  
    		  out.printf("public %s (Integer modifier, String returnType, String name, ArrayList<String> parameterTypes)\n", nameManager.generateUniqueClassName("MethodSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.modifier = modifier;\n");
    		    out.printf("this.returnType = returnType;\n");
    		    out.printf("this.name = name;\n");
    		    out.printf("this.parameterTypes = parameterTypes;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public Integer getModifier()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.modifier;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getReturnType()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.returnType;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public String getName()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.name;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("public ArrayList<String> getParameterTypes()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.parameterTypes;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("@Override\n");
    		  out.printf("public boolean equals (Object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O instanceof %s)\n", nameManager.generateUniqueClassName("MethodSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s M = (%s) O;\n", nameManager.generateUniqueClassName("MethodSignature"), nameManager.generateUniqueClassName("MethodSignature"));
                  out.printf("if ((M.getModifier() == this.modifier) && (M.getReturnType().equals(this.returnType)) && (M.getName().equals(this.name)))\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("ArrayList <String> FA = new ArrayList <String> (M.getParameterTypes());\n");
    		        
    		        out.printf("for (String TAE: this.parameterTypes)\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
        		      out.printf("if (!FA.contains(TAE))\n");
          		      out.printf("{\n");
        		      out.indent(); 
        		      
        		        out.printf("return false;\n");
        		      
          		      out.unindent();
        		      out.printf("}\n"); 
        		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("FA.remove(TAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		      
        		    
        		    out.unindent();
        		    out.printf("}\n\n"); 
        		    
        		    out.printf("ArrayList <String> TA = new ArrayList <String> (this.parameterTypes);\n");
        		    
    		        out.printf("for (String FAE: M.getParameterTypes())\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
      		          out.printf("if (!TA.contains(FAE))\n");
      		          out.printf("{\n");
    		          out.indent(); 
    		      
    		          out.printf("return false;\n");
    		      
      		          out.unindent();
    		          out.printf("}\n");  
    		          
           		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("TA.remove(FAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		    
        		    out.unindent();
        		    out.printf("}\n\n");   
        		    
        		    out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return super.getClass().equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n"); 		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS MethodSignature END*/
    	  }
    	  
    	  if (CP != 0)
    	  {
    		  /*GENERATE CLASS ConstructorSignature BEGIN*/
    		  
    		  out.printf("private static class %s\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
  		      out.printf("private Integer modifier;\n");
    		  out.printf("private ArrayList<String> parameterTypes;\n\n");
    		  
    		  out.printf("public %s (Integer modifier, ArrayList<String> parameterTypes)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		  out.printf("{\n");
    		  out.indent();
    		  
    		    out.printf("this.modifier = modifier;\n");
    		    out.printf("this.parameterTypes = parameterTypes;\n");
    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  
    		  out.printf("public Integer getModifier()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.modifier;\n");
    		  out.unindent();
    		  out.printf("}\n\n");    		 
    		  
    		  out.printf("public ArrayList<String> getParameterTypes()\n");
    		  out.printf("{\n");
    		  out.indent();
    		    out.printf("return this.parameterTypes;\n");
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  out.printf("@Override\n");
    		  out.printf("public boolean equals (Object O)\n");
    		  out.printf("{\n");
    		  out.indent();    		  
    		  
    		    out.printf("if (O instanceof %s)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("%s C = (%s) O;\n", nameManager.generateUniqueClassName("ConstructorSignature"), nameManager.generateUniqueClassName("ConstructorSignature"));
                  out.printf("if (C.getModifier() == this.modifier)\n");
      		      out.printf("{\n");
    		      out.indent();                
 
    		        out.printf("ArrayList <String> FA = new ArrayList <String> (C.getParameterTypes());\n");
    		        
    		        out.printf("for (String TAE: this.parameterTypes)\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
        		      out.printf("if (!FA.contains(TAE))\n");
          		      out.printf("{\n");
        		      out.indent(); 
        		      
        		        out.printf("return false;\n");
        		      
          		      out.unindent();
        		      out.printf("}\n"); 
        		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("FA.remove(TAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		      
        		    
        		    out.unindent();
        		    out.printf("}\n\n"); 
        		    
        		    out.printf("ArrayList <String> TA = new ArrayList <String> (this.parameterTypes);\n");
        		    
    		        out.printf("for (String FAE: C.getParameterTypes())\n");
        		    out.printf("{\n");
        		    out.indent();   
        		    
      		          out.printf("if (!TA.contains(FAE))\n");
      		          out.printf("{\n");
    		          out.indent(); 
    		      
    		          out.printf("return false;\n");
    		      
      		          out.unindent();
    		          out.printf("}\n");  
    		          
           		      out.printf("else\n");
        		      out.printf("{\n");
        		      out.indent();
        		      
        		        out.printf("TA.remove(FAE);\n");
        		      
        		      out.unindent();
        		      out.printf("}\n");
        		    
        		    out.unindent();
        		    out.printf("}\n\n");   
        		    
        		    out.printf("return true;\n");
    		        
      		      out.unindent();
    		      out.printf("}\n"); 
      		      out.printf("else\n");
    		      out.printf("{\n");
    		      out.indent();
    		      
    		        out.printf("return false;\n");
    		      
      		      out.unindent();
    		      out.printf("}\n");                 
                  
    		    out.unindent();
    		    out.printf("}\n");
    		    out.printf("else\n");
    		    out.printf("{\n");
    		    out.indent();
    		    
    		      out.printf("return super.getClass().equals(O);\n");
    		    
    		    out.unindent();
    		    out.printf("}\n");
    		  
    		  out.unindent();
    		  out.printf("}\n"); 		  

    		  
    		  out.unindent();
    		  out.printf("}\n\n");
    		  
    		  /*GENERATE CLASS ConstructorSignature END*/
    	  }    	  
    	  
    	  
    	  
      	  out.printf("private static void %s ()\n", nameManager.generate_unique_method_name(this.implementation, "checkClassStructure"));
    	  out.printf("{\n");
    	  out.indent();

    	  
    	    if (MP != 0)
    	    {
    	    	out.printf("ArrayList <%s> generatedMethods = new ArrayList <%s>();\n\n", nameManager.generateUniqueClassName("MethodSignature"), nameManager.generateUniqueClassName("MethodSignature"));
    	    	
    	    	for (MethodSignature method: this.methods)
    	    	{
    	    		String outParameters = new String();
    	    		
                    for (String parameterType: method.parameterTypes)
                    {
                    	outParameters = outParameters + "\""+ parameterType +"\", "; 
                    }
                    
                    if (method.parameterTypes.size() > 0)
                    {
                    	outParameters = "Arrays.asList(" + outParameters.substring(0, (outParameters.length() - 2))+ ")";
                    }
                    
                    outParameters = "new ArrayList<String>(" + outParameters + ")";
                    
    	    		out.printf("generatedMethods.add (new %s(%d, \"%s\", \"%s\", %s));\n", nameManager.generateUniqueClassName("MethodSignature")
    	    				                                                 , method.modifiers
    	    				                                                 , method.returnType
    	    				                                                 , method.name
    	    				                                                 , outParameters);
    	    	}
    	    	
    	    	switch (MP) 
    	    	{
    	    	
    	    	case 1 : {
    	    		       
    	    		       out.printf("for (%s generatedMethod: generatedMethods)\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		         out.printf("Boolean containsMethod = false;\n");
    	    		         out.printf("for (Method classMethod :%s.class.getMethods())\n", nameManager.getUniqueMachineName());
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
    	    		           
    	    		           out.printf("for (Class  parameterType: classMethod.getParameterTypes())\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n\n");
    	    		           
    	    		           out.printf("if (generatedMethod.equals (new %s(classMethod.getModifiers(), classMethod.getReturnType().getSimpleName(), classMethod.getName(), parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("containsMethod = true;\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n");
    	    		           
    	    		           
    	    		           
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		         out.printf("if (containsMethod == false)\n");
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.out.println (\"Missing method: \" + Modifier.toString(generatedMethod.getModifier()) + \" \" + generatedMethod.getReturnType() + \" \" + generatedMethod.getName() + \" \" + generatedMethod.getParameterTypes().toString());\n" );
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");
    	    		
    	    		
    	    	         } break;
    	    	
    	    	case 2 : {
    	    		       out.printf("for (Method classMethod :%s.class.getMethods())\n", nameManager.getUniqueMachineName());
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		        out.printf("if (!classMethod.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
    	    		        out.printf("{\n");
    	    		        out.indent();
    	    		       
	    		             out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
	    		           
	    		             out.printf("for (Class  parameterType: classMethod.getParameterTypes())\n");
	    		             out.printf("{\n");
	    		             out.indent();
	    		           
	    		               out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
	    		           
	    		             out.unindent();
	    		             out.printf("}\n\n");
    	    		       
    	    		         out.printf("if (!generatedMethods.contains(new %s(classMethod.getModifiers(), classMethod.getReturnType().getSimpleName(), classMethod.getName(), parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.out.println (\"Forbidden method: \" + Modifier.toString(classMethod.getModifiers()) + \" \" + classMethod.getReturnType().getSimpleName() + \" \" + classMethod.getName() + \" \" + parameterTypes.toString());\n");
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		       
    	    		        out.unindent(); 
    	    		        out.printf("}\n");
    	    		        
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");		
    	    		
    	    	         } break;
    	    	
    	    	case 3 :  {
    	    		
	    		       out.printf("for (%s generatedMethod: generatedMethods)\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsMethod = false;\n");
	    		         out.printf("for (Method classMethod :%s.class.getMethods())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
	    		           
	    		           out.printf("for (Class  parameterType: classMethod.getParameterTypes())\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n\n");
	    		           
	    		           out.printf("if (generatedMethod.equals (new %s(classMethod.getModifiers(), classMethod.getReturnType().getSimpleName(), classMethod.getName(), parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsMethod = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsMethod == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Missing method: \" + Modifier.toString(generatedMethod.getModifier()) + \" \" + generatedMethod.getReturnType() + \" \" + generatedMethod.getName() + \" \" + generatedMethod.getParameterTypes().toString());\n" );
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");    	    		
    	    		         
	    		       
	    		       out.printf("for (Method classMethod :%s.class.getMethods())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classMethod.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       
    		             out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
    		           
    		             out.printf("for (Class  parameterType: classMethod.getParameterTypes())\n");
    		             out.printf("{\n");
    		             out.indent();
    		           
    		               out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
    		           
    		             out.unindent();
    		             out.printf("}\n\n");
	    		       
	    		         out.printf("if (!generatedMethods.contains(new %s(classMethod.getModifiers(), classMethod.getReturnType().getSimpleName(), classMethod.getName(), parameterTypes)))\n", nameManager.generateUniqueClassName("MethodSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Forbidden method: \" + Modifier.toString(classMethod.getModifiers()) + \" \" + classMethod.getReturnType().getSimpleName() + \" \" + classMethod.getName() + \" \" + parameterTypes.toString());\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	

    	    	          }break;
    	    	
    	    	
    	    	default : break;
    	        
    	    	}
	
    	    }
    	    
    	    if (CP != 0)
    	    {
    	    	out.printf("ArrayList <%s> generatedConstructors = new ArrayList <%s>();\n\n", nameManager.generateUniqueClassName("ConstructorSignature"), nameManager.generateUniqueClassName("ConstructorSignature"));
    	    	
    	    	for (MethodSignature constructor: this.constructors)
    	    	{
    	    		String outParameters = new String();
    	    		
                    for (String parameterType: constructor.parameterTypes)
                    {
                    	outParameters = outParameters + "\""+ parameterType +"\", "; 
                    }
                    
                    if (constructor.parameterTypes.size() > 0)
                    {
                    	outParameters = "Arrays.asList(" + outParameters.substring(0, (outParameters.length() - 2))+ ")";
                    }
                    
                    outParameters = "new ArrayList<String>(" + outParameters + ")";
                    
    	    		out.printf("generatedConstructors.add (new %s(%d, %s));\n", nameManager.generateUniqueClassName("ConstructorSignature")
    	    				                                                 , constructor.modifiers
    	    				                                                 , outParameters);
    	    	}
    	    	
    	    	switch (CP) 
    	    	{
    	    	
    	    	case 1 : {
    	    		       
    	    		       out.printf("for (%s generatedConstructor: generatedConstructors)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		         out.printf("Boolean containsConstructor = false;\n");
    	    		         out.printf("for (Constructor classConstructor :%s.class.getConstructors())\n", nameManager.getUniqueMachineName());
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
    	    		           
    	    		           out.printf("for (Class  parameterType: classConstructor.getParameterTypes())\n");
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n\n");
    	    		           
    	    		           out.printf("if (generatedConstructor.equals (new %s(classConstructor.getModifiers(), parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		           out.printf("{\n");
    	    		           out.indent();
    	    		           
    	    		             out.printf("containsConstructor = true;\n");
    	    		           
    	    		           out.unindent();
    	    		           out.printf("}\n");
    	    		           
    	    		           
    	    		           
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		         out.printf("if (containsConstructor == false)\n");
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.out.println (\"Missing constructor: \" + Modifier.toString(generatedConstructor.getModifier()) + \" %s \" + generatedConstructor.getParameterTypes().toString());\n", nameManager.getUniqueMachineName());
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		         
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");
    	    		
    	    		
    	    	         } break;
    	    	
    	    	case 2 : {
    	    		       out.printf("for (Constructor classConstructor :%s.class.getConstructors())\n", nameManager.getUniqueMachineName());
    	    		       out.printf("{\n");
    	    		       out.indent();
    	    		       
    	    		        out.printf("if (!classConstructor.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
    	    		        out.printf("{\n");
    	    		        out.indent();
    	    		       
	    		             out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
	    		           
	    		             out.printf("for (Class  parameterType: classConstructor.getParameterTypes())\n");
	    		             out.printf("{\n");
	    		             out.indent();
	    		           
	    		               out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
	    		           
	    		             out.unindent();
	    		             out.printf("}\n\n");
    	    		       
    	    		         out.printf("if (!generatedConstructors.contains(new %s(classConstructor.getModifiers(), parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
    	    		         out.printf("{\n");
    	    		         out.indent();
    	    		         
    	    		           out.printf("System.out.println (\"Forbidden constructor: \" + Modifier.toString(classConstructor.getModifiers()) + \" %s \"  + parameterTypes.toString());\n", nameManager.getUniqueMachineName());
    	    		         
    	    		         out.unindent();
    	    		         out.printf("}\n");
    	    		       
    	    		        out.unindent(); 
    	    		        out.printf("}\n");
    	    		        
    	    		       
    	    		       out.unindent();
    	    		       out.printf("}\n");		
    	    		
    	    	         } break;
    	    	
    	    	case 3 :  {
    	    		
	    		       out.printf("for (%s generatedConstructor: generatedConstructors)\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsConstructor = false;\n");
	    		         out.printf("for (Constructor classConstructor :%s.class.getConstructors())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
	    		           
	    		           out.printf("for (Class  parameterType: classConstructor.getParameterTypes())\n");
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n\n");
	    		           
	    		           out.printf("if (generatedConstructor.equals (new %s(classConstructor.getModifiers(), parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsConstructor = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsConstructor == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Missing constructor: \" + Modifier.toString(generatedConstructor.getModifier()) + \" %s \" + generatedConstructor.getParameterTypes().toString());\n", nameManager.getUniqueMachineName());
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");
	    		       
	    		       out.printf("for (Constructor classConstructor :%s.class.getConstructors())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classConstructor.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       
    		             out.printf("ArrayList <String> parameterTypes = new ArrayList <String>();\n");
    		           
    		             out.printf("for (Class  parameterType: classConstructor.getParameterTypes())\n");
    		             out.printf("{\n");
    		             out.indent();
    		           
    		               out.printf("parameterTypes.add(parameterType.getSimpleName());\n");
    		           
    		             out.unindent();
    		             out.printf("}\n\n");
	    		       
	    		         out.printf("if (!generatedConstructors.contains(new %s(classConstructor.getModifiers(), parameterTypes)))\n", nameManager.generateUniqueClassName("ConstructorSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Forbidden constructor: \" + Modifier.toString(classConstructor.getModifiers()) + \" %s \"  + parameterTypes.toString());\n", nameManager.getUniqueMachineName());
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	    		       

    	    	          }break;
    	    	
    	    	
    	    	default : break;
    	        
    	    	}
	
    	    }
    	    
    	    if (AP != 0)
    	    {
    	    	
    	    	out.printf("ArrayList <%s> generatedAtributes = new ArrayList <%s>();\n\n", nameManager.generateUniqueClassName("AtributeSignature"), nameManager.generateUniqueClassName("AtributeSignature"));
    	    	
    	    	for (AtributeSignature atribute: this.atributes)
    	    	{
    	    		if (atribute.modifiers != 2)
    	    		{
    	    		out.printf("generatedAtributes.add(new %s(%d, \"%s\", \"%s\"));\n", nameManager.generateUniqueClassName("AtributeSignature")
    	    				                                                  , atribute.modifiers
    	    				                                                  , atribute.type
    	    				                                                  , atribute.name);
    	    		}
    	    	}
    	    	
    	    	switch (AP)
    	    	{
    	    	case 1 : {
    	    		
	    		       out.printf("for (%s generatedAtribute: generatedAtributes)\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsAtribute = false;\n");
	    		         out.printf("for (Field classAtribute :%s.class.getFields())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		      	    		           
	    		           out.printf("if (generatedAtribute.equals (new %s(classAtribute.getModifiers(), classAtribute.getType().getSimpleName(), classAtribute.getName())))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsAtribute = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsAtribute == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Missing atribute: \" + Modifier.toString(generatedAtribute.getModifier()) + \" \" + generatedAtribute.getType() + \" \" + generatedAtribute.getName());\n" );
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");    	    		
    	    		        
    	    	         }break;
    	    	
    	    	case 2: {
    	    		
	    		       out.printf("for (Field classAtribute :%s.class.getFields())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classAtribute.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       	       
	    		         out.printf("if (!generatedAtributes.contains(new %s(classAtribute.getModifiers(), classAtribute.getType().getSimpleName(), classAtribute.getName())))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Forbidden atribute: \" + Modifier.toString(classAtribute.getModifiers()) + \" \" + classAtribute.getType().getSimpleName() + \" \" + classAtribute.getName());\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	    	    		
    	    		 
    	    	        }break;
    	    	
    	    	case 3: { 
    	    		
	    		       out.printf("for (%s generatedAtribute: generatedAtributes)\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		         out.printf("Boolean containsAtribute = false;\n");
	    		         out.printf("for (Field classAtribute :%s.class.getFields())\n", nameManager.getUniqueMachineName());
	    		         out.printf("{\n");
	    		         out.indent();
	    		      	    		           
	    		           out.printf("if (generatedAtribute.equals (new %s(classAtribute.getModifiers(), classAtribute.getType().getSimpleName(), classAtribute.getName())))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		           out.printf("{\n");
	    		           out.indent();
	    		           
	    		             out.printf("containsAtribute = true;\n");
	    		           
	    		           out.unindent();
	    		           out.printf("}\n");
	    		           
	    		           
	    		           
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		         out.printf("if (containsAtribute == false)\n");
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Missing atribute: \" + Modifier.toString(generatedAtribute.getModifier()) + \" \" + generatedAtribute.getType() + \" \" + generatedAtribute.getName());\n" );
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		         
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");  
	    		       
	    		       out.printf("for (Field classAtribute :%s.class.getFields())\n", nameManager.getUniqueMachineName());
	    		       out.printf("{\n");
	    		       out.indent();
	    		       
	    		        out.printf("if (!classAtribute.getDeclaringClass().getSimpleName().equals(\"Object\"))\n");
	    		        out.printf("{\n");
	    		        out.indent();
	    		       	       
	    		         out.printf("if (!generatedAtributes.contains(new %s(classAtribute.getModifiers(), classAtribute.getType().getSimpleName(), classAtribute.getName())))\n", nameManager.generateUniqueClassName("AtributeSignature"));
	    		         out.printf("{\n");
	    		         out.indent();
	    		         
	    		           out.printf("System.out.println (\"Forbidden atribute: \" + Modifier.toString(classAtribute.getModifiers()) + \" \" + classAtribute.getType().getSimpleName() + \" \" + classAtribute.getName());\n");
	    		         
	    		         out.unindent();
	    		         out.printf("}\n");
	    		       
	    		        out.unindent(); 
	    		        out.printf("}\n");
	    		        
	    		       
	    		       out.unindent();
	    		       out.printf("}\n");	
    	    		
    	    	        } break;
    	    	
    	    	default: break;
    	    	
    	    	}
    	    		
    	    }

      	  out.unindent();
    	  out.printf("}\n"); 
    	   
      }
    }
    
    private void generateConstructorAccessCheck ()
    {
    	
    	out.printf("private static void %s ()\n", nameManager.generate_unique_method_name(this.implementation, "checkConstructorAccessRight"));
    	out.printf("{\n");
    	out.indent();
    	
    	  out.printf("ArrayList <String> constructorAccessClasses = new ArrayList <String>();\n");
    	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
    	  
    	  for (Module foreignModule: this.project.modules())
    	  {
    		  for (Import foreignImport: foreignModule.implementation().imports())
    		  {
    			  
    			  if (foreignImport.machine().name().equals(this.module.machine().name()))
    			  {
    				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
    				  { 
    					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    					  out.printf("constructorAccessClasses.add(\"%s\");\n", nameManager.getUniqueMachineName(foreignModule.machine().name()));
    				  }
    			  }  
    		  }
    	  }
    	  
  		if (project.rootModules().size() > 0) {
  		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
  		  {
  			out.printf("constructorAccessClasses.add(\"%s\");\n", nameManager.generateMainMachineName());  
  		  }
  		}
			
    	  out.printf("if (!constructorAccessClasses.contains(sun.reflect.Reflection.getCallerClass(3).getSimpleName()))\n");
    	  out.printf("{\n");
    	  out.indent();
    	  
    	    out.printf("System.out.println (\"Class \" + sun.reflect.Reflection.getCallerClass(3).getSimpleName() + \" cannot create instance of class %s\");\n", nameManager.getUniqueMachineName());
    	  
    	  out.unindent();
    	  out.printf("}\n");
    	
    	out.unindent();
    	out.printf("}\n");
        	
    }
    
    
    private void generateMethodAccessRight ()
    {   
    	
    	out.printf("private static void %s(String methodName)\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
    	out.printf("{\n");
    	out.indent();
    	
    	
    	out.printf("ArrayList <String> methodAccessClasses = new ArrayList<String>();\n");
    	
    	for (Module Pmodule: project.modules())
    	{
    		out.printf("methodAccessClasses.add(\"%s\");\n", nameManager.getUniqueMachineName(Pmodule.machine().name()));
    	}
    	
    	out.printf("methodAccessClasses.add(\"%s\");\n", nameManager.generateMainMachineName());
    	
    	out.printf("if (methodAccessClasses.contains(sun.reflect.Reflection.getCallerClass(3).getSimpleName()))\n"); 
    	out.printf("{\n");
    	out.indent();
    	
    	  out.printf("System.out.println(\"Method \" + methodName + \" cannot be called from class \" + sun.reflect.Reflection.getCallerClass(3).getSimpleName());\n"); 
    	
    	out.unindent();
    	out.printf("}\n");
    	
    	
    	out.unindent();
    	out.printf("}\n");
    	
    }
    
    private void generateAccessRightComments()
    {
    	
  	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
	  
  	  out.printf("/*Instances of class %s can be created by classes :\n", nameManager.getUniqueMachineName());
  	  out.indent();
	  for (Module foreignModule: this.project.modules())
	  {
		  for (Import foreignImport: foreignModule.implementation().imports())
		  {
			  
			  if (foreignImport.machine().name().equals(this.module.machine().name()))
			  {
				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
				  { 
					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
					  out.printf("%s\n", nameManager.getUniqueMachineName(foreignModule.machine().name()));
				  }
			  }  
		  }
	  }
	  
		if (project.rootModules().size() > 0) {
		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
		  {
			out.printf("%s\n", nameManager.generateMainMachineName());  
		  }
		}
    	
    	out.unindent();
    	out.printf("*/\n");
    	
    	out.printf("/*Methods of class %s can be invoked by classes :\n", nameManager.getUniqueMachineName());
    	out.indent();
    	
    	for (Module Pmodule: project.modules())
    	{
    		out.printf("%s\n", nameManager.getUniqueMachineName(Pmodule.machine().name()));
    	}
    	
    	out.printf("methodAccessClasses.add(\"%s\");\n", nameManager.generateMainMachineName());
    	out.unindent();
    	out.printf("*/\n");
    }
    
    
    private Hashtable <String, ArrayList<String>> projectConsructorAccessRights = new Hashtable <String, ArrayList<String>>();
    private Hashtable <String, ArrayList<String>> projectMethodAccessRights = new Hashtable <String, ArrayList<String>>();
    
    public Hashtable <String, ArrayList<String>> getProjectConstructorAccessRight()
    {
    	return this.projectConsructorAccessRights;
    }
    
    public Hashtable <String, ArrayList<String>> getProjectMethodAccessRight()
    {
    	return this.projectMethodAccessRights;
    }
    
    private void generateAccessRightReport()
    {
    	
    	  ArrayList <String> constructorAccessClasses = new ArrayList <String>();
    	  
    	  this.projectConsructorAccessRights.put(nameManager.getUniqueMachineName(), new ArrayList<String>());
    	  this.projectMethodAccessRights.put(nameManager.getUniqueMachineName(), new ArrayList<String>());
    	  
    	  for (Module foreignModule: this.project.modules())
    	  {
    		  for (Import foreignImport: foreignModule.implementation().imports())
    		  {
    			  
    			  if (foreignImport.machine().name().equals(this.module.machine().name()))
    			  {
    				  if (!constructorAccessClasses.contains(nameManager.getUniqueMachineName(foreignImport.machine().name())))
    				  { 
    					  constructorAccessClasses.add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    					  this.projectConsructorAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.getUniqueMachineName(foreignModule.machine().name()));
    				  }
    			  }  
    		  }
    	  }
    	  
    		if (project.rootModules().size() > 0) {
    		  if (project.rootModules().get(0).machine().name().equals(this.module.machine().name()))
    		  {
    			  this.projectConsructorAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.generateMainMachineName());
    		  }
    		}
        	
        	
        	for (Module Pmodule: project.modules())
        	{
        		this.projectMethodAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.getUniqueMachineName(Pmodule.machine().name()));
        	}
        	this.projectMethodAccessRights.get(nameManager.getUniqueMachineName()).add(nameManager.generateMainMachineName());
    	
    }
    


	
	private void generate() {

		
		
		/*Class structure Begin */
		
		this.methods = new ArrayList <MethodSignature>();
		this.atributes = new ArrayList <AtributeSignature>();	
		this.constructors = new ArrayList <MethodSignature>();
		
		/*Class structure End*/
	   	

		/*
		for (Set SET: this.implementation.abstractMachine().sets())
		{
			
						System.out.println("SET NAME: " + SET.name());
			
		}*/
		
		System.out.println("Generate code for: "+implementation+"\n"); // je null, why ?
		
		List<Refinement> refinements = new ArrayList<Refinement>();
		AbstractComponent refinedComponent = implementation.refinedComponent();
		while (refinedComponent instanceof Refinement) {
			refinements.add((Refinement) refinedComponent);
			refinedComponent = ((Refinement)refinedComponent).refinedComponent();
		}
		Machine machine = (Machine) refinedComponent;
		
		String generics = "";
		
		if (machine.setParameters().size() > 0) {
			ListFormatter typesList = new ListFormatter();
			for (SetParameter set : machine.setParameters()) {
				typesList.addItem(set.name());
				// TODO: check if extracting from PowerSet is the only way
				typesAssociationTable.put(((PowerSet)set.type()).baseType(), set.name());
			}
			generics = "<"+typesList.text()+">";
		} // TODO DEAD CODE
		for (SeenMachine seenMachine : implementation.seenMachines()) {
			for (EnumeratedSet set : seenMachine.machine().enumeratedSets()) {
				//typesAssociationTable.put(set.type(), seenMachine.machine().name()+"."+set.name());
				//typesAssociationTable.put(((PowerSet)set.type()).baseType(), seenMachine.machine().name()+"."+set.name());
			}
		}
		
		
		
		if (packageName != null)
			out.printf("package %s;\n\n", packageName);
		if(
		MPR.getClassModificationPolicy(machine.name()).getAtributePolicy().index != 0 ||
		MPR.getClassModificationPolicy(machine.name()).getMethodPolicy().index != 0 ||
		MPR.getClassModificationPolicy(machine.name()).getConstructorPolicy().index != 0
		)
		{
			out.printf("import java.lang.reflect.Field;\n");
			out.printf("import java.lang.reflect.Method;\n");
			out.printf("import java.lang.reflect.Modifier;\n");
			out.printf("import java.lang.reflect.Constructor;\n");
			out.printf("import java.util.ArrayList;\n");
			out.printf("import java.util.Arrays;\n");
			
		}
		
		if(
				MPR.getClassModificationPolicy(machine.name()).getAtributePolicy().index == 0 &&
				MPR.getClassModificationPolicy(machine.name()).getMethodPolicy().index == 0 &&
				MPR.getClassModificationPolicy(machine.name()).getConstructorPolicy().index == 0
				)
		{

			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("import java.util.ArrayList;\n");
			}
			
		}
		
	    /* Comented By Ivan Halupka
		out.printf("public class %s%s {\n", machine.name(), generics);*/
		if (APR.applyAllForbyddenExceptAllowedByComment(machine.name()))
		{
		 generateAccessRightComments();
		}
		
		if (APR.applyAllForbyddenExceptAllowedByReport(machine.name()))
		{
			generateAccessRightReport();
		}
		out.printf("public class %s%s {\n", nameManager.getUniqueMachineName(), generics);
		out.indent();
		
		
		// SETS
		for (Component component : module.components()) {
			generateSets(component);
		}
		for (Set set : machine.includedSets()) {
			find(set);
		}
		
		// CONSTANTS
		for (Component component : module.components()) {
			generateConstants(component);
		}
		
		// VARIABLES
		for (Component component : module.components()) {
			generateVariables(component);
		}
		
		// EXTENDED MACHINES
		if (implementation.extendedMachines().size() > 0) {
			out.println("//Extended machines");
			generateCompositionChains(implementation.extendedMachines());
		}
		
		// IMPORTED MACHINES
		if (implementation.imports().size() > 0) {
			out.println("//Imported machines");
			generateCompositionChains(implementation.imports());
		}
		
		
		// SEEN MACHINES
		if (implementation.seenMachines().size() > 0) {
			out.println("//Seen machines");
		}
		for (SeenMachine seenMachine : implementation.seenMachines()) {
			if (seenMachine.isImported()) {
				out.printf("private %s %s;\n", nameManager.getUniqueMachineName(seenMachine.machine().name()), nameManager.componentReferenceName(module, seenMachine));
                
				
				this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(seenMachine.machine().name()), nameManager.componentReferenceName(module, seenMachine)));
				
				// TODO BLINDSIGHT out.printf("private %s %s;\n", seenMachine.machine().name(), nameManager.componentReferenceName(module, seenMachine));
			}
		}
		
		// scalar formal parameters
		if (machine.scalarParameters().size() > 0) {
			out.println();
			out.println("// scalar formal parameters");
		}
		for (ScalarParameter scalar : machine.scalarParameters()) {
			out.printf("private %s %s;\n", javaType(scalar.type()), nameManager.moduleData(module, scalar));
			this.atributes.add(new AtributeSignature (2, javaType(scalar.type()), nameManager.moduleData(module, scalar)));
		}
		// CONSTRUCTOR
		
		MethodSignature mainConstructor = new MethodSignature();
		
		ListFormatter argsList = new ListFormatter();
		for (ScalarParameter scalar : machine.scalarParameters()){
			argsList.addItem(javaType(scalar.type()) +" "+ nameManager.moduleData(module, scalar));
		    mainConstructor.parameterTypes.add(javaType(scalar.type()));
		}
			//argsList.addItem(javaType(scalar.type()) +" "+ scalar.name());
		
		// collect seen references without duplications
		List<SeenMachine> seenMachinesReferences = new ArrayList<SeenMachine>();
		for (SeenMachine seenMachine : module.seesReferencesDelegations()) {
			if (!seenMachinesReferences.contains(seenMachine))
				seenMachinesReferences.add(seenMachine);
		}
		for (SeenMachine seenMachine : implementation.seenMachines()) {
			if (seenMachine.isImported() && !seenMachinesReferences.contains(seenMachine))
				seenMachinesReferences.add(seenMachine);
		}
		for (SeenMachine seenMachine : seenMachinesReferences) {
			argsList.addItem(nameManager.getUniqueMachineName(seenMachine.machine().name())+" "+nameManager.componentReferenceName(module, seenMachine));
			mainConstructor.parameterTypes.add(nameManager.getUniqueMachineName(seenMachine.machine().name()));
			// TODO BLINDSIGHT argsList.addItem(seenMachine.machine().name()+" "+nameManager.componentReferenceName(module, seenMachine));
		}
		
		out.println();
		out.printf("public %s(%s) {\n", nameManager.getUniqueMachineName(machine.name()), argsList.text());
		mainConstructor.modifiers = 1;
		this.constructors.add(mainConstructor);
		
		
		out.indent();
		nameManager.openScope();
	
		if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
		{
			out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkConstructorAccessRight"));
		}
		
		if (MPR.ApplyModificationPolicy(machine.name()))
		{
			out.printf("checkClassStructure();\n");
		}
	
		
		for (ScalarParameter scalar : machine.scalarParameters()) {
			out.printf("this.%s = %s;\n", nameManager.moduleData(module, scalar), nameManager.moduleData(module, scalar));
			
			
			//out.printf("this.%s = %s;\n", scalar.name(), scalar.name());
		}
		
		// Create imported machines instances
		
		//out.println("//unsorted: "+implementation().imports());
		//out.println("//sorted: "+module.sortedImports());
		
		List<Import> allCompositions = new ArrayList<Import>(implementation.imports());
		allCompositions.addAll(implementation.extendedMachines());
		
		allCompositions = module.sortedImports(allCompositions);
		for (Import importedMachine : allCompositions) {
			generateMachinesInstantiation(importedMachine);
		}
		/*
		for (Import importedMachine : module.sortedImports()) {
			generateMachinesInstanciantion(importedMachine);
		}
		*/
		// Assign seen machines references
		for (SeenMachine seenMchine : implementation.seenMachines()) {
			if (seenMchine.isImported()) {
				String seenMachineReference = nameManager.componentReferenceName(module, seenMchine);
				out.printf("this.%s = %s;\n", seenMachineReference, seenMachineReference);
			}
		}
		
		if (implementation.initialization() != null) //TODO this should be modified
			generateInstruction(implementation.initialization());
		
		
		nameManager.closeScope();
		out.unindent();
		
		out.println("}\n"); // constructor
		
		
		// OPERATIONS
		List<OperationB0> operations = implementation.getOperations();
		for (OperationB0 operation : operations) { 
			generateMethodSignature(nameManager.get_resolved_operation_name(implementation, operation.name()), operation.inputParameters(), operation.outputParameters());
          	//		generateMethodSignature(operation.name(), operation.inputParameters(), operation.outputParameters());
			out.print(" {\n");
			out.indent();
			nameManager.openScope();
			//TODO BS
			
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			
			for (OperationOutputParameter outParameter : operation.outputParameters()) {
				out.printf("%s %s;\n", javaType(outParameter.type()), nameManager.moduleData(outParameter));
			}
			generateInstruction(operation.body());
			if (operation.outputParameters().size() == 1) {
				out.printf("return %s;\n", nameManager.moduleData(operation.outputParameters().get(0)));
				//out.printf("return %s;\n", operation.outputParameters().get(0).name());
			}
			if (operation.outputParameters().size() > 1) {
				String retVarName = nameManager.generateUniqueName("ret");// TODO register retVarName in nameManager ? is it needed ?
				//out.printf("java.util.ArrayList<Object> %s = new java.util.ArrayList<Object>(%d);\n", retVarName, operation.outputParameters().size());
				//for (OperationOutputParameter outParam : operation.outputParameters()) {
				//	out.printf("%s.add(%s);\n", retVarName, nameManager.moduleData(outParam));
				//}
				//out.printf("return %s;\n", retVarName);
				
				ListFormatter outParams = new ListFormatter();
				for (OperationOutputParameter outParam : operation.outputParameters()) {
					outParams.addItem(nameManager.moduleData(outParam));
				}
				out.printf("return new Object[]{%s};\n", outParams.text());
			}
			nameManager.closeScope();
			out.unindent();
			out.println("}\n");
		}
		
		List<OperationDeclaration> promotedOperations = new ArrayList<OperationDeclaration>(implementation.promotedOperations());
		for (Import i : implementation.extendedMachines()) {
			promotedOperations.addAll(i.operationsDeclarations());
		}
		for (OperationDeclaration promotedOperation : promotedOperations) {
			generateMethodSignature( nameManager.get_resolved_operation_name(implementation, promotedOperation.operationName()), promotedOperation.inputParameters(), promotedOperation.outputParameters());
			//  generateMethodSignature(promotedOperation.operationName(), promotedOperation.inputParameters(), promotedOperation.outputParameters());
			out.print(" {\n");
			out.indent();
			
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			
			assert promotedOperation.source() != null;
			
			String delegateObject = nameManager.componentReferenceName(promotedOperation.source());
			ListFormatter inputParams = new ListFormatter();
			for (OperationInputParameter inputParam : promotedOperation.inputParameters())
				inputParams.addItem(nameManager.moduleData(inputParam));
			
			//inputParams.addItem(inputParam.name());			
			String format = promotedOperation.outputParameters().size() > 0? "return %s.%s(%s);\n" : "%s.%s(%s);\n";
			out.printf(format, delegateObject, nameManager.get_resolved_operation_name(promotedOperation.source().machine(), promotedOperation.operationName()), inputParams.text());
			//out.printf(format, delegateObject, promotedOperation.operationName(), inputParams.text());
			out.unindent();
			out.println("}");
		}
		
		// generate methods to return imported machines objects
		for (Import importedMachine : implementation.imports()) { // TODO check if other elements uses imported machine access operation
			out.printf("public %s %s() {\n", nameManager.getUniqueMachineName(importedMachine.machine().name()), nameManager.generate_unique_method_name(implementation, nameManager.componentReferenceName(importedMachine)));
//			out.printf("public %s %s() {\n", nameManager.getUniqueMachineName(importedMachine.machine().name()), nameManager.componentReferenceName(importedMachine));
            //AddedByMe

			MethodSignature MS = new MethodSignature();
			MS.name = nameManager.generate_unique_method_name(implementation, nameManager.componentReferenceName(importedMachine));
			MS.modifiers = 1;
			MS.returnType = nameManager.getUniqueMachineName(importedMachine.machine().name());
			this.methods.add(MS);
			
			out.indent();
			if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
			{
				out.printf("%s();\n", nameManager.generate_unique_method_name(this.implementation, "checkMethodAccessRight"));
			}
			out.printf("return %s;\n", nameManager.componentReferenceName(importedMachine));
			out.unindent();
			out.println("}\n");
		}
		
		
		generateModificationPolicyEnforcementMethod(MPR.getClassModificationPolicy(machine.name()).getAtributePolicy().index, MPR.getClassModificationPolicy(machine.name()).getMethodPolicy().index, MPR.getClassModificationPolicy(machine.name()).getConstructorPolicy().index);
		
		
		if (APR.applyAllForbyddenExceptAllowedByCode(machine.name()))
		{
		  generateConstructorAccessCheck();
		  generateMethodAccessRight();
		}
		
		
		out.unindent();
		out.println("}"); // class
		
		
		//recordGenerator.generateRecordsClasses(this);
		
		out.flush();
		//System.out.println(buffer);
		
	 
	}
	
	Object[] ret() {
		return new Object[]{"g", 4};
	}
	private void generateCompositionChains(List<Import> compositions) {
		for (Import importedMachine : compositions) {
			String name = nameManager.componentReferenceName(module, importedMachine);
			if (importedMachine.machine().setParameters().size() > 0) {
				ListFormatter typesList = new ListFormatter();
				for (SetParameter setParam : importedMachine.machine().setParameters()) {
					
					Type paramType = importedMachine.formalParametersValues().get(setParam.index()).type();
					if (paramType instanceof PowerSet) {
						typesList.addItem(javaType(((PowerSet) paramType).baseType()));
					} else {
						throw new JavaCodeGenerationException("Set Parameter has wrong type: "+paramType, module);
					}
					
				}
				out.printf("private %s<%s> %s;\n", nameManager.getUniqueMachineName(importedMachine.machine().name()), typesList.text(), name);
				this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(importedMachine.machine().name()), name));
			} else
			{
				out.printf("private %s %s;\n", nameManager.getUniqueMachineName(importedMachine.machine().name()), name);
				this.atributes.add(new AtributeSignature (2, nameManager.getUniqueMachineName(importedMachine.machine().name()), name));
			}   
		}
	}
	
	private void generateMachinesInstantiation(Import importedMachine) { //TODO deeper analyse
		
		String name = nameManager.componentReferenceName(module, importedMachine);
		
		ListFormatter values = new ListFormatter();
		for (ScalarParameter scalarParam : importedMachine.machine().scalarParameters()) {
			System.out.println(importedMachine);
			System.out.println(importedMachine.machine().scalarParameters());
			values.addItem(generateDataValue(importedMachine.formalParametersValues().get(scalarParam.index())));
		}
		// Seen machines references
		// as refinements must preserve previous sees, to get all seen machines
		// is used last refinement/implementation
		List<SeenMachine> seenReferences = new ArrayList<SeenMachine>();
		for (SeenMachine seen : project.moduleOfMachine(importedMachine.machine()).seesReferencesDelegations()) {
			if (!seenReferences.contains(seen))
				seenReferences.add(seen);
		}
		for (SeenMachine seen : project.moduleOfMachine(importedMachine.machine()).lastComponent().seenMachines()) {
			if (!seenReferences.contains(seen))
				seenReferences.add(seen);
		}
		for (SeenMachine seenMachine : seenReferences) {
			if (module.seesReferencesDelegations().contains(seenMachine)) {
				values.addItem(nameManager.componentReferenceName(module, seenMachine));
			} else {
				List<Import> importPath = module.getImportPath(seenMachine);
				String reference = "this."+nameManager.componentReferenceName(module, importPath.get(0));
				for (int i = 1 ; i < importPath.size(); i++) {
					reference += "."+nameManager.componentReferenceName(project.moduleOfMachine(importPath.get(1).machine()), importPath.get(1))+"()";
				}
				//values.addItem(module.getImportPath(seenMachine).toString());
				values.addItem(reference);
			}
		}
		
		if (importedMachine.machine().setParameters().size() > 0) {
			ListFormatter types = new ListFormatter();
			for (SetParameter setParam : importedMachine.machine().setParameters()) {
				Type paramType = importedMachine.formalParametersValues().get(setParam.index()).type();
				if (paramType instanceof PowerSet) {
					types.addItem(javaType(((PowerSet) paramType).baseType()));
				} else {
					throw new JavaCodeGenerationException("Set Parameter has wrong type: "+paramType, module);
				}
			}
			out.printf("this.%s = new %s<%s>(%s);\n", name, nameManager.getUniqueMachineName(importedMachine.machine().name()), types.text(), values.text(), name);
		} else {
			out.printf("this.%s = new %s(%s);\n", name, nameManager.getUniqueMachineName(importedMachine.machine().name()), values.text(), name);
		}
		
	}
	private void generateMethodSignature(String operationName, List<OperationInputParameter> inputParameters, List<OperationOutputParameter> outputParameters) {
		//format input parameters to Java
		
		//AddedbyMe
		MethodSignature MS = new MethodSignature();
		MS.name = operationName;
		
		ListFormatter inputParams = new ListFormatter();
		for (OperationInputParameter inputParam : inputParameters){
			inputParams.addItem(javaType(inputParam.type()) +" "+ nameManager.moduleData(module, inputParam));
			//AddedByMe
			MS.parameterTypes.add(javaType(inputParam.type()));
			//inputParams.addItem(javaType(inputParam.type()) +" "+ inputParam.name());
		}
		
		
		String outParamType;
		switch (outputParameters.size()) {
			case 0: outParamType = "void";break;
			case 1: outParamType = javaType(outputParameters.get(0).type());break;
			default: outParamType = "Object[]";
		}
		
		out.printf("public %s %s(%s)", outParamType, operationName, inputParams.text());
		//out.printf("public %s %s(%s)", outParamType, operationName, inputParams.text());
		
		//AddedByMe
        MS.modifiers = 1;
        MS.returnType = outParamType;
        this.methods.add(MS);
	
	}
	
	private void generateSets(Component component) {
		if (component.enumeratedSets().size() > 0) {
			out.println();
			out.println("// sets of "+component);
			for (EnumeratedSet set : component.enumeratedSets()) {
				String javaName = nameManager.moduleData(module, set);
				out.printf("public enum %s {\n", javaName);
				out.indent();
				//for (Item item : set.items()) {
				//	out.printf("%s,\n", item.name());
				//}
				
				int index = 0;
				for (Item item : set.items()) {
					out.printf("%s(%d),\n", nameManager.moduleData(module, item), index++);
				}
				out.println(";");
				out.println("public final int index;\n");
				out.printf("%s(int index) {\n", javaName);
				out.println("\tthis.index = index;");
				out.println("}");
				
				out.unindent();
				out.println("}");
			}
		} 
		
		// Deferred sets ?
		/*
		for (Set unknown_set: component.sets())
		{
			boolean is_deffered_set = true;
			for (EnumeratedSet enumerated_set: component.enumeratedSets())
			{
				if (unknown_set.name().equals(enumerated_set.name()))
				{
					is_deffered_set = false;
				}
			}
			
			if (is_deffered_set == true)
			{
				if (this.implementation.valuations().get(unknown_set) instanceof IntervalB0)
				{
					IntervalB0 interval = (IntervalB0) this.implementation.valuations().get(unknown_set);
					out.printf ("public static class " + unknown_set.name() + "{\n\n");
					out.indent();
					out.printf ("private static Integer minimum = " + generateDataValue (interval.minimum()) + ";\n");
					out.printf ("private static Integer maximum = " + generateDataValue (interval.maximum()) + ";\n\n");
					
					out.printf ("public static Integer getMinimum() {\n");
					out.indent();
					out.printf ("return "+ unknown_set.name() + ".minimum;\n");
					out.unindent();
					out.printf ("}\n\n");
					
					out.printf ("public static Integer getMaximum() {\n");
					out.indent();
					out.printf ("return "+ unknown_set.name() + ".maximum;\n");
					out.unindent();
					out.printf ("}\n");
			
					out.unindent();
					out.printf ("}\n");
				}
				
				System.out.println ("Hello :" + javaType (this.implementation.valuations().get(unknown_set).type()));
				System.out.println ("Hello :" + this.implementation.valuations().get(unknown_set).getClass().toString());
				
				

				
			}
		}*/
	}
	

	
	private void generateConstants(Component component) {
		if (component.concreteConstants().size() < 1)
			return;
		out.println();
		out.println("// constants of "+component);
		generateConstants: for (ConcreteConstant constant : component.concreteConstants()) {
			//out.println(implementation.valueOf(constant));
			//out.println(implementation.valuations2.get(constant.name()));
			
			B0DataValue value = implementation.valuations2.get(constant.name());
			// check all imported machines (transitively) for such data
			for (Import imp : implementation.imports()) {
				for (ConcreteConstant con : imp.machine().concreteConstants()) {
					if (con.name().equals(constant.name())) {
						out.printf("//included constant %s is in imported machine %s", constant, imp.machine().name());
						continue generateConstants;
					}
				}
			}
			// null value should have imported
			//if (value != null) {
				//out.printf("public static %s %s = %s;\n", javaType(constant.type()), constant.name(), generateDataValue(value));
				System.out.println(constant);
				System.out.println(value);
				String assignment = generateAssigment(constant, value, new StaticCodeGenerator());
				out.printf("public static %s %s;\n", javaType(constant.type()), assignment);
				this.atributes.add(new AtributeSignature(9, javaType(constant.type()), nameManager.moduleData(constant))); // TODo check
			//}
		}
	}
	private void generateVariables(Component component) {
				
		if (component.concreteVariables().size() < 1)
			return;
		out.println(); 
		out.println("// variables of "+component);
		for (ConcreteVariable variable : component.concreteVariables()) {
		
			boolean generateSimpleVariable = true;
			
			if (nameManager.moduleData(variable).contains("."))
			{
				generateSimpleVariable = false;
			}
			
			for (IncludedMachine included : this.implementation.abstractMachine().includedMachines())
			{
				for (ConcreteVariable concreteVariable: included.includedConcreteVariables())
				{
					if (concreteVariable.name().equals(variable.name()))
					{
						generateSimpleVariable = false;
					}	
				}
			}
				
				// TODO DO FOR REFINEMENTS, BUT FIRST MODIFI PARSER ITSELF

			
			
			
			
			
			if (generateSimpleVariable)
			{
			out.printf("private %s %s;\n", javaType(variable.type()), nameManager.moduleData(variable));
			this.atributes.add(new AtributeSignature (2,javaType(variable.type()), nameManager.moduleData(variable)));
			}
			
		
		
		
		
		
		}
		
		
		
		
		/*NATIVE METHOD
		if (component.concreteVariables().size() < 1)
			return;
		out.println(); 
		out.println("// variables of "+component);
		for (ConcreteVariable variable : component.concreteVariables()) {
			out.printf("private %s %s;\n", javaType(variable.type()), nameManager.moduleData(variable));
		}*/
		
	}
	
	private void generateInstruction(Instruction instruction) {
		if (instruction instanceof SequenceInstruction)
			generateInstruction((SequenceInstruction) instruction);
		else if (instruction instanceof BlockInstruction)
			generateInstruction((BlockInstruction) instruction);
		else if (instruction instanceof BecomesEqualInstruction)
			generateInstruction((BecomesEqualInstruction) instruction);
		
		else if (instruction instanceof IfConditionalInstruction)
			generateInstruction((IfConditionalInstruction) instruction);
		else if (instruction instanceof LocalVariableInstruction)
			generateInstruction((LocalVariableInstruction) instruction);
		else if (instruction instanceof CallupInstruction)
			generateInstruction((CallupInstruction) instruction);
		else if (instruction instanceof WhileInstruction)
			generateInstruction((WhileInstruction) instruction);
		else if (instruction instanceof CaseConditionalInstruction)
			generateInstruction((CaseConditionalInstruction) instruction);
		
		else if (instruction instanceof IdentityInstruction)
			;
		else if (instruction instanceof AssertionInstruction)
			;
		
		else throw new RuntimeException("Unsupported instruction type: "+instruction.getClass());
	}
	
	
	private String generateLabel(B0DataValue value) {
		if (value instanceof B0VariableReference) {
			DataField var = ((B0VariableReference) value).variable();
			return generateLabel(var);
		}
		return generateDataValue(value);
	}
	
	private String generateLabel(DataField data) {
		if (data instanceof ForeignVariable) {
			return generateLabel(((ForeignVariable) data).data());
		} else if (data instanceof EnumeratedSet.Item) {
			EnumeratedSet.Item item = (EnumeratedSet.Item) data;
			return nameManager.moduleData(module, item);
		}
		throw new JavaCodeGenerationException("wrong case label", module);
	}
	
	private void generateInstruction(CaseConditionalInstruction instruction) {
		boolean convertToBool = instruction.expression().type().isEqual(BOOL);
		out.printf(convertToBool ? "switch (%s ? 1 : 0) {\n"
		                         : "switch (%s) {\n", generateDataValue(instruction.expression()));
		
		out.indent();
		for (sk.tuke.bkpi.model.instruction.CaseConditionalInstruction.Branch branch : instruction.branches()) {
			for (B0DataValue literal : branch.literalConstants()) {
				out.printf(convertToBool? "case %s ? 1 : 0:\n"
										: "case %s:\n", generateLabel(literal));
			}
			out.indent();
			generateInstruction(branch.substitution());
			out.println("break;");
			out.unindent();
		}
		if (instruction.elseSubstitution() != null) {
			out.println("default:");
			out.indent();
			generateInstruction(instruction.elseSubstitution());
			out.unindent();
		}
		out.unindent();
		out.println("}");
	}

	private void generateInstruction(LocalVariableInstruction instruction) {
		out.println("{");
		out.indent();
		nameManager.openScope();
		for (LocalVariable variable : instruction.localVariables()) {
			// declaration of local variables
			if (!variable.isTyped())
				throw new JavaCodeGenerationException("variable not typed: "+variable, module);
			out.printf("%s %s;\n", javaType(variable.type()), nameManager.moduleData(variable));
		}
		// substitution code
		generateInstruction(instruction.substitution());
		nameManager.closeScope();
		out.unindent();
		out.println("}");
	}
	
	private void generateInstruction(BlockInstruction block) {
		//out.println("{");
		generateInstruction(block.body());
		//out.println("}");
	}
	
	private void generateInstruction(SequenceInstruction sequence) {
		for (Instruction instruction : sequence.instructions()) {
			generateInstruction(instruction);
		}
	}
	
	private void generateInstruction(IfConditionalInstruction instruction) {
		out.printf("if (%s) {\n", generateCondition(instruction.ifBranch().condition()));
		out.indent();
		generateInstruction(instruction.ifBranch().substitution());
		out.unindent();
		out.println("}");
		for (Branch elseif : instruction.elseifBranches()) {
			out.printf("else if (%s) {\n", generateCondition(elseif.condition()));
			out.indent();
			generateInstruction(elseif.substitution());
			out.unindent();
			out.println("}");
		}
		if (instruction.elseBranch() != null) {
			out.println("else {");
			out.indent();
			generateInstruction(instruction.elseBranch());
			out.unindent();
			out.println("}");
		}
	}
	
	private void generateInstruction(BecomesEqualInstruction instruction) {
		out.printf("%s;\n", generateAssigment(instruction.variable(), instruction.value(), new MethodCodeGenerator()));
		if (true) return;
		if (instruction.value() instanceof ArrayExpression) {
			out.println(arrayGenerator.arrayInitialization(instruction.variable(), (ArrayExpression) instruction.value(), null));
		} else {
			
			if (instruction.variable().type() instanceof Record) {
				out.println("//check record type: "+instruction.variable().type()+" data: "+instruction.value().type());
			}
			//out.printf("%s = %s;\n", generateVariable(instruction.variable()), generateDataValue(instruction.value()));
			out.printf("%s;\n", generateAssigment(instruction.variable(), instruction.value(), new MethodCodeGenerator()));
		}
	}
	
	private void generateInstruction(CallupInstruction instruction) {
		ListFormatter argumentsList = new ListFormatter();
		for (B0DataValue argument : instruction.arguments()) {
			argumentsList.addItem(generateDataValue(argument));
		}
		
		String instanceName = instruction.declaration().source() == null? "this" : nameManager.componentReferenceName(module, instruction.declaration().source());
		String call = format("%s.%s(%s)", instanceName, nameManager.get_resolved_operation_name(instruction.declaration().source().machine(), instruction.declaration().operationName()), argumentsList.text());
		//  String call = format("%s.%s(%s)", instanceName, instruction.declaration().operationName(), argumentsList.text());
		if (instruction.assignVariables().size() == 0) {
			out.printf("%s;\n", call);
		} else if (instruction.assignVariables().size() == 1) {
			out.printf("%s = %s;\n", nameManager.moduleData(instruction.assignVariables().get(0)), call);
			//out.printf("%s = %s;\n", instruction.assignVariables().get(0).name(), call); 
		} else {
			LocalVariable result = new LocalVariable(nameManager.generateUniqueName("out"));
			// Note: cannot directly use result.name(), cause it will not register
			// local variable and names collisions can occur later in code
			out.printf("Object[] %s = %s;\n", nameManager.moduleData(result), call);
			int i = 0;
			for (DataField variable : instruction.assignVariables()) {
				out.printf("%s = (%s) %s[%d];\n", nameManager.moduleData(variable), javaType(variable.type()),
						nameManager.moduleData(result), i++);
			}
		}
	}
	
	private void generateInstruction(WhileInstruction instruction) {
		out.printf("while(%s) {\n", generateCondition(instruction.condition()));
		out.indent();
		generateInstruction(instruction.body());
		out.unindent();
		out.println("}");
	}
	
	String generateDataField(DataField data, Module module) {
		//System.out.println("generate data field "+data);
		String prefix = "";
		String name = null;
		if (data instanceof ForeignVariable) { //&& ((ForeignData) data).data() instanceof ConcreteConstant or set
			Module m = project.moduleOfMachine(((ForeignVariable) data).componentReference().machine());
			DataField d = ((ForeignVariable) data).data();
			
			prefix = nameManager.componentReferenceName(((ForeignVariable) data).componentReference())+".";
			name = nameManager.globalModuleData(m, data);
			module = m;
			data = d;
		}
		if (data instanceof ForeignData) {
			module = project.moduleOfMachine(((ForeignData) data).machine());
			DataField d = ((ForeignData) data).data();
			prefix = nameManager.getUniqueMachineName(((ForeignData) data).machine().name())+"."; 
			name = nameManager.globalModuleData(module, data);
			data = d;
		}
		if (data instanceof Item) {
			Item item = (Item) data;
			prefix += nameManager.globalModuleData(module, item.type().set())+".";
			name = nameManager.moduleData(module, item);
		}
		
		if (name == null)
			name = nameManager.moduleData(module, data);

		return prefix+name;
	}
	
	String generateDataField(DataField data) {
		return generateDataField(data, module);
	}
	
	String generateVariable(B0Variable variable) {
		//System.out.println("generate variable: "+variable);
		if (variable instanceof B0VariableReference) {
			DataField data = ((B0VariableReference) variable).variable();
			return generateDataField(data);
		}
		if (variable instanceof RecordFieldVariable) {
			// TODO: move to record generator
			RecordFieldVariable record = (RecordFieldVariable) variable;    //System.out.println("KKT : " + javaType (record.recordVariable().type()));
			return format("%s.%s", generateVariable(record.recordVariable()), nameManager.record_data(javaType (record.recordVariable().type()), record.field())); 
			//return format("%s.%s", generateVariable(record.recordVariable()), record.field()); 
			//return format("%s.%s", generateVariable(((RecordFieldVariable) variable).recordVariable()), ((RecordFieldVariable) variable).field());
		}
		if (variable instanceof ArrayFieldVariable) {
			return arrayGenerator.getElement((ArrayFieldVariable) variable);
		}
		return "Unsupported variable :"+variable + " class: "+variable.getClass();
	}
	
	String generateDataValue(B0DataValue value) {
		return generateDataValue(value.type(), value, null);
	}

	private String generateCondition(Condition condition) {
		if (condition instanceof BinaryCondition)
			return generateCondition((BinaryCondition) condition);
		else if (condition instanceof ComparisonCondition)
			return generateCondition((ComparisonCondition) condition);
		else if (condition instanceof NegationCondition)
			return "!("+generateCondition(((NegationCondition) condition).getOriginalCondition())+")";
		return "unimplemented";
	}
	
	private String generateCondition(BinaryCondition condition) {
		String operator;
		switch (condition.getOperator()) {
			case CONJUNCTION: operator = "&&"; break;
			case DISJUNCTION: operator = "||"; break;
			default: throw new JavaCodeGenerationException("Unknown Binary Condition operator: "+condition.getOperator(), module);
		}
		//return generateCondition(condition.getLeft())+operator+generateCondition(condition.getRight());
		return format("(%s %s %s)", generateCondition(condition.getLeft()), operator, generateCondition(condition.getRight()));
	}
	
	private String generateCondition(ComparisonCondition condition) {
		Operator operator = condition.operator();
		/* transformation to nicer code
		if (condition.right().getType().isEqual(BasicType.BOOL)) {
			if (operator == Operator.EQUAL)
				return generateTerm(condition.left());
			else if (operator == Operator.NOT_EQUAL)
		}
		*/
		String javaOperator;
		switch (operator) {
			case EQUAL: 			javaOperator = "=="; break;
			case NOT_EQUAL: 		javaOperator = "!="; break;
			case LESS: 				javaOperator = "<";  break;
			case LESS_OR_EQUAL: 	javaOperator = "<="; break;
			case GREATER: 			javaOperator = ">";  break;
			case GREATER_OR_EQUAL: 	javaOperator = ">="; break;
			default: throw new JavaCodeGenerationException("Unimplemented Comparison Condition operator: "+operator, module);
		}
		return format("%s %s %s", generateDataValue(condition.left()), javaOperator, generateDataValue(condition.right()));
	}
	
	private TypeResolver INDEXES = new TypeResolver();
	private TypeResolver ELEMENT_TYPE = new TypeResolver();
	private Type ARRAY_TYPE = PowerSet(CartesianProduct(INDEXES, ELEMENT_TYPE));
	
	
	private Map<Type, String> typesAssociationTable = new Hashtable<Type, String>();
	
	String javaType(Type type) {
		for (Type t1 : implementation.typesSubstitutions().keySet()) {
			if (type.contains(t1)) {
				type = type.replace(t1, implementation.typesSubstitutions().get(t1));
			}
		}
		
		if (typesAssociationTable.containsKey(type))
			return typesAssociationTable.get(type);
		
		// or add Basic Types to table ?
		else if (type.isEqual(Z))
			return "Integer";
		else if (type.isEqual(BOOL))
			return "Boolean";
		else if (type.isEqual(STRING))
			return "String";
		
		else if (type instanceof SetType) {
			Set set = ((SetType) type).set();
			if (set instanceof EnumeratedSet) {
				if (includedMapping.containsKey(set)) {
					DataField v = includedMapping.get(set);
					return nameManager.getUniqueMachineName(setsMapping.get(v).machine().name())+"."+nameManager.moduleData(module, v); 
				}
				return nameManager.getUniqueMachineName(setsMapping.get(set).machine().name())+"."+nameManager.moduleData(module, set); 
			}
			return nameManager.moduleData(module, set); //+generateDataField(set);
		}
		
		else if (type instanceof Record)
			return recordGenerator.recordType((Record) type);
		else if (type.isSubtype(ARRAY_TYPE))
			return arrayGenerator.arrayType(type);
		
		if (type.isEqual(PowerSet(Z))) {
			return "Interval";
		}
		throw new JavaCodeGenerationException("Unimplemented Type:"+type);
	}
	
	String generateAssigment(B0Variable variable, B0DataValue value, CodeGenerator codeGenerator) {
		String javaVariable = generateVariable(variable);
		String javaValue = generateDataValue(variable.type(), value, codeGenerator);
		return format("%s = %s", javaVariable, javaValue);
	}
	
	String generateAssigment(DataField data, B0DataValue value, CodeGenerator codeGenerator) {
		String javaValue = generateDataValue(data.type(), value, codeGenerator);
		codeGenerator.close();
		return format("%s = %s", nameManager.moduleData(module, data), javaValue);
	}


	/* ***************************************
	 * ****** generation of expressions ******
	 * ***************************************/
	public String generateDataValue(Type type, B0DataValue value, CodeGenerator codeGenerator) {
		try {
			Method m = JavaGenerator.class.getMethod("generateDataValue", new Class[]{Type.class, value.getClass(), CodeGenerator.class});
			return (String) m.invoke(this, type, value, codeGenerator);
		} catch (Exception e) {
			try {
				System.out.println("-------\ngenerateDataValue");
				System.out.println("Value: "+value);
				System.out.println("Class: "+value.getClass());
				System.out.println("SuperClass: "+value.getClass().getSuperclass());
				System.out.println("-------");
				Method m = JavaGenerator.class.getMethod("generateDataValue", new Class[]{Type.class, value.getClass().getSuperclass(), CodeGenerator.class});
				System.out.println("after JavaGenerator.class.getMethod");
				return (String) m.invoke(this, type, value, codeGenerator);
			} catch (Exception ex) {
				ex.printStackTrace();
				throw new JavaCodeGenerationException(ex.getMessage());
			}
		}
	}
	
	public String generateDataValue(Type type, BooleanTerm value, CodeGenerator codeGenerator) {
		return value.value()? "true" : "false";
	}
	
	public String generateDataValue(Type type, IntegerTerm value, CodeGenerator codeGenerator) {
		switch(value.value()) {
			case Integer.MAX_VALUE: return "Integer.MAX_VALUE";
			case Integer.MIN_VALUE: return "Integer.MIN_VALUE";
			default: return value.value().toString();
		}
	}
	
	public String generateDataValue(Type type, NegativeTerm value, CodeGenerator codeGenerator) {
		String javaValue = generateDataValue(value.expression());
		return value.expression() instanceof ArithmeticalBinaryExpression? "-("+javaValue+")" : "-"+javaValue;
	}
	
	public String generateDataValue(Type type, RecordTerm record, CodeGenerator codeGenerator) {
		Record recordType = (Record) type;
		ListFormatter fields = new ListFormatter();
		for (RecordTerm.RecordField field : record.fields()) {
			Type fieldType = recordType.fieldType(field.label());
			fields.addItem(generateDataValue(fieldType, field.value(), codeGenerator));
		}
		return format("new %s(%s)", recordGenerator.recordType(recordType), fields.text());
	}
	
	public String generateDataValue(Type type, ArrayExpression value, CodeGenerator codeGenerator) {
		return arrayGenerator.arrayExpression(value, codeGenerator);
	}
	
	public String generateDataValue(Type type, ArrayFieldVariable value, CodeGenerator codeGenerator) {
		return arrayGenerator.getElement(value);
	}
	
	public String generateDataValue(Type type, B0Variable value, CodeGenerator codeGenerator) {
		return generateVariable(value);
	}
	
	public String generateDataValue(Type type, BoolConversionValue value, CodeGenerator codeGenerator) {
		return generateCondition(value.predicate());
	}
	
	public String generateDataValue(Type type, StringValue string, CodeGenerator codeGenerator) {
		return "\""+string.value()+"\"";
	}
	
	public String generateDataValue(Type type, ArithmeticalBinaryExpression expr, CodeGenerator codeGenerator) {
		String operator;
		B0DataValue left = expr.left();
		B0DataValue right = expr.right();
		
		switch (expr.operator()) {
			case PLUS: operator = "+";break;
			case MINUS:
				operator = "-";
				if (right instanceof NegativeTerm) {
					//return generateDataValue(expr.left())+"+"+generateDataValue(((NegativeTerm) expr.right()).expression());
					
					right = negativeTerm(right);
					if (negativeTermsCount(expr.right()) % 2 == 1) {
						right = ((NegativeTerm) right).expression();
						operator = "+";
					}
				}
				
				break;
			case TIMES: operator = "*";break;
			case DIVIDE: operator = "/";break;
			case MODULO: operator = "%";break;
			case POWER: return format("Math.pow(%s, %s)", expr.left(), expr.right());
			default: throw new JavaCodeGenerationException("Unsuported Arithmetical Binary operator: "+expr.operator(), module);
		}
		//String format = expr.right() instanceof NegativeTerm? "" : "";
		String format;
		
		format = left instanceof ArithmeticalBinaryExpression &&
			((ArithmeticalBinaryExpression) left).operator().priority>expr.operator().priority? "(%s)" : "%s";
		format += operator;
		format += right instanceof ArithmeticalBinaryExpression &&
			(expr.operator() == ArithmeticalBinaryExpression.Operator.MINUS
				|| ((ArithmeticalBinaryExpression) right).operator().priority>expr.operator().priority)? "(%s)" : "%s";
		
		//return format("(%s)%s(%s)", generateDataValue(left), operator, generateDataValue(right)); // safe way
		return format(format, generateDataValue(left), generateDataValue(right));
		//return "("+generateDataValue(left)+operator+generateDataValue(right)+")";
	}
	
	public String generateDataValue(Type type, IntervalB0 interval, CodeGenerator codeGenerator) {
		return format("new Interval(%s, %s)", generateDataValue(interval.minimum()), generateDataValue(interval.maximum()));
		
	}

	public B0DataValue negativeTerm2(B0DataValue term) {
		int count = 0;
		B0DataValue x = term;
		while(term instanceof NegativeTerm) {
			count++;
			term = ((NegativeTerm) term).expression();
		}
		return count % 2 == 0? term : new NegativeTerm(term);
	}
	
	public B0DataValue negativeTerm(B0DataValue term) {
		int count = 0;
		B0DataValue x = term;
		while(term instanceof NegativeTerm) {
			count++;
			if (count % 2 == 1) {
				x = term;
			}
			term = ((NegativeTerm) term).expression();
		}
		return count % 2 == 0? term : x;
	}
	
	public int negativeTermsCount(B0DataValue term) {
		int count = 0;
		while(term instanceof NegativeTerm) {
			count++;
			term = ((NegativeTerm) term).expression();
		}
		return count;
	}
	
	public abstract class CodeGenerator {
		protected int scope;
		
		public void openScope() {
			scope++;
			out.print(" {\n");
			out.indent();
		}
		
		void closeScope() {
			if (scope > 0) {
				scope--;
				out.unindent();
				out.println("}");
			} else {
				throw new IllegalStateException();
			}
		}
		
		void close() {}
		
		abstract void addDeclaration(String declaration);
		abstract void addStatement(String statement);
	}
	
	public class MethodCodeGenerator extends CodeGenerator {

		@Override
		void addDeclaration(String declaration) {
			out.printf(declaration);
		}

		@Override
		void addStatement(String statement) {
			out.printf(statement);
		}		
	}
	
	public class StaticCodeGenerator extends CodeGenerator {
		
		boolean declarationMode = true;
		
		public void addDeclaration(String declaration) {
			if (scope == 0) {
				if (!declarationMode) {
					out.println("}");
					declarationMode = true;
				}
				out.println("private static "+declaration);
			} else {
				out.println(declaration);
			}
		}
		
		public void close() {
			if (!declarationMode) {
				out.unindent();
				out.println("}");
			}
		}
		
		void addStatement(String statement) {
			if (scope == 0) {
				if (declarationMode) {
					out.println("static {");
					out.indent();
					declarationMode = false;
				}
				out.printf(statement);
			} else {
				out.printf(statement);
			}
		}
	}
}
