package sk.tuke.bkpi.generator;

import java.io.PrintWriter;
import java.io.Writer;

/**
 * Writer implementation that implements support for indentation.
 * Every <code>println</code> method as well as <code>printf</code>
 * method will be indented.
 * 
 * @author Marcel Dancak
 *
 */
public class JavaWriter extends PrintWriter {
	private int indent;
	
	public JavaWriter(Writer writer) {
		super(writer);
	}
	
	@Override
	public PrintWriter printf(String format, Object... args) {
		doIndent();
		return super.printf(format, args);
	}
	
	@Override
	public void println(String x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(boolean x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(double x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(int x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(float x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(long x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(char x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(char[] x) {
		doIndent();
		super.println(x);
	}
	
	@Override
	public void println(Object x) {
		doIndent();
		super.println(x);
	}

	public void indent() {
		indent++;
	}
	public void unindent() {
		if (indent > 0)
			indent--;
	}
	
	public void doIndent() {
		for (int i = 0; i < indent; i++) {
			super.print("    ");
		}
	}
	
	public static String indent(int count) {
		String indent = "";
		for (int i = 0; i < count; i++) {
			indent += "    ";
		}
		return indent;
	}
}
