package sk.tuke.bkpi.generator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.OperationB0;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;

public class OperationsNameManager {
	
	private Collection<String> excluded_words = Collections.emptyList();
	
	public void setExcluded_words(String ... excluded_words) {
		this.excluded_words = Arrays.asList(excluded_words);
	}
	
	
	public void setExcluded_words(Collection<String> excluded_words) {
		this.excluded_words = excluded_words;
	}
	
	private Map <String, Hashtable <String, String>> operation_names = new Hashtable <String, Hashtable <String, String>>();
	
	private Map <String, Hashtable <String, String>> other_generated_methods = new Hashtable <String, Hashtable <String, String>>(); 
	
	private Map <String, List <String>> implementations_per_machine = new Hashtable <String, List<String>>();

	public Map<String, List<String>> getImplementations_per_machine() {
		return implementations_per_machine;
	}


	public void generate_operation_names (Project project)
	{
		for (Implementation implementation : project.implementations())
		{
			
			/*Insert implementation name in the list of all implementations per abstraction BEGIN */
			
			if (this.implementations_per_machine.containsKey(implementation.abstractMachine().name()))
			{
				if ((!this.implementations_per_machine.get(implementation.abstractMachine().name()).contains(implementation.name())) && ((implementation.getOperations().size() > 0) || (implementation.promotedOperations().size() > 0)))
						{
					     this.implementations_per_machine.get(implementation.abstractMachine().name()).add(implementation.name());
						}
			}
			else
			{
				if (((implementation.getOperations().size() > 0) || (implementation.promotedOperations().size() > 0)))
				{
				this.implementations_per_machine.put(implementation.abstractMachine().name(), new ArrayList <String>());
				this.implementations_per_machine.get(implementation.abstractMachine().name()).add(implementation.name());
				}
			}
			
			/*Insert implementation name in the list of all implementations per abstraction END */
		
		}
		
		for (Implementation implementation : project.implementations())
		{

			/*Create operations hash table per implementation BEGIN */
			
			if (((implementation.getOperations().size() > 0) || (implementation.promotedOperations().size() > 0)) && (!this.operation_names.containsKey(implementation.name())))
			{
				this.operation_names.put(implementation.name(), new Hashtable<String, String>());
			}
			
			/*Create operations hash table per implementation END */			
			
			/*Resolve local operation names and abstraction operation names per implementation BEGIN */
			
			for (OperationB0 operationB0 : implementation.getOperations())
			{
				if (!this.operation_names.get(implementation.name()).containsKey(operationB0.name()))
				{
					
				String proposal = new String (operationB0.name());
				while ((this.excluded_words.contains(proposal)) || (this.operation_names.get(implementation.name()).containsValue(proposal)))
				{
					proposal = proposal + "_op";
				}
				
				//this.operation_names.get(implementation.name()).put(operationB0.name(), proposal);
				
				for (Implementation Count_implementation: project.implementations())
				{
                   if (this.implementations_per_machine.get(implementation.abstractMachine().name()).contains(Count_implementation.name()))
                   {
                	   
                	   if (Count_implementation.getOperations().contains(operationB0))
                	   {
                		   
               			if ((!this.operation_names.containsKey(Count_implementation.name())))
            			{
            				this.operation_names.put(Count_implementation.name(), new Hashtable<String, String>());
            			}
               			
               			this.operation_names.get(Count_implementation.name()).put(operationB0.name(), proposal);
                	
                	   }
                	   else
                	   {
                		   boolean contains_as_promoted_operation = false;
                		   for (OperationDeclaration promoted_operation: Count_implementation.promotedOperations())
                		   {
                			   if (promoted_operation.operationName().equals(operationB0.name()))
                			   {
                				   contains_as_promoted_operation = true;
                			   }
                			   
                		   }
                		   if (contains_as_promoted_operation == true)
                		   {
                      			if ((!this.operation_names.containsKey(Count_implementation.name())))
                    			{
                    				this.operation_names.put(Count_implementation.name(), new Hashtable<String, String>());
                    			}
                       			
                       			this.operation_names.get(Count_implementation.name()).put(operationB0.name(), proposal);
                		   }   
                	   } 
                   }
					
				}
				 }
			} 

			/*Resolve local operation names and abstraction operation names per implementation END */
			
			/*Resolve promoted operation names per implementation BEGIN */
			
		    for (OperationDeclaration operation_declaration : implementation.promotedOperations())
		    {
				if (!this.operation_names.get(implementation.name()).containsKey(operation_declaration.operationName()))
				{
		    	String proposal = new String (operation_declaration.operationName());	
		    	while ((this.excluded_words.contains(proposal)) || (this.operation_names.get(implementation.name()).containsValue(proposal)))
		    	{
		    		proposal = proposal + "_op";
		    	}
		    	//this.operation_names.get(implementation.name()).put(operation_declaration.operationName(), proposal);
		    	
				for (Implementation Count_implementation: project.implementations())
				{
                   if (this.implementations_per_machine.get(implementation.abstractMachine().name()).contains(Count_implementation.name()))
                   {
                	   
                	   if (Count_implementation.promotedOperations().contains(operation_declaration))
                	   {
                		   if (!this.operation_names.containsKey(Count_implementation.name()))
                		   {
                			   this.operation_names.put(Count_implementation.name(), new Hashtable<String, String>());   
                		   }
                		   
                		   this.operation_names.get(Count_implementation.name()).put(operation_declaration.operationName(), proposal);
                		   
                		   
                	   }
                	   else
                	   {
                		   
                		   boolean contains_as_operation_of_abstraction = false;
                		   for (OperationB0 operation: Count_implementation.getOperations())
                		   {
                			   if (operation.name().equals(operation_declaration.operationName()))
                			   {
                				   contains_as_operation_of_abstraction = true;
                			   }
                		   }
                		   
                		   if (contains_as_operation_of_abstraction == true)
                		   {
                			   
                    		   if (!this.operation_names.containsKey(Count_implementation.name()))
                    		   {
                    			   this.operation_names.put(Count_implementation.name(), new Hashtable<String, String>());   
                    		   }
                    		   
                    		   this.operation_names.get(Count_implementation.name()).put(operation_declaration.operationName(), proposal);                			   
                			   
                		   }
                		   
                	   }
                	   
                   }
				}

				}
				
		    }
			
			/*Resolve promoted operation names per implementation END */			
				
		}
		
		/* PRINT RESOLVED OPERATION NAMES FOR ALL IMPLEMENTATIONS
		 
		for (String machine :this.implementations_per_machine.keySet())
		{
			
			System.out.println("Generating operation names for machine: " + machine);
			
			for (String implementation: this.implementations_per_machine.get(machine))
			{
				System.out.println("Generationg operation names for implementation: " + implementation);
				
				for (String operation: this.operation_names.get(implementation).keySet())
				{
				
					System.out.println ("Operation name: " + operation +" Resolved operation name: " + this.operation_names.get(implementation).get(operation));
					
				}
			}
		}
		*/

	}
	
	public String get_resolved_operation_name (Implementation implementation, String operation_name)
	{
		String result = new String();
		
		if (this.operation_names.containsKey(implementation.name()))
		{
			if (this.operation_names.get(implementation.name()).containsKey(operation_name))
			{
			 result = this.operation_names.get(implementation.name()).get(operation_name);	
			}
			else
			{   //result = null;
                return generate_unique_method_name(implementation, operation_name);				
				
			}
		}
		else
		{
			return generate_unique_method_name(implementation, operation_name);	
			//result = null;
		}

		return result;
	}
	
	public String get_resolved_operation_name (Machine machine, String operation_name)
	{
		String result = new String();
		
		if (this.implementations_per_machine.containsKey(machine.name())){
			
			if (this.operation_names.get(this.implementations_per_machine.get(machine.name()).get(0)).containsKey(operation_name))
			{
				result = operation_names.get(this.implementations_per_machine.get(machine.name()).get(0)).get(operation_name);
			}
			else
			{
				return generate_unique_method_name(machine, operation_name);	
				//result = null;
			}
		}
		else
		{
			return generate_unique_method_name(machine, operation_name);
			//result = null;
		}

		return result;
	}
	
	public String generate_unique_method_name (Machine machine, String proposal)
	{
        String result = new String();
        
        if (!this.other_generated_methods.containsKey(machine.name()))
        {
        	this.other_generated_methods.put(machine.name(), new Hashtable <String, String>());
        }
        
        if (this.other_generated_methods.get(machine.name()).containsKey(proposal))
        {
        	result = this.other_generated_methods.get(machine.name()).get(proposal);
        }
        else
        {
        	result = proposal;
        	boolean resolution_required;
        	do
        	{
        		resolution_required = false;
        		if (this.implementations_per_machine.containsKey(machine.name()))
        		{
        		for (String Cimplementation: this.implementations_per_machine.get(machine.name()))
        		{
        			if (this.operation_names.get(Cimplementation).containsValue(result))
        			{
        				resolution_required = true;
        			}
        		}
        		}
        		
        		// BEGNA
        		if (this.excluded_words.contains(result))
        		{
        			resolution_required = true;
        		}
        		//ENDA
        		
        		if (resolution_required == true)
        		{
        			result = result + "_met";
        		}
        		
        	} while (resolution_required == true);

        }

        this.other_generated_methods.get(machine.name()).put(proposal, result);		
		return result;
	}
	
	public String generate_unique_method_name (Implementation implementation, String proposal)
	{
		String result = new String();
		
        if (!this.other_generated_methods.containsKey(implementation.abstractMachine().name()))
        {
        	this.other_generated_methods.put(implementation.abstractMachine().name(), new Hashtable <String, String>());
        }
        
        if (this.other_generated_methods.get(implementation.abstractMachine().name()).containsKey(proposal))
        {
        	result = this.other_generated_methods.get(implementation.abstractMachine().name()).get(proposal);
        }
        else
        {
        	
        	result = proposal;
        	boolean resolution_required;
        	do
        	{
        		resolution_required = false;
        		
        		// BEGNA
        		if (this.implementations_per_machine.containsKey(implementation.abstractMachine().name()))
        		{//ENDA
        			
        		
        		for (String Cimplementation: this.implementations_per_machine.get(implementation.abstractMachine().name()))
        		{
        			if (this.operation_names.get(Cimplementation).containsValue(result))
        			{
        				resolution_required = true;
        			}
        		}
        		// BEGNA
        		}
        		//ENDA
        		
        		// BEGNA
        		if (this.excluded_words.contains(result))
        		{
        			resolution_required = true;
        		}
        		//ENDA
        		
        		if (resolution_required == true)
        		{
        			result = result + "_met";
        		}
        		
        	} while (resolution_required == true);

        }

        this.other_generated_methods.get(implementation.abstractMachine().name()).put(proposal, result);		
		return result;
	}
	
	
	
		
	
	
	
	
	
	
	
	

}
