package sk.tuke.bkpi.generator;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Logger;

import sk.tuke.bkpi.generator.Utilities.ListFormatter;
import sk.tuke.bkpi.model.types.Record;
import sk.tuke.bkpi.model.types.Type;

/**
 * Translates B-Method record expressions to their corresponding
 * new Java classes in separated files. It provides their names
 * generation and mapping from B-Method record types to these
 * Java classes.
 * 
 * @author Marcel Dancak
 */
public class RecordGenerator1 implements RecordGenerator {

	private int recordsCounter = 0;
	private Map<Record, String> records = new Hashtable<Record, String>();
	
	private JavaGenerator generator;
	
	public RecordGenerator1(JavaGenerator generator) {
		this.generator = generator;
	}
	
	@Override
	public String recordType(Record type) {
		System.out.println("generate record: "+type);
		System.out.println("Record Type Cache: "+records);
		for (Type t : records.keySet()) {
			//if (t.isEqual(type)) {
			if (t.isCompatibile(type)) {
				System.out.println("Record Type for: "+type+" from cache: "+records.get(t));
				return records.get(t);
			}
		}
		// recursively generate new Record types for all fields, that
		// contain another Records
		for (Record.Field f : type.fields()) {
			generator.javaType(f.type); 
		}
		
		String recordName = generator.getNameManager().generateUniqueClassName("Record"+recordsCounter);
		records.put(type, recordName);
		generateRecordClass(type, recordName);
		System.out.println("Record Type for: "+type + " "+"new Record type: "+recordName);
		
		recordsCounter++;
		return recordName;
	}
	
	private void generateRecordClass(Record record, String name) {
		JavaWriter out = null;
		try {
			out = new JavaWriter(generator.newFileStream(name));
			out.printf("public class %s {\n", records.get(record));
			out.indent();
			
			ListFormatter arguments = new ListFormatter();
			for (Record.Field field : record.fields()) {
				String javaType = generator.javaType(field.type);
				String fieldName = field.label;
				
				if (fieldName == null)
					throw new NullPointerException();
				out.printf("public %s %s;\n", javaType, generator.getNameManager().record_data(records.get(record), fieldName));
				//out.printf("public %s %s;\n", javaType, fieldName);
				arguments.addItem(javaType+" "+generator.getNameManager().record_data(records.get(record), fieldName));
				//arguments.addItem(javaType+" "+fieldName);
			}
			
			// Record constructor
			out.println();
			out.printf("public %s(%s) {\n", records.get(record), arguments.text());
			out.indent();
			for (Record.Field field : record.fields()) {
				out.printf("this.%s = %s;\n", generator.getNameManager().record_data(records.get(record),field.label), generator.getNameManager().record_data(records.get(record),field.label)); 
				//out.printf("this.%s = %s;\n", field.label, field.label);
			}
			out.unindent();
			out.println("}");
			
			out.unindent();
			out.println("}");
		} catch (IOException e) {
			Logger.getLogger(RecordGenerator1.class.getName()).severe("Cannot generate Record class: "+record);
			e.printStackTrace();
		} finally {
			if (out != null) {
				out.close();
			}
		}
	}
	
}
