package sk.tuke.bkpi.generator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Set of utilities that simplify translation B-Method source files
 * to the target language.
 * 
 * @author Marcel Dancak
 */
public class Utilities {
	
	@SuppressWarnings("unchecked")
	public static <T, E> List<E> filter(List<T> list, Class<E> clazz) {
		List<E> ret = new ArrayList<E>();
		for (T o : list) {
			if (clazz.isInstance(o)) {
				ret.add((E) o);
			}
		}
		return ret;
	}
	
	public static String repeat(String string, int count) {
		StringBuilder buffer = new StringBuilder();
		for (int i = 0; i < count; i++)
			buffer.append(string);
		return buffer.toString();
	}
	
	public static String listOfStrings(List<String> list) {
		StringBuilder oneString = new StringBuilder();
		Iterator<String> it = list.iterator();
		while (it.hasNext()) {
			oneString.append(String.format("%s%s", it.next(), it.hasNext()? ", " : ""));
		}
		return oneString.toString();
	}
	
	//Formats text items by adding specified separator between them to build final text.
	/**
	 * Formats text items to the one text containing these items
	 * separated by specified separator.
	 * 
	 * @author Marcel Dancak 
	 */
	public static class ListFormatter {
		
		private StringBuilder oneString = new StringBuilder();
		private String separator = ", ";
		private int itemsCount = 0;
		
		public ListFormatter() {}
		
		public ListFormatter(String separator) {
			this.separator = separator;
		}
		
		public void addItem(String item) {
			if (itemsCount > 0)
				oneString.append(separator);
			oneString.append(item);
			itemsCount++;
		}
		
		public int itemsCount() {
			return itemsCount;
		}
		
		public String text() {
			return oneString.toString();
		}
	}
}
