package sk.tuke.bkpi.gui;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Map;

import sk.tuke.bkpi.model.Project;

/*
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.exception.MethodInvocationException;
import org.apache.velocity.exception.ParseErrorException;
import org.apache.velocity.exception.ResourceNotFoundException;
*/
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;

/**
 * Generates project graph image by using FreeMarker template engine library.
 * From project model and template file is generated graph description file
 * in dot language for Graphviz library. This file is used to generate image file
 * by execution dot program from Graphviz library. Graphic format is derived
 * from output file's extension.
 * 
 * @author Marcel Dancak
 */
public class GraphvizVisualization {
	
	private Project project;
    private File templateFile;
	private File outputFile;
	private String extension = "png";
	private String dotFileName;
	
	
	public GraphvizVisualization(Project project, File template, File outputFile) {
		this.project = project;
        this.templateFile = template;
		this.outputFile = outputFile;
		
		if (outputFile.getName().lastIndexOf(".") > 0) {
			int lastDot = outputFile.getAbsolutePath().lastIndexOf(".");
			extension = outputFile.getAbsolutePath().substring(lastDot+1);
			dotFileName = outputFile.getAbsolutePath().substring(0, lastDot)+".dot";
		} else {
			dotFileName = outputFile.getAbsolutePath()+".dot";
		}
	}
	
	/**
	 * Generates graph description file with ".dot" extension and image file.
	 * @throws IOException
	 * @throws TemplateException
	 */
	public void generateGraph() throws IOException, TemplateException {
		//createGraph(graph2);
		//createGraphVelocity();
		createGraphFreeMarker();
	}
	
	private void createGraphFreeMarker() throws IOException, TemplateException {
		
		Configuration config = new Configuration();
        config.setDirectoryForTemplateLoading(templateFile.getAbsoluteFile().getParentFile());
		//config.setObjectWrapper(new DefaultObjectWrapper());
		Map<String, Object> context = new Hashtable<String, Object>();
		context.put("project", project);
		//context.put("nodes", project.rootModulesGraphs().values());
		
		FileWriter dotFile = null;
		try {
			dotFile = new FileWriter(dotFileName);

			Template template = config.getTemplate("freemarker.template");
			template.process(context, dotFile);
			
		} finally {
			if (dotFile != null) {
				try {
					dotFile.close();
				} catch (IOException e) {} // at least we tried
			}
		}
		executeDotCommand();
	}
	/*
	private void createGraphVelocity() {
		VelocityContext context = new VelocityContext();
		context.put("project", project);
		
		FileInputStream template = null;
		FileWriter dotFile = null;
		try {
			//template = new FileInputStream("project_graph.template");
			dotFile = new FileWriter(dotFileName);
			Velocity.mergeTemplate("velocity_graph.template", "utf-8", context, dotFile);
			//Velocity.evaluate(context, dotFile, "", template);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			if (template != null) {
				try {
					template.close();
				} catch (Exception e) {}
			}
			if (dotFile != null) {
				try {
					dotFile.close();
				} catch (IOException e) {}
			}
		}
		executeDotCommand();
	}
	*/
	
	private void executeDotCommand() {
		System.out.println(String.format("dot -T%s \"%s\"", extension, dotFileName));
		try {
			Process p = Runtime.getRuntime().exec(String.format("dot -T%s \"%s\"", extension, dotFileName));
			InputStream input = p.getInputStream();
			
			FileOutputStream graphWriter = new FileOutputStream(outputFile);
			int c;
		    while ((c = input.read()) != -1) {
		    	graphWriter.write(c);
		    }
		    input.close();
		    graphWriter.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
