/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package sk.tuke.bkpi.gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextPane;
import javax.swing.text.DefaultCaret;
import javax.swing.text.Element;
import javax.swing.text.html.HTMLDocument;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import sk.tuke.bkpi.complexGui.MainWindow;

/**
 * Swing component for displaying text in HTML format with support for text styles.
 * 
 * @author Marcel Dancak
 */
public class HtmlConsole extends JTextPane{

	private static final long serialVersionUID = 1L;
	private JPopupMenu consoleMenu = new JPopupMenu();
    private JMenuItem clear = new JMenuItem("clear");
    private boolean cleared = true;
    
    public HtmlConsole() {
        consoleMenu.add(clear);
        setComponentPopupMenu(consoleMenu);
        clear.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent evt) {
                clear();
            }
        });

        setEditable(false);
        setCaret(new DefaultCaret());
        setCaretPosition(0);
        setFont(new java.awt.Font("Tahoma", 0, 11));
        setContentType("text/html");
        setText("<html> <body> </body> </html>");
    }

    /**
     * Clears component's content (display area).
     */
    public void clear() {
        setText("<html> <body> </body> </html>");
        cleared = true;
    }

    public boolean isCleared() {
    	return cleared;
    }
    
    /**
     * Encodes <code>text</code> to HTML format and appends it at the end
     * of component's content.
     * @param text input text
     */
    public void write(final String text) {
        writeHTML(encodeHTML(text));
    }

    /**
     * Appends text that should already be in HTML format and appends it
     * at the end of component's content.
     * @param text in HTML format
     */
    public void writeHTML(final String text) {
    	cleared = false;
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
        		HTMLDocument doc = (HTMLDocument) getDocument();
                
                Element root = doc.getDefaultRootElement();
                Element body = root.getElement(1);
                Element p = body.getElement(0);
                try {   
                    doc.insertBeforeEnd(p, text);
                } catch (Exception ex) {
                	//Logger.getLogger(SimpleGui.class.getName()).log(Level.ERROR, null, ex);
                	Logger.getLogger(MainWindow.class.getName()).log(Level.ERROR, null, ex);
                	
                }
            }
        });
    }

    /**
     * Encodes <code>text</code> to HTML format
     * @param text input
     * @return text encoded in HTML format
     */
    public static String encodeHTML(String text) {
        StringBuffer sb = new StringBuffer();
        int len = text.length();
        char c;

        for (int i = 0; i < len; i++) {
            c = text.charAt(i);

            switch (c) {
            case '<': sb.append("&lt;"); break;
            case '>': sb.append("&gt;"); break;
            case '&': sb.append("&amp;"); break;
            case '"': sb.append("&quot;"); break;
			 case '\u00e0': sb.append("&agrave;");break;
			 //case 'à': sb.append("&agrave;");break;
			 
            case '\u00c0': sb.append("&Agrave;");break;
            //case 'À': sb.append("&Agrave;");break;
			 
            case '\u00e2': sb.append("&acirc;");break;
            //case 'â': sb.append("&acirc;");break;				 
			 
            case '\u00c2': sb.append("&Acirc;");break;
            //case 'Â': sb.append("&Acirc;");break;
			 
            case '\u00e4': sb.append("&auml;");break;
            //case 'ä': sb.append("&auml;");break;				 
			 
            case '\u00c4': sb.append("&Auml;");break;
            //case 'Ä': sb.append("&Auml;");break;				 
			 
            case '\u00e5': sb.append("&aring;");break;
            //case 'å': sb.append("&aring;");break;				 
			 
            case '\u00c5': sb.append("&Aring;");break;
            //case 'Å': sb.append("&Aring;");break;				 
			 
            case '\u00e6': sb.append("&aelig;");break;
            //case 'æ': sb.append("&aelig;");break;				 
			 
            case '\u00c6': sb.append("&AElig;");break;
            //case 'Æ': sb.append("&AElig;");break;				 
			 
            case '\u00e7': sb.append("&ccedil;");break;
            //case 'ç': sb.append("&ccedil;");break;				 
			 
            case '\u00c7': sb.append("&Ccedil;");break;
            //case 'Ç': sb.append("&Ccedil;");break;				 
			 
            case '\u00e9': sb.append("&eacute;");break;
            //case 'é': sb.append("&eacute;");break;				 
			 
            case '\u00c9': sb.append("&Eacute;");break;
            //case 'É': sb.append("&Eacute;");break;				 
			 
            case '\u00e8': sb.append("&egrave;");break;
            //case 'è': sb.append("&egrave;");break;				 
			 
            case '\u00c8': sb.append("&Egrave;");break;
            //case 'È': sb.append("&Egrave;");break;				 
			 
            case '\u00ea': sb.append("&ecirc;");break;
            //case 'ê': sb.append("&ecirc;");break;				 
			 
            case '\u00ca': sb.append("&Ecirc;");break;
            //case 'Ê': sb.append("&Ecirc;");break;				 
			 
            case '\u00eb': sb.append("&euml;");break;
            //case 'ë': sb.append("&euml;");break;				 
			 
            case '\u00cb': sb.append("&Euml;");break;
            //case 'Ë': sb.append("&Euml;");break;				 
			 
            case '\u00ef': sb.append("&iuml;");break;
            //case 'ï': sb.append("&iuml;");break;				 
			 
            case '\u00cf': sb.append("&Iuml;");break;
            //case 'Ï': sb.append("&Iuml;");break;				 
			 
            case '\u00f4': sb.append("&ocirc;");break;
            //case 'ô': sb.append("&ocirc;");break;				 
			 
            case '\u00d4': sb.append("&Ocirc;");break;
            //case 'Ô': sb.append("&Ocirc;");break;				 
			 
            case '\u00f6': sb.append("&ouml;");break;
            //case 'ö': sb.append("&ouml;");break;				 
			 
            case '\u00d6': sb.append("&Ouml;");break;
            //case 'Ö': sb.append("&Ouml;");break;				 
			 
            case '\u00f8': sb.append("&oslash;");break;
            //case 'ø': sb.append("&oslash;");break;				 
			 
            case '\u00d8': sb.append("&Oslash;");break;
            //case 'Ø': sb.append("&Oslash;");break;				 
			 
            case '\u00df': sb.append("&szlig;");break;
            //case 'ß': sb.append("&szlig;");break;				 
			  
            case '\u00f9': sb.append("&ugrave;");break;
            //case 'ù': sb.append("&ugrave;");break;				 
			 
            case '\u00d9': sb.append("&Ugrave;");break;
            //case 'Ù': sb.append("&Ugrave;");break;				 
			 
            case '\u00fb': sb.append("&ucirc;");break;
            //case 'û': sb.append("&ucirc;");break;				 
			 
			 case '\u00db': sb.append("&Ucirc;");break;
            //case 'Û': sb.append("&Ucirc;");break;
			 				 
            case '\u00fc': sb.append("&uuml;");break;
            //case 'ü': sb.append("&uuml;");break;				 
			 
            case '\u00dc': sb.append("&Uuml;");break;
            //case 'Ü': sb.append("&Uuml;");break;				 
			 
            case '\u00ae': sb.append("&reg;");break;
            //case '®': sb.append("&reg;");break;				 
			 
            case '\u00a9': sb.append("&copy;");break;
            //case '©': sb.append("&copy;");break;				 
			 
            case '\u0080': sb.append("&euro;"); break;
                 // be careful with this one (non-breaking white space)
                 case ' ': sb.append("&nbsp;");break;
                 case '\n': sb.append("<br/>");break;

                 //default:  sb.append(c); break;
                default:
                    int ci = 0xffff & c;
                    if (ci < 160) // nothing special only 7 Bit
                    {
                        sb.append(c);
                    } else {
                        // Not 7 Bit use the unicode system
                        sb.append("&#");
                        sb.append(new Integer(ci).toString());
                        sb.append(';');
                    }
              }
        }
        return sb.toString();
    }
    
    private void htmlStructure() {
        HTMLDocument doc = (HTMLDocument) getDocument();
        elementStructure(doc.getDefaultRootElement(), "");
        System.out.println("___________");
    }
    
    private void elementStructure(Element e, String indent) {
        System.out.print(indent+e);
        for (int i = 0; i < e.getElementCount(); i++) {
            elementStructure(e.getElement(i), indent+" ");
        }
    }
}
