package sk.tuke.bkpi.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.component.Component;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.component.Refinement;
import sk.tuke.bkpi.parser.Couple;

/**
 * B-Method project module, collecting all components from abstract machine
 * specification to its last refinement (possibly implementation). It also 
 * contains information about dependencies between modules.
 * 
 * @author Marcel Dancak
 */
public class Module {

	private Machine machine;
	private List<Refinement> refinements = Collections.emptyList();
	private Implementation implementation;
	// machine, refinements and implementation in one list
	private List<Component> components;
	
	private List<Module> importedModules;
	
	private List<SeenMachine> seesReferencesDelegations = new ArrayList<SeenMachine>(2);
	private List<Couple<Machine, Machine>> seesDependencies = new ArrayList<Couple<Machine, Machine>>();
	private Map<SeenMachine, List<Import>> map = new Hashtable<SeenMachine, List<Import>>();
	
	public Module(Machine machine, List<Refinement> refinements, Implementation implementation) {
		this(machine, refinements);
		
		if (implementation != null) { 
			
            // TU SA SHIT ISTOTNE VETVI`
			/*
			if (implementation.name().equals("abstraction_i"))
			{*/
				/*System.out.print("Blu:" + implementation.name());*/
				components.add(implementation);
				this.implementation = implementation;
				System.out.println("jj : "+ this.implementation.name());
			/*}*/
			
			
		}
	}
	
	public Module(Machine machine, List<Refinement> refinements) {
		this(machine);
		this.refinements = refinements;
		components.addAll(refinements);
	}
	
	public Module(Machine machine) {
		this.machine = machine;
		components = new ArrayList<Component>();
		components.add(machine);
	}
	
	public Machine machine() {
		return machine;
	}
	
	public Implementation implementation() {
		return implementation;
	}
	
	public List<Refinement> refinements() {
		return refinements;
	}
	
	/**
	 * Returns all components of the module
	 */
	public List<Component> components() {
		return components;
	}
	
	/**
	 * Returns last component of the module.
	 */
	public Component lastComponent() {
		return components.get(components.size()-1);
	}
	
	public void setImportedModules(List<Module> modules) {
		this.importedModules = modules;
	}
	
	public List<Module> getImportedModules() {
		return importedModules;
	}
	
	/**
	 * Adds sees machine link with its corresponding imports links path starting
	 * from implementation of this module toward to the seen machine. This module
	 * must be first common ancestor of source and target machine of <code>seesLink</code>
	 * sees link.
	 *  
	 * @param seesLink sees link
	 * @param path list of the imports links
	 */
	public void addImportPath(SeenMachine seesLink, List<Import> path) {
		map.put(seesLink, path);
	}
	
	/**
	 * Returns corresponding imports path for sees link <code>seesLink</code>. 
	 * @param seesLink
	 */
	public List<Import> getImportPath(SeenMachine seesLink) {
		return map.get(seesLink);
	}
	
	/**
	 * Adds dependency between <code>source</code> and <code>target</code> machines.
	 * Both machines must be directly imported by this module implementation.
	 * This dependency must be caused by sees link, which source component must
	 * be <code>source</code> or its descendant in imports subgraph, and which
	 * target machine must be <code>source</code> or its descendant in imports subgraph.
	 * This module must also be the first common ancestor of source and target machine
	 * of that sees link.
	 * 
	 * @param source machine of dependency, directly imported by implementation of this module
	 * @param target machine of dependency, directly imported by implementation of this module
	 */
	public void addSeesDependency(Machine source, Machine target) {
		Couple<Machine, Machine> dependency = new Couple<Machine, Machine>(source, target);
		if (!seesDependencies.contains(dependency)) {
			seesDependencies.add(dependency);
		}
	}
	
	/**
	 * Returns dependencies between directly imported machines by implementation
	 * of this module.
	 */
	public List<Couple<Machine, Machine>> seesDependencies() {
		return seesDependencies;
	}
	
	/**
	 * Returns specified imports links sorted by dependencies introduced by sees links.
	 * @param imports imports link to sort
	 */
	public List<Import> sortedImports(List<Import> imports) {
		List<Couple<Import, Integer>> dependencies = new ArrayList<Couple<Import,Integer>>(seesDependencies.size());
		for (Import imported : imports) {
			
			int counter = 0;
			Machine mch = imported.machine();
			dependencyCycle: while (mch != null) {
				if (counter > seesDependencies.size())
					throw new IllegalStateException("Cycle in Sees!");
				for (Couple<Machine, Machine> dependency : seesDependencies) {
					if (mch == dependency.first()) {
						counter++;
						mch = dependency.second();
						continue dependencyCycle;
					}
				}
				mch = null;
			}
			dependencies.add(new Couple<Import, Integer>(imported, counter));
		}
		Collections.sort(dependencies, new Comparator<Couple<Import, Integer>>() {
			@Override
			public int compare(Couple<Import, Integer> arg0, Couple<Import, Integer> arg1) {
				return arg0.second()-arg1.second();
			}});
		List<Import> sorted = new ArrayList<Import>();
		for (Couple<Import, Integer> dependency : dependencies) {
			sorted.add(dependency.first());
		}
		return sorted;
	}
	
	/**
	 * Returns imports links sorted by dependencies introduced by sees links.
	 */
	public List<Import> sortedImports() {
		List<Couple<Import, Integer>> dependencies = new ArrayList<Couple<Import,Integer>>(seesDependencies.size());
		for (Import imported : implementation.imports()) {
			
			int counter = 0;
			Machine mch = imported.machine();
			dependencyCycle: while (mch != null) {
				if (counter > seesDependencies.size())
					throw new IllegalStateException("Cycle in Sees!");
				for (Couple<Machine, Machine> dependency : seesDependencies) {
					if (mch == dependency.first()) {
						counter++;
						mch = dependency.second();
						continue dependencyCycle;
					}
				}
				mch = null;
			}
			dependencies.add(new Couple<Import, Integer>(imported, counter));
		}
		Collections.sort(dependencies, new Comparator<Couple<Import, Integer>>() {
			@Override
			public int compare(Couple<Import, Integer> arg0, Couple<Import, Integer> arg1) {
				return arg0.second()-arg1.second();
			}});
		List<Import> sorted = new ArrayList<Import>();
		for (Couple<Import, Integer> dependency : dependencies) {
			sorted.add(dependency.first());
		}
		return sorted;
	}
	
	public void addSeenReferenceDelegation(SeenMachine seenMchine) {
		for (SeenMachine seen : seesReferencesDelegations) {
			if (seen.prefixString().equals(seenMchine.prefixString()) && seen.machine() == seenMchine.machine())
				return;
		}
		seesReferencesDelegations.add(seenMchine);
	}
	
	public List<SeenMachine> seesReferencesDelegations() {
		return seesReferencesDelegations;
	}
	
	@Override
	public String toString() {
		return "Module: "+machine.name();
	}
}
