package sk.tuke.bkpi.model.b0data;

import static sk.tuke.bkpi.model.types.TypeBuilder.CartesianProduct;
import static sk.tuke.bkpi.model.types.TypeBuilder.PowerSet;

import java.util.List;

import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.types.CartesianProduct;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeResolver;

public class ArrayFieldVariable extends B0Variable {

	private B0Variable arrayVariable;
	private List<B0DataValue> indexes;
	
	private TypeResolver INDEXES = new TypeResolver();
	private TypeResolver ELEMENT_TYPE = new TypeResolver();
	private Type ARRAY_TYPE = PowerSet(CartesianProduct(INDEXES, ELEMENT_TYPE));
	
	public ArrayFieldVariable(B0Variable arrayVariable, List<B0DataValue> indexes) {
		this.arrayVariable = arrayVariable;
		this.indexes = indexes;

		arrayVariable.typeCheckAndResolve(ARRAY_TYPE);
		Type indexesType = INDEXES.resolvedType();
		
		for (B0DataValue index : indexes) {
			Type indexType;
			if (indexesType instanceof CartesianProduct) {
				indexType = ((CartesianProduct) indexesType).first();
				indexesType    = ((CartesianProduct) indexesType).second();
			} else {
				indexType = indexesType;
				indexesType = null;
			}
			
			index.typeCheck(indexType);
		}
		type = ELEMENT_TYPE.resolvedType();
	}
	
	public B0Variable array() {
		return arrayVariable;
	}
	
	public List<B0DataValue> indexes() {
		return indexes;
	}
	
	@Override
	public String toString() {
		return arrayVariable+indexes.toString()+" (type: "+type+")";
	}
	
}
