package sk.tuke.bkpi.model.clause;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.OperationInputParameter;
import sk.tuke.bkpi.model.data.OperationOutputParameter;
import sk.tuke.bkpi.model.data.SetParameter;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Type;

public class Import extends ComponentReference {
	
	private String instanceName;
	private Machine machine;
	private List<B0DataValue> parameters;
	List<OperationDeclaration> operationsDeclarations;
	
	public Import(String name, Machine machine, B0DataValue ... parameters) {
		this(name, machine, Arrays.asList(parameters));
	}
	
	public Import(String name, Machine machine, List<B0DataValue> parameters) {
		super(name, machine);
		this.instanceName = name;
		this.machine = machine;
		this.parameters = parameters != null ? parameters : Collections.<B0DataValue>emptyList();
		
		operationsDeclarations = new ArrayList<OperationDeclaration>();
		for (Operation operation : machine.operations()) {
			List<OperationInputParameter> actualisedInputParameters = new ArrayList<OperationInputParameter>();
			List<OperationOutputParameter> actualisedOutputParameters = new ArrayList<OperationOutputParameter>();
			for (OperationInputParameter inParam : operation.inputParameters()) {
				OperationInputParameter inputParam = inParam;
				System.out.println(this.parameters);
				for (SetParameter setParam : machine.setParameters()) {
					Type t1 = ((PowerSet) setParam.type()).baseType();
					if (inputParam.type().contains(t1)) {
						System.out.println("\t"+setParam);
						System.out.println("\t"+setParam.index());
						System.out.println("\t"+this.parameters.get(setParam.index()));
						Type t2 = ((PowerSet) this.parameters.get(setParam.index()).type()).baseType();
						Type newType = inputParam.type().replace(t1, t2);
						inputParam = inputParam.newData();
						inputParam.typeData(newType);
					}
				}
				actualisedInputParameters.add(inputParam);
			}
			for (OperationOutputParameter outParam : operation.outputParameters()) {
				OperationOutputParameter outputParam = outParam;
				for (SetParameter setParam : machine.setParameters()) {
					Type t1 = ((PowerSet) setParam.type()).baseType();
					if (outputParam.type().contains(t1)) {
						Type t2 = ((PowerSet) this.parameters.get(setParam.index()).type()).baseType();
						Type newType = outputParam.type().replace(t1, t2);
						outputParam = outputParam.newData();
						outputParam.typeData(newType);
					}
				}
				actualisedOutputParameters.add(outputParam);
			}
			
			operation.outputParameters();
			operationsDeclarations.add(new OperationDeclaration(this, operation.name(), actualisedInputParameters, actualisedOutputParameters));
		}
	}	
	
	public Machine machine() {
		return machine;
	}
	
	public List<B0DataValue> formalParametersValues() {
		return Collections.unmodifiableList(parameters);
	}
	
	public Collection<OperationDeclaration> operationsDeclarations() {
		return operationsDeclarations;
	}
	
	@Override
	public String toString() {
		return "Import: "+machine +"("+prefixString()+")";
	}
	
}
