package sk.tuke.bkpi.model.clause;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.data.AbstractVariable;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.DataValue;
import sk.tuke.bkpi.model.data.OperationInputParameter;
import sk.tuke.bkpi.model.data.OperationOutputParameter;
import sk.tuke.bkpi.model.data.SetParameter;
import sk.tuke.bkpi.model.expression.ExpressionReference;
import sk.tuke.bkpi.model.predicate.Predicate;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Type;

public class IncludedMachine extends ComponentReference {

	private Machine machine;
	private List<DataValue> parameters;
	
	private Predicate includedInvariant; 
	
	private List<AbstractVariable> includedAbstractVariables = new ArrayList<AbstractVariable>();
	private List<ConcreteVariable> includedConcreteVariables = new ArrayList<ConcreteVariable>();
	
	private List<Operation> includedOperations = new ArrayList<Operation>();
	
	public IncludedMachine(Machine machine, List<DataValue> parameters) {
		this(null, machine, parameters);
	}
	public IncludedMachine(String name, Machine machine, List<DataValue> parameters) {
		super(machine);
		this.machine = machine;
		this.parameters = parameters;
		name = name == null? "" : name+".";
		
		System.out.println("INCLUDING MACHINE: "+machine.name());
		
		Map<Object, Object> substitution = new Hashtable<Object, Object>();
		Map<Type, Type> typeSubstitution = new Hashtable<Type, Type>();
		
		// replace formal parameters
		int index = 0;
		for (DataField formalParam : machine.formalParameters()) {
			if (formalParam instanceof SetParameter) {
				DataValue value = parameters.get(index);
				substitution.put(ExpressionReference.newReference(formalParam), value);
				typeSubstitution.put(((PowerSet)formalParam.type()).baseType(), ((PowerSet)value.type()).baseType());
			}
			index++;
		}
		
		for (Operation operation : machine.operations()) {
			List<OperationInputParameter> newInputParameters = new ArrayList<OperationInputParameter>();
			for (OperationInputParameter inParam : operation.inputParameters()) {
				Type newType = inParam.type();
				for (Type type : typeSubstitution.keySet()) {
					if (inParam.type().contains(type)) {
						newType = newType.replace(type, typeSubstitution.get(type));
					}
				}
				if (newType != inParam.type()) {
					OperationInputParameter newInParam = new OperationInputParameter(inParam.name());
					newInParam.typeData(newType);
					newInputParameters.add(newInParam);
				} else {
					newInputParameters.add(inParam);
				}
			}
			List<OperationOutputParameter> newOutputParameters = new ArrayList<OperationOutputParameter>();
			for (OperationOutputParameter outParam : operation.outputParameters()) {
				Type newType = outParam.type();
				for (Type type : typeSubstitution.keySet()) {
					if (outParam.type().contains(type)) {
						newType = newType.replace(type, typeSubstitution.get(type));
					}
				}
				if (newType != outParam.type()) {
					OperationOutputParameter newOutParam = new OperationOutputParameter(outParam.name());
					newOutParam.typeData(newType);
					newOutputParameters.add(newOutParam);
				} else {
					newOutputParameters.add(outParam);
				}
			}
			includedOperations.add(new Operation(operation.name(), newInputParameters, newOutputParameters, operation.body()));
		}
		
		
		// replace all variables with new one without type, so they will be typed in new invariant
		for (AbstractVariable variable : machine.abstractVariables()) {
			AbstractVariable includedVariable = new AbstractVariable(name+variable.name());
			includedAbstractVariables.add(includedVariable);
			substitution.put(variable, includedVariable);
		}
		for (ConcreteVariable variable : machine.concreteVariables()) {
			ConcreteVariable includedVariable = new ConcreteVariable(name+variable.name());
			includedConcreteVariables.add(includedVariable);
			substitution.put(variable, includedVariable);
		}
		
		// replace only variables containing formal set parameter in type
		/*
		for (AbstractVariable variable : machine.abstractVariables()) {
			boolean containsSetParameterType = false;
			for (SetParameter setParameter : machine.setParameters()) {
				if (variable.type().contains(((PowerSet) setParameter.type()).baseType())) {
					AbstractVariable includedVariable = new AbstractVariable(variable.name());
					includedVariables.add(includedVariable);
					substitution.put(variable, includedVariable);
					containsSetParameterType = true;
					//substitution.put(ExpressionReference.newReference(variable), ExpressionReference.newReference(includedVariable));
				}	
			}
			 if (!containsSetParameterType) {
				System.out.println("^^^^ ADD VARIABLE "+variable);
				includedVariables.add(variable);
			}
		}
		*/
		/*
		int index = 0;
		for (DataField formalParam : machine.formalParameters()) {
			if (formalParam instanceof SetParameter) {
				System.out.println("ABSTRACT VARS: "+machine.abstractVariables());
				for (AbstractVariable variable : machine.abstractVariables()) {
					if (variable.type().contains(((PowerSet) formalParam.type()).baseType())) {
						AbstractVariable includedVariable = new AbstractVariable(variable.name());
						includedVariables.add(includedVariable);
						substitution.put(variable, includedVariable);
						//substitution.put(ExpressionReference.newReference(variable), ExpressionReference.newReference(includedVariable));
					} else {
						System.out.println("^^^^ ADD VARIABLE "+variable);
						includedVariables.add(variable);
					}
				}
				
				Object value = parameters.get(index);
				
				String msg = String.format("REPLACE %s BY %s", ExpressionReference.newReference(formalParam), value);
				String msg2 = String.format("REPLACE %s BY %s", formalParam.getClass(), value.getClass());
				//System.out.println(msg);
				//System.out.println(msg2);
				substitution.put(ExpressionReference.newReference(formalParam), value);
			}
			index++;
		}
		*/
		System.out.println("template invariant: "+machine.invariant());
		System.out.println("Substitution: "+substitution);
		
		if (machine.invariant() != null) {
			includedInvariant = machine.invariant().instance(substitution);
			System.out.println("included invariant: "+includedInvariant);
		}
		
		System.out.println("INCLUDED VARIABLES");
		for (DataField variable : includedAbstractVariables) {
			System.out.println("variable: "+variable);
		}
		
		System.out.println("END OF INCLUDING MACHINE "+machine().name());
		System.out.println("////////////////");
	}
	
	public Machine machine() {
		return machine;
	}
	
	public List<DataValue> parameters() {
		return parameters;
	}
	
	public Predicate includedInvariant() {
		return includedInvariant;
	}
	
	public List<AbstractVariable> includedAbstractVariables() {
		return includedAbstractVariables;
	}
	public List<ConcreteVariable> includedConcreteVariables() {
		return includedConcreteVariables;
	}
	
	public List<Operation> includedOperations() {
		return includedOperations;
	}
}
