package sk.tuke.bkpi.model.clause;

import java.util.Collections;
import java.util.List;

import sk.tuke.bkpi.model.Identifier;
import sk.tuke.bkpi.model.data.OperationInputParameter;
import sk.tuke.bkpi.model.data.OperationOutputParameter;
import sk.tuke.bkpi.model.substitution.Substitution;
import sk.tuke.bkpi.model.types.TypeCheckException;

public class Operation {

	private String name;
	private List<OperationInputParameter> inputParameters;
	private List<OperationOutputParameter> outputParameters;
	private Substitution body;
	
	public Operation(String name,
			List<OperationInputParameter> inputParameters,
			List<OperationOutputParameter> outputParameters,
			Substitution body) {
		this.name = name;
		this.inputParameters  = inputParameters  != null? inputParameters : Collections.<OperationInputParameter>emptyList();
		this.outputParameters = outputParameters != null? outputParameters : Collections.<OperationOutputParameter>emptyList();;
		for (OperationInputParameter inParam : this.inputParameters) {
			if (!inParam.isTyped())
				throw new TypeCheckException("Operation input parameter: "+inParam.name()+" is not typed");
		}
		for (OperationOutputParameter outParam : this.outputParameters) {
			if (!outParam.isTyped())
				throw new TypeCheckException("Operation output parameter: "+outParam.name()+" is not typed");
		}
		this.body = body;
	}
	
	public String name() {
		return name;
	}
	
	public List<OperationInputParameter> inputParameters() {
		return inputParameters;
	}
	
	public List<OperationOutputParameter> outputParameters() {
		return outputParameters;
	}
	
	public Substitution body() {
		return body;
	}
}
