package sk.tuke.bkpi.model.component;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sk.tuke.bkpi.generator.Utilities;
import sk.tuke.bkpi.model.clause.EnumeratedSet;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.clause.Set;
import sk.tuke.bkpi.model.data.AbstractConstant;
import sk.tuke.bkpi.model.data.ConcreteConstant;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.ScalarParameter;
import sk.tuke.bkpi.model.data.SetParameter;
import sk.tuke.bkpi.model.predicate.Predicate;

public abstract class Component {
	
	protected Header header;
	protected List<Set> sets;
	protected List<SeenMachine> seenMachines;
	protected List<ConcreteConstant> concreteConstants;
	protected List<ConcreteVariable> concreteVariables;
	protected Predicate properties;
	protected Predicate invariant;
	protected List<OperationDeclaration> promotedOperations;
	
	public Component(Header header,
			List<SeenMachine> seenMachines,
			List<Set> sets,
			List<ConcreteConstant> concreteConstants,
			List<ConcreteVariable> concreteVariables,
			Predicate properties,
			Predicate invariant) {
		this.header = header;
		this.sets = sets != null? sets : new ArrayList<Set>();
		this.seenMachines = seenMachines != null? seenMachines : Collections.<SeenMachine>emptyList();
		this.concreteConstants = concreteConstants != null? concreteConstants : new ArrayList<ConcreteConstant>();
		this.concreteVariables = concreteVariables != null? concreteVariables : new ArrayList<ConcreteVariable>();
		this.properties = properties;
		this.invariant = invariant;
	}
	
	public List<DataField> formalParameters() {
		return header.parameters();
	}
	
	public List<SetParameter> setParameters() {
		return Utilities.filter(header.parameters(), SetParameter.class);
	}
	
	public List<ScalarParameter> scalarParameters() {
		return Utilities.filter(header.parameters(), ScalarParameter.class);
	}
	
	public String name() {
		return header.name();
	}
	
	public List<SeenMachine> seenMachines() {
		return Collections.unmodifiableList(seenMachines);
	}
	
	public List<Set> sets() {
		return Collections.unmodifiableList(sets);
	}
	
	public List<EnumeratedSet> enumeratedSets() {
		return Utilities.filter(sets, EnumeratedSet.class);
	}
	
	public List<ConcreteConstant> concreteConstants() {
		return concreteConstants;
	}
	
	public List<ConcreteVariable> concreteVariables() {
		return concreteVariables;
	}
	
	public Predicate getProperties() {
		return properties;
	}
	
	public Predicate invariant() {
		return invariant;
	}
	
	public Predicate properties() {
		return properties;
	}
	
	public List<OperationDeclaration> promotedOperations() {
		return promotedOperations;
	}
	
	public DataField dataByName(String name) {
		for (ConcreteConstant constant : concreteConstants) {
			if (constant.name().equals(name))
				return constant;
		}
		for (Set set : sets) {
			if (set.name().equals(name))
				return set;
		}
		return null;
	}
}
