package sk.tuke.bkpi.model.component;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.model.Identifier;
import sk.tuke.bkpi.model.clause.DeferredSet;
import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.OperationB0;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.clause.Set;
import sk.tuke.bkpi.model.clause.Values;
import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.data.ConcreteConstant;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.DataValue;
import sk.tuke.bkpi.model.instruction.Instruction;
import sk.tuke.bkpi.model.predicate.Predicate;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Type;

public class Implementation extends Component {

	private Machine abstractMachine;
	private AbstractComponent refinedComponent;
	private List<Import> extendedMachines;
	private List<Import> imports;
	//private Values valuations;
	private Map<DataField, B0DataValue> valuations;
	public Map<String, B0DataValue> valuations2 = new Hashtable<String, B0DataValue>();
	private Instruction initialization;
	private List<OperationB0> operations;
	
	private Map<Type, Type> typesSubstitutions = new Hashtable<Type, Type>();
	
	public Implementation(Header header,
			AbstractComponent refinedComponent,
			List<Import> extendedMachines,
			List<Import> imports,
			List<SeenMachine> seenMachines,
			List<Set> sets,
			List<ConcreteConstant> concreteConstants,
			List<ConcreteVariable> concreteVariables,
			Predicate properties,
			Predicate invariant,
			Map<DataField, B0DataValue> valuations,
			Instruction initialization,
			List<OperationB0> operations,
			List<OperationDeclaration> promotedOperations) {
		super(header, seenMachines, sets, concreteConstants, concreteVariables, properties, invariant);
		this.refinedComponent = refinedComponent;
		this.extendedMachines =  extendedMachines != null? extendedMachines : Collections.<Import>emptyList();
		this.imports =  imports != null? imports : Collections.<Import>emptyList();
		//this.valuations = valuations != null? valuations : Collections.<OperationB0>emptyList();
		this.valuations = valuations != null? valuations : Collections.<DataField, B0DataValue>emptyMap();
		this.initialization = initialization;
		this.operations = operations != null? operations : Collections.<OperationB0>emptyList();
		this.promotedOperations = promotedOperations != null? promotedOperations : Collections.<OperationDeclaration>emptyList(); // TODO move to Component
		
		List<Refinement> refinements = new ArrayList<Refinement>();
		//Component refinedComponent = implementation.getRefinedComponent();
		while (refinedComponent instanceof Refinement) {
			refinements.add((Refinement) refinedComponent);
			refinedComponent = ((Refinement)refinedComponent).refinedComponent();
		}
		abstractMachine = (Machine) refinedComponent;
		
		for (DataField data: this.valuations.keySet()) {
			valuations2.put(data.name(), this.valuations.get(data));
			if (data instanceof DeferredSet) {
				Type t1 = ((PowerSet) data.type()).baseType();
				Type t2 = ((PowerSet) valuations.get(data).type()).baseType();
				typesSubstitutions.put(t1, t2);
			}
		}
				
		/* 
		for (String dataName : this.valuations.keySet()) {
			DataField data = dataByName(dataName);
			System.out.println("VALUATION "+data+ " = "+valuations.get(dataName));
			if (data != null) {
				data.typeData(valuations.get(dataName).type());
			}
		}
		*/
	}
	
	public Map<Type, Type> typesSubstitutions() {
		return typesSubstitutions;
	}
	
	public AbstractComponent refinedComponent() {
		return refinedComponent;
	}
	
	public List<Import> extendedMachines() {
		return extendedMachines;
	}
	
	public List<Import> imports() {
		return Collections.unmodifiableList(imports);
	}
	
	public B0DataValue valueOf(DataField data) {
		return valuations.get(data.name());
	}
	
	public B0DataValue valueOf(String constantName) {
		return valuations.get(constantName);
	}
	
	public Map<DataField, B0DataValue> valuations() {
		return valuations;
	}
	
	public Machine abstractMachine() {
		return abstractMachine;
	}
	
	public Instruction initialization() {
		return initialization;
	}
	
	//public Values getValues() {
	//	return valuations;
	//}
	
	/*
	public Valuation getValuationOf(String name) {
		for (Valuation valuation : valuations) {
			if (valuation.getConstant().equals(name))
				return valuation;
		}
		return null;
	}
	*/
	public List<OperationB0> getOperations() {
		return operations;
	}
	
	@Override
	public String toString() {
		return "Implementation: " + header.name();
	}
	
}
