package sk.tuke.bkpi.model.component;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import sk.tuke.bkpi.model.clause.IncludedMachine;
import sk.tuke.bkpi.model.clause.Operation;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.clause.Set;
import sk.tuke.bkpi.model.data.AbstractConstant;
import sk.tuke.bkpi.model.data.AbstractVariable;
import sk.tuke.bkpi.model.data.ConcreteConstant;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.predicate.BinaryPredicate;
import sk.tuke.bkpi.model.predicate.Predicate;
import sk.tuke.bkpi.model.predicate.BinaryPredicate.Operator;

public class Machine extends Component implements AbstractComponent {
	
	private List<AbstractConstant> abstractConstants;
	private List<AbstractVariable> abstractVariables;
	private List<IncludedMachine> includedMachines;
	private List<IncludedMachine> extendedMachines;
	private Predicate constraints;
	private List<Operation> operations;
	
	private List<Set> includedSets = new ArrayList<Set>();
	
	public Machine(Header header,
			Predicate constraints,
			List<SeenMachine> seenMachines,
			List<IncludedMachine> includedMachines,
			List<IncludedMachine> extendedMachines,
			List<Set> sets,
			List<ConcreteConstant> concreteConstants,
			List<AbstractConstant> abstractConstants,
			List<ConcreteVariable> concreteVariables,
			List<AbstractVariable> abstractVariables,
			Predicate properties,
			Predicate invariant,
			List<Operation> operations) {
		super(header, seenMachines, sets, concreteConstants, concreteVariables, properties, invariant);
		this.includedMachines  = includedMachines  != null? includedMachines  : Collections.<IncludedMachine>emptyList();
		this.extendedMachines  = extendedMachines  != null? extendedMachines  : Collections.<IncludedMachine>emptyList();
		this.abstractConstants = abstractConstants != null? abstractConstants : new ArrayList<AbstractConstant>();
		this.abstractVariables = abstractVariables != null? abstractVariables : new ArrayList<AbstractVariable>();
		this.constraints = constraints;
		this.operations = operations != null? operations : Collections.<Operation>emptyList();
		
		for (IncludedMachine includedMachine : this.extendedMachines) {
			this.abstractVariables.addAll(includedMachine.includedAbstractVariables());
			this.concreteVariables.addAll(includedMachine.includedConcreteVariables());
			
			this.sets.addAll(includedMachine.machine().sets);
			this.abstractConstants.addAll(includedMachine.machine().abstractConstants);
			this.concreteConstants.addAll(includedMachine.machine().concreteConstants);
			this.invariant = new BinaryPredicate(Operator.CONJUNCTION, this.invariant, includedMachine.includedInvariant());
		}
		for (IncludedMachine includedMachine : this.includedMachines) {
			this.abstractVariables.addAll(includedMachine.includedAbstractVariables());
			System.out.println(includedMachine.includedConcreteVariables());
			this.concreteVariables.addAll(includedMachine.includedConcreteVariables());
			
			
			includedSets.addAll(includedMachine.machine().sets());
			//this.sets.addAll(includedMachine.machine().sets());
			this.abstractConstants.addAll(includedMachine.machine().abstractConstants);
			this.concreteConstants.addAll(includedMachine.machine().concreteConstants);
			this.invariant = new BinaryPredicate(Operator.CONJUNCTION, this.invariant, includedMachine.includedInvariant());
			this.operations.addAll(includedMachine.includedOperations());
			//System.out.println("INCLUDED VARIABLES: "+includedMachine.includedAbstractVariables());
		}
		//System.out.println("FINAL VARIABLES: "+this.abstractVariables);
		//System.out.println("FINAL INVARIANT: "+this.invariant);
	}

	public List<Set> includedSets() {
		return includedSets;
	}
	
	public Predicate getConstraints() {
		return constraints;
	}
	
	public List<IncludedMachine> includedMachines() {
		return includedMachines;
	}
	
	public List<IncludedMachine> extendedMachines() {
		return extendedMachines;
	}
	
	@Override
	public List<AbstractConstant> abstractConstants() {
		return abstractConstants;
	}
	
	@Override
	public List<AbstractVariable> abstractVariables() {
		return abstractVariables;
	}
	
	public List<Operation> operations() {
		return operations;
	}
	
	@Override
	public String toString() {
		return "Machine: " + header.name();
	}
}
