package sk.tuke.bkpi.model.expression.function;

import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeResolver;
import static sk.tuke.bkpi.model.types.TypeBuilder.*;

/**
 * Evaluation of the function. Definition: argument &isin; dom(function) &and;
 * function &isin; dom(function) &rarr; ran(function)
 */
public class FunctionEvaluation extends Expression {

	TypeResolver T1 = new TypeResolver();
	TypeResolver T2 = new TypeResolver();
	Type FUNCTION_TYPE = PowerSet(CartesianProduct(T1, T2));
	
	private Expression function;
	private Expression argument;
	
	/**
	 * Evaluation of the function, which type will be T2
	 * @param function of type P(T1xT2)
	 * @param argument of type T1
	 * 
	 * @throws TypeCheckException
	 */
	public FunctionEvaluation(Expression function, Expression argument) {
		function.typeCheckAndResolve(FUNCTION_TYPE);
		argument.typeCheck(T1.resolvedType());
		type = T2.resolvedType();
		this.function = function;
		this.argument = argument;
	}
	
	@Override
	public String toString() {
		return "FunctionEvaluation [function: "+function+ " argument: "+argument;
	}
}
