package sk.tuke.bkpi.model.expression.relation;

import static sk.tuke.bkpi.model.types.TypeBuilder.CartesianProduct;
import static sk.tuke.bkpi.model.types.TypeBuilder.PowerSet;
import sk.tuke.bkpi.model.expression.BinaryExpression;
import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.types.CartesianProduct;
import sk.tuke.bkpi.model.types.GenericType;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeCheckException;
import sk.tuke.bkpi.model.types.TypeResolver;

public class Composition extends BinaryExpression {

	private static TypeResolver T = new TypeResolver();
	private static TypeResolver U1 = new TypeResolver();
	private static TypeResolver U2 = new TypeResolver();
	private static TypeResolver V = new TypeResolver();
	private static Type RELATION_TYPE1 = PowerSet(CartesianProduct(T, U1));
	private static Type RELATION_TYPE2 = PowerSet(CartesianProduct(U2, V));
	
	/**
	 * Composition expression of type P(TxV)
	 * 
	 * B Reference Manual page 55
	 * @param left expression of type P(TxU)
	 * @param right expression of type P(UxV)
	 * 
	 * @throws TypeCheckException
	 */
	public Composition(Expression left, Expression right) {
		super(left, right);
		
		Type XT = GenericType.GENERIC;
		Type XU1 = GenericType.GENERIC;
		Type XU2 = GenericType.GENERIC;
		Type XV = GenericType.GENERIC;
		
		if (left.type().isCompatibile(PowerSet(CartesianProduct(GenericType.GENERIC, GenericType.GENERIC)))) {
			if (left.type() instanceof PowerSet &&
				((PowerSet) left.type()).baseType() instanceof CartesianProduct) {
					XT = ((CartesianProduct) ((PowerSet) left.type()).baseType()).first();
					XU1 = ((CartesianProduct) ((PowerSet) left.type()).baseType()).second();
			}
		} else {
			throw new TypeCheckException("Not relation type ...");
		}
		
		if (right.type().isCompatibile(PowerSet(CartesianProduct(GenericType.GENERIC, GenericType.GENERIC)))) {
			if (right.type() instanceof PowerSet &&
				((PowerSet) right.type()).baseType() instanceof CartesianProduct) {
					XU2 = ((CartesianProduct) ((PowerSet) right.type()).baseType()).first();
					XV = ((CartesianProduct) ((PowerSet) right.type()).baseType()).second();
			}
		} else {
			throw new TypeCheckException("Not relation type ...");
		}
		
		/*
		if (left.type() instanceof PowerSet) {
			if (((PowerSet) left.type()).baseType() instanceof CartesianProduct) {
				XT = ((CartesianProduct) ((PowerSet) left.type()).baseType()).first();
				XU1 = ((CartesianProduct) ((PowerSet) left.type()).baseType()).second();
			} else if (!(((PowerSet) left.type()).baseType() instanceof GenericType)) {
				throw new TypeCheckException("Not relation type ...");
			}
		} else if (!(left.type() instanceof GenericType)) {
			throw new TypeCheckException("Not relation type ...");
		}
		if (right.type() instanceof PowerSet) {
			if (((PowerSet) right.type()).baseType() instanceof CartesianProduct) {
				XT = ((CartesianProduct) ((PowerSet) right.type()).baseType()).first();
				XU2 = ((CartesianProduct) ((PowerSet) right.type()).baseType()).second();
			} else if (!(((PowerSet) right.type()).baseType() instanceof GenericType)) {
				throw new TypeCheckException("Not relation type ...");
			}
		} else if (!(right.type() instanceof GenericType)) {
			throw new TypeCheckException("Not relation type ...");
		}
		*/
		
		
		if (!XU1.isCompatibile(XU2))
			throw new TypeCheckException("Incompatibile relations types "+XU1 + " and "+XU2);
		type = PowerSet(CartesianProduct(XT, XV));
		
		if (!RELATION_TYPE1.isCompatibile(left.type()) ||
				!RELATION_TYPE2.isCompatibile(right.type())) {
			throw new TypeCheckException("Not relation type ...");
		}
		RELATION_TYPE1.resolveType(left.type());
		RELATION_TYPE2.resolveType(right.type());
		if (!U1.resolvedType().isCompatibile(U2.resolvedType())) {
			throw new TypeCheckException("Incompatibile relations types ...");
		}
		
		typeCheckLeft (RELATION_TYPE1);
		typeCheckRight(RELATION_TYPE2);
		if (!U1.resolvedType().isCompatibile(U2.resolvedType()))
			throw new TypeCheckException("Incompatibile relations types for composition: "+U1+" and "+U2);
		
		type = PowerSet(CartesianProduct(T.resolvedType(), V.resolvedType()));
	}
}
