package sk.tuke.bkpi.model.expression.relation;

import static sk.tuke.bkpi.model.types.TypeBuilder.CartesianProduct;
import static sk.tuke.bkpi.model.types.TypeBuilder.PowerSet;
import sk.tuke.bkpi.model.expression.BinaryExpression;
import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeCheckException;
import sk.tuke.bkpi.model.types.TypeResolver;

public class DirectProduct extends BinaryExpression {

	private static TypeResolver T1 = new TypeResolver();
	private static TypeResolver T2 = new TypeResolver();
	private static TypeResolver U = new TypeResolver();
	private static TypeResolver V = new TypeResolver();
	private static Type RELATION_TYPE1 = PowerSet(CartesianProduct(T1, U));
	private static Type RELATION_TYPE2 = PowerSet(CartesianProduct(T2, V));
	
	/**
	 * Direct Product type P(Tx(UxV))
	 * 
	 * @param left expression of type P(TxU)
	 * @param right expression of type P(TxV)
	 */
	public DirectProduct(Expression left, Expression right) {
		super(left, right);
		typeCheckLeft (RELATION_TYPE1);
		typeCheckRight(RELATION_TYPE2);
		if (!T1.resolvedType().isCompatibile(T2.resolvedType()))
			throw new TypeCheckException("Incompatibile types for Direct Product: "+T1+" and "+T2);
		type = PowerSet(CartesianProduct(T1.resolvedType(), CartesianProduct(U.resolvedType(), V.resolvedType())));
	}
}
