package sk.tuke.bkpi.model.instruction;

import java.util.List;

import sk.tuke.bkpi.model.Identifier;
import sk.tuke.bkpi.model.clause.OperationDeclaration;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.types.TypeCheckException;

public class CallupInstruction extends Instruction {

	private OperationDeclaration declaration;
	private List<DataField> assignVariables;
	private List<B0DataValue> arguments;
	
	public CallupInstruction(OperationDeclaration declaration, List<DataField> assignVariables, List<B0DataValue> arguments) {
		// local variables can be typed here 
		int variableIndex = 0;
		for (DataField variable : assignVariables) {
			if (!variable.isTyped()) {
				variable.typeData(declaration.outputParameters().get(variableIndex).type());
			} else {
				// TODO: check types
				if (!variable.type().isCompatibile(declaration.outputParameters().get(variableIndex).type())) {
					String errmsg = String.format(
							"variable %s has type: %s which is incompatibile with type: %s",
							variable, variable.type(), declaration.outputParameters().get(variableIndex).type());
					throw new TypeCheckException(errmsg);
				}
			}
			variableIndex++;
		}
		this.declaration = declaration;
		this.assignVariables = assignVariables;
		this.arguments = arguments;
	}
	
	public OperationDeclaration declaration() {
		return declaration;
	}
	
	public List<DataField> assignVariables() {
		return assignVariables;
	}
	
	public List<B0DataValue> arguments() {
		return arguments;
	}
}
