package sk.tuke.bkpi.model.instruction;

import java.util.List;

import sk.tuke.bkpi.model.data.B0DataValue;

public class CaseConditionalInstruction extends Instruction {
	
	private B0DataValue expression;
	private List<Branch> alternatives;
	private Instruction elseSubstitution;

	public CaseConditionalInstruction(B0DataValue expression, List<Branch> alternatives) {
		this.expression = expression;
		if (alternatives.size() >= 2) {
			this.alternatives = alternatives;
		} else {
			throw new IllegalArgumentException();
		}
	}
	
	public CaseConditionalInstruction(B0DataValue expression, List<Branch> alternatives, Instruction elseSubstitution) {
		this(expression, alternatives);
		this.elseSubstitution = elseSubstitution;
	}
		
	public B0DataValue expression() {
		return expression;
	}
	
	public List<Branch> branches() {
		return alternatives;
	}

	public Instruction elseSubstitution() {
		return elseSubstitution;
	}
	
	public static class Branch {
		private List<B0DataValue> literalConstants;
		private Instruction substitution;
	
		public List<B0DataValue> literalConstants() {
			return literalConstants;
		}
		
		public Instruction substitution() {
			return substitution;
		}
		
		public Branch(List<B0DataValue> literalConstants, Instruction substitution) {
			if (literalConstants.size() > 0) {
				this.literalConstants = literalConstants;
			} else {
				throw new IllegalArgumentException();
			}
			this.substitution = substitution;
		}
	}
}
