package sk.tuke.bkpi.model.predicate;

import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.expression.ExpressionReference;
import sk.tuke.bkpi.model.types.PowerSet;
import sk.tuke.bkpi.model.types.Type;
import sk.tuke.bkpi.model.types.TypeCheckException;
import sk.tuke.bkpi.model.types.TypeResolver;
import static sk.tuke.bkpi.model.types.TypeBuilder.*;

public class BelongingPredicate extends BinaryExpressionPredicate {
	
	public enum Operator {
		BELONGING,
		NON_BELONGING
	}
	
	private Operator operator;
	
	private TypeResolver T = new TypeResolver();
	private Type SET = PowerSet(T);
	
	public BelongingPredicate(Expression left, Operator operator, Expression right) {
		super(left, right);
		
		//typeCheck(right, SET);
		/*
		System.out.println("predicate " +left+" : "+right);
		System.out.println(left.getClass());
		System.out.println("types " +left.getType()+" : "+right.getType());
		*/
		
		
		if (!right.isTyped())
			throw new TypeCheckException("expression not typed yet: "+right);
		if (!right.type().isCompatibile(SET))
			throw new TypeCheckException("expected type: Powerset(?)");
		SET.resolveType(right.type());
		
		if (!left.isTyped() && left instanceof ExpressionReference && operator == Operator.BELONGING) {
			((ExpressionReference)left).getData().typeData(T.resolvedType());
		} else {
			if (!left.type().isCompatibile(T.resolvedType()))
				throw new TypeCheckException(left.type()+" not compatibile with "+T.resolvedType()+"\nleft:"+left+" right: "+right);
			if (left.type().inference(T.resolvedType()).containsGenricType())
				throw new TypeCheckException("predicate contains Generic Type: "+left.type().inference(T.resolvedType()).containsGenricType());
		}
		
		this.operator = operator;
	}
	
	public Operator getOperator() {
		return operator;
	}
	
	@Override
	public String toString() {
		//return "BelongsPredicate";
		return left+" belongs "+right;
	}

	@Override
	public BinaryExpressionPredicate newInstance(Expression left, Expression right) {
		return new BelongingPredicate(left, operator, right);
	}
}
