package sk.tuke.bkpi.model.predicate;

import java.util.Map;

import sk.tuke.bkpi.model.expression.BinaryExpression;
import sk.tuke.bkpi.model.expression.Expression;

public class BinaryPredicate extends Predicate {

	public enum Operator {
		CONJUNCTION,
		DISJUNCTION,
		IMPLICATION,
		EQUIVALENCE
	}
	private Predicate left;
	private Predicate right;
	private Operator operator;
	
	public BinaryPredicate(Operator operator, Predicate left, Predicate right) {
		this.operator = operator;
		this.left = left;
		this.right = right;
	}
	
	public Predicate left() {
		return left;
	}
	
	public Predicate right() {
		return right;
	}
	
	public Operator operator() {
		return operator;
	}
	
	@Override
	public String toString() {
		return String.format("[%s %s %s]",left, operator, right);
	}
	
	@Override
	public Predicate instance(Map<Object, Object> substitution) {
		System.out.println("TEST "+left+ " | "+right);
		Predicate leftInstance  = substitution.containsKey(left)? (Predicate) substitution.get(left) : left.instance(substitution);
		Predicate rightInstance = substitution.containsKey(right)? (Predicate) substitution.get(right) : right.instance(substitution);
		if (left == leftInstance && right == rightInstance) {
			return this;
		} else {
			return new BinaryPredicate(operator, leftInstance, rightInstance);
		}
	}
	
}
