package sk.tuke.bkpi.model.predicate;

import sk.tuke.bkpi.model.expression.Expression;
import sk.tuke.bkpi.model.expression.ExpressionReference;
import sk.tuke.bkpi.model.types.TypeCheckException;

public class InclusionPredicate extends BinaryExpressionPredicate {

	public enum Operator {
		INCLUSION,
		STRICT_INCLUSION,
		NON_INCLUSION,
		STRICT_NON_INCLUSION
	}
	
	private Operator operator;
	
	public InclusionPredicate(Expression left, Operator operator, Expression right) {
		super(left, right);
		if (left.type() == null) {
			if (left instanceof ExpressionReference) {
				((ExpressionReference)left).getData().typeData(right.type());
			} else {
				throw new TypeCheckException("wrong typing predicate");
			}
		} else if (!left.type().isCompatibile(right.type())) {
			throw new TypeCheckException("types are not compatibile: "+left.type()+" and "+right.type());
		}
		if (left.type().inference(right.type()).containsGenricType())
			throw new TypeCheckException("predicate contains Generic Type");
		
		this.operator = operator;
		
	}
	@Override
	public String toString() {
		return String.format("Inclusion Predicate:[%s %s %s]", left, operator, right);
	}
	@Override
	public BinaryExpressionPredicate newInstance(Expression left, Expression right) {
		return new InclusionPredicate(left, operator, right);
	}
}
