package sk.tuke.bkpi.model.term;

import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.types.BasicType;
import sk.tuke.bkpi.model.types.TypeCheckException;

public class ArithmeticalBinaryExpression extends B0DataValue {
	public enum Operator {
		PLUS(4),
		MINUS(4),
		TIMES(3),
		DIVIDE(3),
		MODULO(3),
		POWER(2);
		
		public final int priority;
		Operator(int priority) {
			this.priority = priority;
		}
	}
	
	private Operator operator;
	private B0DataValue left;
	private B0DataValue right;
	
	public ArithmeticalBinaryExpression(Operator operator, B0DataValue left, B0DataValue right) {
		if (!left.type().isEqual(BasicType.Z))
			throw new TypeCheckException("Left expression must be of type: "+BasicType.Z+", but is: "+left.type());
			
		if (!right.type().isEqual(BasicType.Z))
			throw new TypeCheckException("Right expression must be of type: "+BasicType.Z+", but is: "+right.type());
		
		this.operator = operator;
		this.left = left;
		this.right = right;
		type = BasicType.Z;
	}
	
	public B0DataValue left() {
		return left;
	}
	
	public B0DataValue right() {
		return right;
	}
	
	public Operator operator() {
		return operator;
	}
	/*
	public int evaluate() {
		switch (operator) {
		case PLUS: return evaluate(left)+evaluate(right);
			
		default:
			break;
		}
	}
	*/
	@Override
	public String toString() {
		return String.format("(%s %s %s)", left, operator, right);
	}
}
