package sk.tuke.bkpi.model.term;

import java.util.List;

import sk.tuke.bkpi.model.data.B0DataValue;
import sk.tuke.bkpi.model.types.Type;
import static sk.tuke.bkpi.model.types.TypeBuilder.*;

public class CoupleSet extends ArrayExpression {

	List<Couple> couples;
	
	public CoupleSet(List<Couple> couples) {
		if (couples == null)
			throw new NullPointerException("Couples must be!");
		if (couples.size() < 1)
			throw new IllegalArgumentException("Cluple list cannot be empty!");
		
		this.couples = couples;
		
		Type T = couples.get(0).type();
		for (Couple couple : couples) {
			if (!T.toString().equals(couple.type().toString()))
				throw new IllegalArgumentException("All couples must be of the same type");
		}
		type = PowerSet(T);
	}
	
	public List<Couple> couples() {
		return couples;
	}
	
	public static class Couple extends B0DataValue {
		
		//types: Couple, Term
		private B0DataValue left;
		private B0DataValue right;
		
		public Couple(B0DataValue left, B0DataValue right) {
			this.left  = left;
			this.right = right;
			type = CartesianProduct(left.type(), right.type());
		}
		
		public B0DataValue left() {
			return left;
		}
		
		public B0DataValue right() {
			return right;
		}
		
		@Override
		public String toString() {
			return left+"->"+right;
		}
	}
	
	@Override
	public String toString() {
		return "CoupleSet"+couples;
	}
}
