package sk.tuke.bkpi.model.types;

import static sk.tuke.bkpi.model.types.TypeBuilder.*;

public final class CartesianProduct extends Type {

	private Type first;
	private Type second;
	
	public static Type instance(Type first, Type second) {
		if (first instanceof PowerSet && second instanceof PowerSet)
			return new PowerSet(new CartesianProduct(((PowerSet)first).baseType(), ((PowerSet)second).baseType()));
		return new CartesianProduct(first, second);
	}
	
	CartesianProduct(Type first, Type second) {
		if (first == null)
			throw new NullPointerException("First type is null");
		if (second == null)
			throw new NullPointerException("Second type is null");
		
		this.first  = first;
		this.second = second;
	}
	
	public Type first() {
		return first;
	}
	
	public Type second() {
		return second;
	}
	
	@Override
	public boolean isEqual(Type t) {
		if (!(t instanceof CartesianProduct))
			return false;
		return first.isEqual(((CartesianProduct)t).first) && second.isEqual(((CartesianProduct)t).second);
	}
	
	@Override
	public final boolean isCompatibile(Type t) {
		if ((t instanceof CartesianProduct))
			return first.isCompatibile(((CartesianProduct)t).first) && second.isCompatibile(((CartesianProduct)t).second);
		return t instanceof GenericType;
	}
	
	@Override
	public boolean isSubtype(Type t) {
		if (t instanceof CartesianProduct) {
			return first.isSubtype(((CartesianProduct)t).first) && second.isSubtype(((CartesianProduct)t).second);
		}
		return t instanceof GenericType;
	}
	
	@Override
	public void resolveType(Type t) {
		super.resolveType(t);
		if (t instanceof CartesianProduct) {
			first.resolveType (((CartesianProduct)t).first );
			second.resolveType(((CartesianProduct)t).second);
		} else if (!(t instanceof GenericType)) {
			throw new TypeCheckException("incompatibile types");
		}
	}
	@Override
	public boolean containsGenricType() {
		return first.containsGenricType() || second.containsGenricType();
	}
	
	public boolean contains(Type type) {
		return isEqual(type) || first.contains(type) || second.contains(type);
	}
	
	@Override
	public Type inference(Type t) {
		if (t instanceof CartesianProduct)
			return CartesianProduct(first.inference(((CartesianProduct)t).first), second.inference(((CartesianProduct)t).second));
		if (t instanceof GenericType)
			return this;
		throw new TypeCheckException("union faild");
	}
	
	@Override
	public Type replace(Type oldType, Type newType) {
		if (isEqual(oldType))
			return newType;
		
		Type newFirst = first.replace(oldType, newType);
		Type newSecond = second.replace(oldType, newType);
		if (newFirst != first || newSecond != second) {
			System.out.println("create new CartesianProduct: "+CartesianProduct(newFirst, newSecond));
			return CartesianProduct(newFirst, newSecond);
		}
		return this;
	}
	/*
	@Override
	public final boolean isGeneric() {
		return first.isGeneric() && second.isGeneric();
	}
	*/
	@Override
	public String toString() {
		String f = first instanceof CartesianProduct ? "("+first+")" : first.toString();
		String s = second instanceof CartesianProduct ? "("+second+")" : second.toString();
		return f+" x "+s;
	}
}
