package sk.tuke.bkpi.model.types;

import static sk.tuke.bkpi.model.types.TypeBuilder.*;

public final class PowerSet extends Type {
	
	private Type set;
	
	PowerSet(Type type) {
		if (type == null)
			throw new NullPointerException("Type cannot be null");
		this.set = type;
	}
	
	public Type baseType() {
		return set;
	}
	
	@Override
	public boolean isEqual(Type t) {
		if (t instanceof PowerSet) {
			return set.isEqual(((PowerSet)t).baseType());
		}
		return false;
	}
	
	@Override
	public final boolean isCompatibile(Type t) {
		/*
		if (t instanceof PowerSet) {
			return type.isCompatibile(((PowerSet) t).baseType());
		}
		return false;
		*/
		return t instanceof PowerSet? set.isCompatibile(((PowerSet) t).baseType()) : t instanceof GenericType;
	}
	
	@Override
	public boolean isSubtype(Type t) {
		if (t instanceof PowerSet) {
			return set.isSubtype(((PowerSet)t).set);
		}
		return t instanceof GenericType;
	}
	
	@Override
	public void resolveType(Type t) {
		super.resolveType(t);
		if (t instanceof PowerSet) {
			//System.out.println("resolving Powerset "+type+".resolveType("+((PowerSet)t).type+")");
			set.resolveType(((PowerSet)t).set);
		} else if (!(t instanceof GenericType)) {
			throw new TypeCheckException("incompatibile types");
		}
	}
	
	@Override
	public boolean containsGenricType() {
		return set.containsGenricType();
	}
	
	public boolean contains(Type type) {
		return isEqual(type) || this.set.contains(type);
	}
	
	@Override
	public Type inference(Type t) {
		if (t instanceof PowerSet)
			return PowerSet(set.inference(((PowerSet)t).baseType()));
		if (t instanceof GenericType)
			return this;
		throw new TypeCheckException("inference of "+this+" and "+t+"faild");
	}
	
	@Override
	public Type replace(Type oldType, Type newType) {
		Type t = set.replace(oldType, newType);
		if (t != set)
			return PowerSet(t);
		else
			return super.replace(oldType, newType);
	}
	/*
	@Override
	public final boolean isGeneric() {
		return type.isGeneric();
	}
	*/
	@Override
	public String toString() {
		return "PowerSet("+set+")";
	}
}
