package sk.tuke.bkpi.model.types;

import java.util.ArrayList;
import java.util.List;

import sk.tuke.bkpi.generator.Utilities.ListFormatter;

public class Record extends Type {
	
	private List<Field> fields;
	
	public Record(List<Field> fields) {
		this.fields = fields;
	}
	
	public boolean hasField(String fieldName) {
		for (Field f : fields) {
			if (f.label.equals(fieldName))
				return true;
		}
		return false;
	}
	
	public Type fieldType(String fieldName) {
		for (Field f : fields) {
			if (f.label.equals(fieldName))
				return f.type;
		}
		return null;
	}
	
	/**
	 * 
	 */
	@Override
	public boolean isCompatibile(Type t) {
		if ((t instanceof Record)) {
			if (((Record)t).fields.size() == fields.size()) {
				int index = 0;
				for (Record.Field field : ((Record)t).fields) {
					if (!field.isCompatibile(fields.get(index)))
						return false;
					index++;
				}
				return true;
			}
		}
		return t instanceof GenericType;
	}
	
	@Override
	public boolean isEqual(Type t) {
		if (t instanceof Record) {
			if (((Record) t).fields().size() == fields.size()) {
				int index = 0;
				for (Record.Field field : ((Record)t).fields) {
					if (!field.isEqual(fields.get(index++)))
						return false;
				}
				return true;
			}
		}
		return false;
	}
	
	@Override
	public boolean isSubtype(Type t) {
		if (t instanceof Record) {
			int index = 0;
			// TODO if label is null
			for (Record.Field field : ((Record)t).fields) {
				if (!field.label.equals(fields.get(index)) || !field.type.isSubtype(fields.get(index).type)) {
					return false;
				}
				index++;
			}
			return true;
		}
		return t instanceof GenericType;
	}
	
	@Override
	public boolean containsGenricType() {
		for (Record.Field field : fields) {
			if (field.type.containsGenricType())
				return true;
		}
		return false;
	}
	
	public boolean contains(Type type) {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public void resolveType(Type t) {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public Type inference(Type t) {
		if (t instanceof Record) {
			if (!this.isEqual(t)) {
				int index = 0;
				List<Record.Field> newFields = new ArrayList<Field>();
				for (Record.Field field : ((Record)t).fields) {
					if (!field.label.equals(fields.get(index).label))
						throw new TypeCheckException("union faild: records labels are not equal: "+field.label+" and "+fields.get(index).label);
					newFields.add(new Record.Field(field.label, field.type.inference(fields.get(index).type)));
					index++;
				}
				return new Record(newFields);
			}
			return this;
		}
		if (t instanceof GenericType)
			return this;
		throw new TypeCheckException("type inference faild: T1: "+this+" T2: "+t);
	}
	
	public List<Field> fields() {
		return fields;
	}

	@Override
	public String toString() {
		ListFormatter fieldsList = new ListFormatter();
		for (Field field : fields) {
			fieldsList.addItem(field.label+":"+field.type);
		}
		return "Record(" +fieldsList.text()+ ")";
	}
	
	public static class Field {
		public final String label;
		public final Type type;
		
		public Field(String label, Type type) {
			this.label = label;
			this.type = type;
		}
		
		public boolean isCompatibile(Field field) {
			if (label != null && field.label != null) {
				return type.isCompatibile(field.type) && label.equals(field.label);
			}
			return type.isCompatibile(field.type);
		}
		
		public boolean isEqual(Field f) {
			//if (f.label == null) return type.isEqual(f.type);
			//return f.label.equals(label) && type.isEqual(f.type);
			return type.isEqual(f.type) && (f.label == null || label == null || f.label.equals(label));
		}
		
		@Override
		public String toString() {
			return "field; label="+label+" type="+type;
		}
	}
}
