package sk.tuke.bkpi.parser;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import sk.tuke.bkpi.model.Identifier;

/**
 * Data structure class that groups all preprocessed information.
 * <ul>
 *  <li>position information of particular clauses</li>
 *  <li>list of imported, extended and included machines</li>
 * </ul>
 * 
 * @author Marcel Dancak
 */
public class PreprocessData {

	public enum Clause {
		DEFINITIONS(0),
				
		REFINES(1),

		SEES(1),
		USES(1),
		SETS(1),
		
		ABSTRACT_CONSTANTS(3),
		CONCRETE_CONSTANTS(3),
		CONSTANTS(3),
		
		ABSTRACT_VARIABLES(3),
		CONCRETE_VARIABLES(3),
		VARIABLES(3),
		
		CONSTRAINTS(4),
		PROPERTIES(4),
		VALUES(4),//?
		
		INVARIANT(6),
		ASSERTIONS(6),
		
		IMPORTS(5),
		EXTENDS(5),
		INCLUDES(5),
		PROMOTES(7),
		
		INITIALISATION(8),
		LOCAL_OPERATIONS(9),
		OPERATIONS(10)
		;
		
		public final int orderNumber;
		Clause(int number) {
			this.orderNumber = number;
		}
	}
	
	public enum Chain {
		IMPORTS,
		INCLUDES,
		SEES,
		USES,
		EXTENDS,
		REFINES
	}
	
	private List<ClausePosition> clausePositions = new ArrayList<ClausePosition>();
	private Map<Chain, List<String>> chains = new Hashtable<Chain, List<String>>();
	private List<String> importedMachines = Collections.emptyList();
	
	private DefinitionsContext definitions;
	private List<Identifier> operations = new ArrayList<Identifier>();
	
	public void addClausePosition(Clause clause, int line, int column) {
		clausePositions.add(new ClausePosition(clause, line, column));
	}
	
	public void setImportedMachines(List<String> list) {
		this.importedMachines = list;
	}
	
	public List<String> importedMachines() {
		return importedMachines;
	}
	
	public void addChain(Chain chain, String machineName) {
		if (!chains.containsKey(chain)) {
			chains.put(chain, new ArrayList<String>());
		}
		chains.get(chain).add(machineName);
	}
	
	public Map<Chain, List<String>> getChains() {
		return chains;
	}
	
	/*
	public Position getClausePosition(Clause clause) {
		return clausePositions.get(clause);
	}
	*/
	public List<ClausePosition> getClausePositions() {
		return clausePositions;
	}
	
	public void setDefinitionContext(DefinitionsContext definitions) {
		this.definitions = definitions;
	}
	
	public DefinitionsContext getDefinitionContext() {
		return definitions;
	}
	
	public void addOperationName(Identifier name) {
		operations.add(name);
	}
	
	public void setOperationsNames(List<Identifier> operations) {
		this.operations = operations;
	}
	
	public List<Identifier> getOperationsNames() {
		return operations;
	}
	
	public static class ClausePosition {
		public final Clause clause;
		public final int line;
		public final int column;
		
		public ClausePosition(Clause clause, int line, int column) {
			this.clause = clause;
			this.line = line;
			this.column = column;
		}
		
		@Override
		public String toString() {
			return "Clause: "+clause +"("+clause.orderNumber+")"+ "["+line+","+column+"]";
			
		}
	}
}
